/*
 * Random.c
 */

#include "random.h"
#include "random_cfg.h"

/* Go see http://euclid.ucsd.edu/~kschneid/spring104/ */
#define a	175

#ifdef INT32_T

typedef union
{
    unsigned char[8]	b;
    INT32_T		x[2];
    long		l;
} rand_type;

#define random_cycle_0\
    if (((rand_type*)r)->x[0] < 0) \
        ((rand_type*)r)->x[0] = (((rand_type*)r)->x[0] << 1) ^ a;\
    else\
        ((rand_type*)r)->x[0] = (((rand_type*)r)->x[0] << 1);

#define random_cycle_1\
    if (((rand_type*)r)->x[1] < 0) \
        ((rand_type*)r)->x[1] = (((rand_type*)r)->x[1] << 1) ^ a;\
    else\
        ((rand_type*)r)->x[1] = (((rand_type*)r)->x[1] << 1);

#ifdef RAND_LE
#  define random_cycle_lo random_cycle_0
#  define random_cycle_hi random_cycle_1
#else
#  define random_cycle_lo random_cycle_1
#  define random_cycle_hi random_cycle_0
#endif

#else /* #ifdef INT32_T */

typedef union
{
    unsigned char[8]	b;
    unsigned int	x;
    long		l;
} rand_type;

#define random_cycle_lo \
{ \
    unsigned int x_lo = ((rand_type*)r)->x & 0x00000000ffffffffL; \
    unsigned int x_hi = ((rand_type*)r)->x & 0xffffffff00000000L; \
    if (x_lo & 0x0000000080000000L) \
	x_lo = (x_lo << 1) ^ a; \
    else \
	x_lo = (x_lo << 1); \
    ((rand_type*)r)->x = (x_hi | x_lo); \
}

#define random_cycle_hi \
{ \
    unsigned int x_lo = ((rand_type*)r)->x & 0x00000000ffffffffL; \
    unsigned int x_hi = ((rand_type*)r)->x & 0xffffffff00000000L; \
    if (x_hi & 0x8000000000000000L) \
	x_hi = (x_hi << 1) ^ (a << 32); \
    else \
	x_hi = (x_hi << 1); \
    ((rand_type*)r)->x = (x_hi | x_lo); \
}

#endif /* #ifdef INT32_T */

int random_seed(random_t *r, int seed_hi, int seed_lo)
{
#ifdef INT32_T
#  ifdef RAND_LE
    ((rand_type*)r)->x[0] = seed_lo;
    ((rand_type*)r)->x[1] = seed_hi;
#  else
    ((rand_type*)r)->x[0] = seed_hi;
    ((rand_type*)r)->x[1] = seed_lo;
#  endif
#else
    ((rand_type*)r)->x = ((seed_hi << 32) | (seed_lo & 0x00000000ffffffffL);
#endif
    return(1);
}

/*
 * random_int()
 */
int random_int(random_t *r)
{
    int i;
    
    random_cycle_lo;
#ifdef INT32_T
#  ifdef RAND_LE
    i = ((rand_type*)r)->x[0];
#  else
    i = ((rand_type*)r)->x[1];
#  endif
#else /* sign extend to 64 bits */
    i = (((rand_type*)r)->x) & 0x00000000ffffffffL;
    if (i & 0x0000000080000000)
	i |= 0xffffffff00000000;
#endif
    return (i);
}

/*
 * random_char()
 */
char random_char(random_t *r)
{
    random_cycle_lo;
#ifdef RAND_LE
    return (r->b[7]);
#else
    return (r->b[0]);
#endif
}

/*
 * random_short()
 */
short random_short(random_t *r)
{
    short i = random_int(r) & 0xffff;
#if   SIZEOF_SHORT == 4
    /* sign extend to 32 bits */
    if (i & 0x8000)
	i |= 0xffff0000;
#elif SIZEOF_SHORT == 8
    /* sign extend to 64 bits */
    if (i & 0x8000)
	i |= 0xffffffffffff0000;
#endif
    return (i);
}

/*
 * random_long()
 */
#ifndef ALLOW_64BIT_LONG
    long random_long(random_t *r)
    {
	return(random_int(r));
    }
#else
#  ifndef RAND_64
#    error 64 bit long not supported
#  else
    long random_long(random_t *r)
    {
	random_cycle_lo;
	random_cycle_hi;
	return (((rand_type*)r)->l);
    }
#  endif
#endif

/*
  These only work under IEEE.  Try a simpler
  (but slower) solution for portability.
  
  float random_float(random_t *r)
  {
  int i = random_int(r) & 0xff7fffff;    
  return(*((float*)&(i)));
  }
  
  double random_double(random_t *r)
  {
  int i[2];
  
  random_cycle_hi;
  random_cycle_lo;
  i[0] = r->x[0];
  i[1] = r->x[1];
  #ifdef RAND_LE
  i[1] &= 0xffefffff;
  #else
  i[0] &= 0xffefffff;
  #endif
  return(*((double*)i));
  }
 */

float random_float(random_t *r)
{
    int n = random_int(r);
    int d = random_int(r);

    if ((n == 0) || (d == 0))
	return (0.0);
    else
	return ((float)n / (float)d);
}

double random_double(random_t *r)
{
    int n = random_int(r);
    int d = random_int(r);

    if ((n == 0) || (d == 0))
	return (0.0);
    else
	return ((double)n / (double)d);
}
