dnl Get things in a better order than default  
AC_DEFUN(LAC_INIT,
[

AC_PREREQ(2.12)

AC_CONFIG_AUX_DIR(lac_config_path)

dnl The following two lines are stolen from AC_OUTPUT().
dnl We needed to hoist the setting of prefix and exec_prefix
dnl so that could get real values for things like bindir.
test "x$prefix" = xNONE && prefix=$ac_default_prefix
test "x$exec_prefix" = xNONE && exec_prefix='${prefix}'

dnl includedir needs to be architecture dependent, so rewrite it here
dnl to be relative to exec_prefix instead of prefix
if test "x${includedir}" = 'x${prefix}/include' ; then
    includedir='${exec_prefix}/include'
fi

AC_ARG_WITH(tmpdir,
        [  --with-tmpdir=DIR         temporary files in DIR [PREFIX/tmp]],
        [tmpdir="$withval"],
        [tmpdir='${prefix}/tmp'])

AC_SUBST(tmpdir)


AC_ARG_WITH(deploy_prefix,
        [  --with-deploy-prefix=DIR       root uid files in DIR [EPREFIX]],
        [deploy_prefix="$withval"],
        [deploy_prefix="${exec_prefix}"])

AC_SUBST(deploy_prefix)





AC_CANONICAL_HOST
AC_CANONICAL_TARGET
service=$target
echo "checking service system type... $service"

LAC_TARGET_ARCH

AC_PROG_LN_S


AC_PROG_INSTALL


dnl The found install program on the quad.mcs.anl.gov
dnl (/usr/bin/installbsd) is broken.  If the INSTALL 
dnl variable is set to this value, change it to the install
dnl script supplied by autoconf.

if test  "X$INSTALL" = "X/usr/bin/installbsd -c" ; then
   INSTALL="$ac_install_sh"
   echo reseting BSD compatible install... $INSTALL
fi



AC_PATH_PROGS(SH, sh bash)
AC_PATH_PROG(RM, rm)
AC_PATH_PROG(MKDIR, mkdir)
AC_PATH_PROG(CP, cp)
AC_PATH_PROG(SED, sed)
AC_PROG_AWK
AC_PATH_PROG(GREP, grep)
AC_PATH_PROG(NM, "nm")
AC_PATH_PROG(CHMOD, chmod)
AC_PATH_PROG(TOUCH, touch)
AC_PATH_PROGS(CUT, cut)

AC_PROG_MAKE_SET

OFILE="o"
AC_SUBST(OFILE)

topsrcdir="${srcdir}/lac_base_path"
topbuilddir="lac_base_path"
configurationdir="${srcdir}/lac_config_path"
AC_SUBST(topsrcdir)
AC_SUBST(topbuilddir)
AC_SUBST(configurationdir)

configurationfiles=""
for i in lac_config_files; do
    configurationfiles="$configurationfiles ${srcdir}/$i"
done
AC_SUBST(configurationfiles)

makefile_common_vars=$srcdir/lac_config_path/makefile.common.vars
AC_SUBST_FILE(makefile_common_vars)

makefile_common_rules=$srcdir/lac_config_path/makefile.common.rules
AC_SUBST_FILE(makefile_common_rules)

makefile_nexus_objs=$srcdir/lac_config_path/makefile.nexus.objs
AC_SUBST_FILE(makefile_nexus_objs)

makefile_gram_objs=$srcdir/lac_config_path/makefile.gram.objs
AC_SUBST_FILE(makefile_gram_objs)

makefile_gssapi_cleartext_objs=$srcdir/lac_config_path/makefile.gssapi_cleartext.objs
AC_SUBST_FILE(makefile_gssapi_cleartext_objs)

makefile_gssapi_spkm_objs=$srcdir/lac_config_path/makefile.gssapi_spkm.objs
AC_SUBST_FILE(makefile_gssapi_spkm_objs)

makefile_gssapi_objs=$srcdir/lac_config_path/makefile.gssapi_cleartext.objs
AC_SUBST_FILE(makefile_gssapi_objs)

])



dnl Fix up definintion of PROG_LEX so that it can find
dnl lex library.
AC_DEFUN(LAC_PROG_LEX,
[AC_PROVIDE([AC_PROG_LEX])
AC_CHECK_PROG(LEX, flex, flex, lex)
dnl Get full path to Lex that we are using
AC_PATH_PROG(ac_lex_path, $LEX)
ac_lex_path=`echo $ac_lex_path | sed s/$LEX//`
if test -z "$LEXLIB"
then
  case "$LEX" in
  flex*) ac_lib=fl ;;
  *) ac_lib=l 
         case "$target" in
	    *solaris2*) LEX_CXXFLAGS="-D__EXTERN_C__" ;;
	 esac
  ;;
  esac
  dnl Try to figure out correct path to lex library
  for ac_extension in a so sl; do
    for ac_dir in \
        /usr/local/lib \
        /usr/local/gnu/lib \
        /usr/gnu/lib \
        $ac_lex_path \
        $ac_lex_path/lib \
	`echo $ac_lex_path | sed s/bin/lib/` \
        ; \
    do 
      if test -r "$ac_dir/lib${ac_lib}.$ac_extension"; then
        ac_lex_lib=$ac_dir
        break
      fi
    done
  done
  AC_CHECK_LIB($ac_lib, yywrap, LEXLIB="-l$ac_lib")
fi
if test -n "$ac_lex_lib"; then 
   LEXLIB="-L$ac_lex_lib $LEXLIB"
fi
AC_SUBST(LEXLIB)
AC_SUBST(LEX_CXXFLAGS)
AC_SUBST(LEXDEFS)])

dnl Search for a library in a list of possible locations.  
dnl lex library.
AC_DEFUN(LAC_FIND_AND_CHECK_LIB,
[ac_link_save=$ac_link
 for ac_extension in a so sl; do
    for ac_dir in $3; do
      if test -r "$ac_dir/lib${ac_lib}.$ac_extension"; then
        ac_lib_dir=$ac_dir
        ac_link="$ac_link -L$ac_lib_dir"   
        break
      fi
    done
  done
  AC_CHECK_LIB($1, $2, [lac_lib_found=yes],[lac_lib_found=no]) 
  LIBS="-l$$1 $LIBS"
  if test ac_lib_found=yes -a -n "$ac_lib_dir"; then
      LIBS=-L$ac_lib_dir
 fi
])

dnl Same as AC version, but saves result of egrep in file conftest.out
dnl Because this macro is used by AC_PROG_GCC_TRADITIONAL, which must
dnl come early, it is not included in AC_BEFORE checks.
dnl LAC_EGREP_CPP(PATTERN, PROGRAM, ACTION-IF-FOUND [,
dnl              ACTION-IF-NOT-FOUND])
AC_DEFUN(LAC_EGREP_CPP_RESULT,
[AC_REQUIRE_CPP()dnl
cat > conftest.$ac_ext <<EOF
[#]line __oline__ "configure"
#include "confdefs.h"
[$2]
EOF
dnl eval is necessary to expand ac_cpp.
dnl Ultrix and Pyramid sh refuse to redirect output of eval, so use subshell.
if (eval "$ac_cpp conftest.$ac_ext") 2>&AC_FD_CC |
  egrep "$1" >conftest.out 2>&1; then
  ifelse([$3], , :, [$3
 rm -rf conftest*])
ifelse([$4], , , [else
  rm -rf conftest*
  $4
])dnl
fi
rm -f conftest*
])

dnl Use list of makefiles to include a common makefile_header
dnl into each makefile.
AC_DEFUN(LAC_MAKEFILES,
[define(lac_makefile_header,$1)dnl
 lac_makefiles=`echo '$2' | sed 's/\n/ /g'`
 lac_makefile_headers="`echo '$2' | sed 's|Makefile|makefile_header:lac_config_path/$1.in |g'`"
])


dnl Insert a standard header into a makefile.  The tricky
dnl bit is we have to run variable substitutions over the
dnl header before we insert it.  We also need to get
dnl reletive directories correct. 
AC_DEFUN(LAC_INSERT_MAKEFILE_HEADERS,[
lac_headers=`echo $lac_makefiles | sed 's/Makefile/lac_makefile_header/g'`
 while test -n "$lac_makefiles"; do
   set $lac_headers; lac_header=[$]1; shift; lac_headers=[$]*
   set $lac_makefiles; lac_makefile=[$]1; shift; lac_makefiles=[$]*
   sed "/@lac_makefile_header@/r $lac_header
        s/@lac_makefile_header@//g" $lac_makefile > conf.tmp
   mv conf.tmp $lac_makefile
 done
])

dnl LAC_PATH_PROG(VARIABLE, PROG-TO-CHECK-FOR [, VALUE-IF-NOT-FOUND [, PATH]])
dnl This is just like AC_PATH_PROG, except that a fully qualified path
dnl will simply be returned.
AC_DEFUN(LAC_PATH_PROG,
[
case $2 in
    /* )
        $1=$2
    ;;
    * )
        AC_PATH_PROG($1, $2, $3, $4)
    ;;
esac
])






dnl
dnl Set TARGET_ARCH_*
dnl
AC_DEFUN(LAC_TARGET_ARCH,
[
case "$target" in
  *sunos4* | *solaris1* )
    AC_DEFINE(TARGET_ARCH_SUNOS41)
  ;;
  *sunos5* | *solaris2* )
    AC_DEFINE(TARGET_ARCH_SOLARIS)
  ;;
  *-ibm-aix3* )
    AC_DEFINE(TARGET_ARCH_AIX)
    AC_DEFINE(TARGET_ARCH_AIX3)
  ;;
  *-ibm-aix4* )
    AC_DEFINE(TARGET_ARCH_AIX)
    AC_DEFINE(TARGET_ARCH_AIX4)
  ;;
  *-*-hpux* )
    AC_DEFINE(TARGET_ARCH_HPUX)
  ;;
  i860-intel-osf* )
    AC_DEFINE(TARGET_ARCH_PARAGON)
  ;;
  mips-sgi-irix5* )
    AC_DEFINE(TARGET_ARCH_SGI)
    AC_DEFINE(TARGET_ARCH_IRIX)
    AC_DEFINE(TARGET_ARCH_IRIX5)
  ;;
  mips-sgi-irix6* )
    AC_DEFINE(TARGET_ARCH_SGI)
    AC_DEFINE(TARGET_ARCH_IRIX)
    AC_DEFINE(TARGET_ARCH_IRIX6)
  ;;
  *dec* )
    AC_DEFINE(TARGET_ARCH_AXP)
    AC_DEFINE(TARGET_ARCH_OSF1)
    dnl --- more stuff to go here.. need to check with John
  ;;
  *c90* )
    AC_DEFINE(TARGET_ARCH_CRAYC90)
  ;;
  alpha-cray-unicosmk* )
    AC_DEFINE(TARGET_ARCH_CRAYT3E)
  ;;
  *freebsd* )
    AC_DEFINE(TARGET_ARCH_FREEBSD)
  ;;
  *nextstep* )
    AC_DEFINE(TARGET_ARCH_NEXTSTEP)
  ;;
  * )
	echo "platform not configured with TARGET_ARCH_*"
  ;;
esac
])

