
AC_DEFUN(LAC_NEXUS_ARGS,
[

AC_ARG_ENABLE(macros,
	[  --disable-macros              use function calls instead of macros],
	[lac_cv_macros="$enableval"],
	[lac_cv_macros=${lac_cv_macros='yes'}])

dnl AC_ARG_ENABLE(profile,
dnl	[  --enable-profile              compile in profiling features],
dnl	[lac_cv_profile="$enableval"],
dnl	[lac_cv_profile=${lac_cv_profile='no'}])
lac_cv_profile='no'

AC_ARG_ENABLE(critical-path-timer,
	[  --enable-critical-path-timer  time the critical path],
	[lac_cv_critical_path_timer="$enableval"],
	[lac_cv_critical_path_timer=${lac_cv_critical_path_timer='no'}])

AC_ARG_ENABLE(poll-range-checking,
	[  --enable-poll-range-checking  check poll ranges],
	[lac_cv_poll_range_checking="$enableval"],
	[lac_cv_poll_range_checking=${lac_cv_poll_range_checking='no'}])

AC_ARG_ENABLE(sigsegv-catching,
	[  --disable-sigsegv-catching     disable catching of SIGSEGV],
	[lac_cv_sigsegv_catching="$enableval"],
	[lac_cv_sigsegv_catching=${lac_cv_sigsegv_catching='yes'}])

AC_ARG_ENABLE(malloc-debugging,
	[  --enable-malloc-debugging      use debugging malloc],
	[lac_cv_malloc_debugging="$enableval"],
	[lac_cv_malloc_debugging=${lac_cv_malloc_debugging='no'}])

AC_ARG_ENABLE(malloc-padding,
	[  --enable-malloc-padding        pad mallocs for debugging],
	[lac_cv_malloc_padding="$enableval"],
	[lac_cv_malloc_padding=${lac_cv_malloc_padding='no'}])

AC_ARG_ENABLE(resource,
	[  --enable-resource		nexus resource abstractions],
	[lac_cv_resource="$enableval"],
	[lac_cv_resource=${lac_qos='no'}])

AC_ARG_ENABLE(qos,
	[  --enable-qos			quality of service overseer],
	[lac_cv_qos="$enableval"],
	[lac_cv_qos=${lac_qos='no'}])


dnl -----------------------------------------------------------------
dnl -- Protocols
dnl --- tcp, udp, inx, shm, mpinx, mpl
dnl -- Startups
dnl --- fork, rsh, ss, ibmds, soldl
dnl -- Database
dnl --- file
dnl -----------------------------------------------------------------

proto_tcp="notset"
proto_udp="notset"
proto_inx="notset"
proto_shm="notset"
startup_inx="notset"
startup_fork="notset"
startup_rsh="notset"
startup_ss="notset"
startup_ibmds="notset"

AC_ARG_WITH(tcp,
	[  --with-tcp                    include the TCP/IP protocol],
	[proto_tcp="$withval"],
	[proto_tcp="notset"])
AC_ARG_WITH(udp,
	[  --with-udp                    include the UDP/IP protocol],
	[proto_udp="$withval"],
	[proto_udp="notset"])
AC_ARG_WITH(inx,
	[  --with-inx                    include the Intel NX protocols],
	[proto_inx="$withval"; startup_inx="$withval"],
	[proto_inx="notset"; startup_inx="notset"])
AC_ARG_WITH(shm,
	[  --with-shm                    include the System V shared memory protocols],
	[proto_shm="$withval"],
	[proto_shm="notset"])
AC_ARG_WITH(fork,
	[  --with-fork                   include the fork startup module],
	[startup_fork="$withval"],
	[startup_fork="notset"])
AC_ARG_WITH(rsh,
	[  --with-rsh                    include the rsh startup module],
	[startup_rsh="$withval"],
	[startup_rsh="notset"])
AC_ARG_WITH(ss,
	[  --with-ss                     include the secure server startup module],
	[startup_ss="$withval"],
	[startup_ss="notset"])
AC_ARG_WITH(ibmds,
	[  --with-ibmds                  include the IBM AIX dynaload startup module],
	[startup_ibmds="$withval"],
	[startup_ibmds="notset"])
])


AC_DEFUN(LAC_NEXUS,
[

LAC_NEXUS_ARGS

if test "$lac_cv_macros" = "yes" ; then
    AC_DEFINE(BUILD_USING_MACROS)
fi

if test "$lac_cv_profile" = "yes" ; then
    AC_DEFINE(BUILD_PROFILE)
fi

if test "$lac_cv_critical_path_timer" = "yes" ; then
    AC_DEFINE(NEXUS_CRITICAL_PATH_TIMER)
fi

if test "$lac_cv_poll_range_checking" = "yes" ; then
    AC_DEFINE(NEXUS_POLL_RANGE_CHECK)
fi

if test "$lac_cv_sigsegv_catching" = "no" ; then
    AC_DEFINE(NEXUS_DONT_CATCH_SIGSEGV)
fi

if test "$lac_cv_malloc_debugging" = "yes" ; then
    AC_DEFINE(PORTS0_MALLOC_DEBUG)
fi
if test "$lac_cv_malloc_padding" = "yes" ; then
    AC_DEFINE(PORTS0_MALLOC_PAD)
fi

if test "$lac_cv_resource" = "yes" ; then
    AC_DEFINE(BUILD_RESOURCE)
fi

if test "$lac_cv_qos" = "yes" ; then
    AC_DEFINE(HAVE_QOS)
    shared_overseer_objs="$shared_overseer_objs overseer_qos.\$(OFILE) qoslib.\$(OFILE)"

fi

if test "$lac_cv_debug" = "yes"; then
    AC_DEFINE(BUILD_DEBUG)
fi

if test "$lac_threads_type" = "no" ; then
    AC_DEFINE(BUILD_LITE)
fi


dnl
dnl Set the thread module
dnl
case "$lac_threads_type" in
    pthreads )
        lac_th_module="p0_th_p"
    ;;
    cthreads )
        lac_th_module="p0_th_c"
    ;;
    solaristhreads )
        lac_th_module="p0_th_sol"
    ;;
    quickthreads )
        lac_th_module="p0_th_qt"
    ;;
    no )
        lac_th_module="p0_th_null"
    ;;
    * )
        AC_MSG_ERROR([Invalid threads type: $lac_threads_type])
    ;;
esac
PORTS0_THREAD_MODULE="$lac_th_module"
AC_DEFINE_UNQUOTED(PORTS0_THREAD_MODULE, $lac_th_module)
AC_SUBST(PORTS0_THREAD_MODULE)


dnl
dnl Figure out all of the startup and protocol modules
dnl
shared_startup_objs=""
static_startup_objs=""
shared_proto_objs=""
static_proto_objs=""
shared_transform_objs=""
static_transform_objs=""

LAC_SYS_IBMDS
LAC_SYS_MPL
LAC_SYS_INX
LAC_SYS_TCP_UDP
LAC_SYS_SHM
LAC_SYS_FORK
LAC_SYS_RSH
LAC_SYS_SS

#shared_transform_objs="transform_gzip.\$(OFILE)"

NEXUS_SHARED_PROTO_OBJS="$shared_proto_objs"
NEXUS_STATIC_PROTO_OBJS="$static_proto_objs"
NEXUS_SHARED_STARTUP_OBJS="$shared_startup_objs"
NEXUS_STATIC_STARTUP_OBJS="$static_startup_objs"
NEXUS_SHARED_TRANSFORM_OBJS="$shared_transform_objs"
NEXUS_STATIC_TRANSFORM_OBJS="$static_transform_objs"
NEXUS_SHARED_OVERSEER_OBJS="$shared_overseer_objs"
NEXUS_STATIC_OVERSEER_OBJS="$static_overseer_objs"
AC_SUBST(NEXUS_SHARED_PROTO_OBJS)
AC_SUBST(NEXUS_STATIC_PROTO_OBJS)
AC_SUBST(NEXUS_SHARED_STARTUP_OBJS)
AC_SUBST(NEXUS_STATIC_STARTUP_OBJS)
AC_SUBST(NEXUS_SHARED_TRANSFORM_OBJS)
AC_SUBST(NEXUS_STATIC_TRANSFORM_OBJS)
AC_SUBST(NEXUS_SHARED_OVERSEER_OBJS)
AC_SUBST(NEXUS_STATIC_OVERSEER_OBJS)

NEXUS_TH_MODULE="th_p0"
NEXUS_MD_MODULE="md_unix"
NEXUS_NL_MODULE="nl_file"
AC_SUBST(NEXUS_TH_MODULE)
AC_SUBST(NEXUS_MD_MODULE)
AC_SUBST(NEXUS_NL_MODULE)

LAC_DATA_CONVERSION

dnl
dnl Miscellaneous other checks
dnl
LAC_FUNC_FILE_LOCK
LAC_NEXUS_ALIGN
LAC_CHECK_POSIX_SIGNALS
LAC_CHECK_SELECT_PROTOTYPE
AC_TYPE_SIGNAL  dnl #define RETSIGTYPE to int or void
LAC_SYS_X_ACC
LAC_SYS_X_OK

AC_CHECK_HEADERS(libc.h)
AC_CHECK_HEADERS(malloc.h)
AC_CHECK_HEADERS(unistd.h)
AC_HEADER_STDC
AC_HEADER_TIME
AC_CHECK_HEADERS(sys/types.h sys/param.h sys/access.h sys/errno.h sys/file.h)
AC_CHECK_HEADERS(sys/sysmp.h sys/lwp.h sys/stat.h sys/file.h sys/uio.h)
AC_CHECK_HEADERS(sys/time.h sys/signal.h sys/select.h sys/ioctl.h)
AC_CHECK_HEADERS(dce/cma.h)
AC_CHECK_HEADERS(dce/cma_ux.h)
AC_CHECK_HEADERS(sys/cnx_pattr.h)
AC_CHECK_HEADERS(netdb.h)
AC_HEADER_SYS_WAIT

AC_CHECK_FUNCS(waitpid)
AC_CHECK_FUNCS(strtoul)
AC_CHECK_FUNCS(wait3)
dnl AC_FUNC_WAIT3
AC_CHECK_FUNCS(sighold)
AC_CHECK_FUNCS(sigblock)
AC_CHECK_FUNCS(getwd)
AC_CHECK_FUNCS(getcwd)
AC_CHECK_FUNCS(memmove)

AC_DEFINE_UNQUOTED(NEXUS_LISTENER,"$prefix/bin/listener")


dnl
dnl Set TARGET_ARCH_*
dnl
dnl These parameters are now set in acmisc.m4

dnl LAC_TARGET_ARCH

])

dnl ---------------------------------------------------------------------
dnl LAC_SYS_TCP_UDP()
dnl
dnl Determine if this system supports TCP/IP communication.
dnl Add pr_tcp to the build if it does.
dnl ---------------------------------------------------------------------
AC_DEFUN(LAC_SYS_TCP_UDP,
[
AC_CHECK_FUNCS(socket connect listen bind accept select getsockname,
                [have_tcp_funcs="yes"],
                [have_tcp_funcs="no"])
AC_CHECK_FUNCS(sysconf, [ ] , [ ] )
if test "$have_tcp_funcs" = "yes" ; then
    AC_DEFINE(HAVE_TCP)
    if test "$proto_tcp" != "no" ; then
        proto_tcp="yes"
        AC_DEFINE(HAVE_TCP_PROTO)
        shared_proto_objs="$shared_proto_objs pr_tcp.\$(OFILE)"
    fi
    if test "$proto_udp" != "no" ; then
        proto_udp="yes"
        AC_DEFINE(HAVE_UDP_PROTO)
        shared_proto_objs="$shared_proto_objs pr_udp.\$(OFILE)"
    else
        proto_udp="no"
    fi
fi
])

dnl ---------------------------------------------------------------------
dnl LAC_SYS_SHM()
dnl
dnl Determine if this system supports System V
dnl shared memory communication.
dnl Add pr_shm to the build if it does.
dnl ---------------------------------------------------------------------
AC_DEFUN(LAC_SYS_SHM,
[
if test "$proto_shm" = "yes" ; then
    AC_CHECK_HEADERS(sys/ipc.h sys/sem.h sys/shm.h, 
                      [], [proto_shm="no"])
    if test "$proto_shm" != "no"; then
        AC_DEFINE(HAVE_SHM_PROTO)
        shared_proto_objs="$shared_proto_objs pr_shm.\$(OFILE)"
        AC_TRY_COMPILE(
            #include <sys/ipc.h>
            #include <sys/sem.h> 
            ,
            union semun foo;
            ,
            [AC_DEFINE(HAVE_UNION_SEMUN)] )
    fi
fi
])

dnl ---------------------------------------------------------------------
dnl LAC_SYS_FORK()
dnl
dnl Determine if this system supports
dnl creating processes with the fork system call.
dnl Add st_fork to the build if it does.
dnl ---------------------------------------------------------------------
AC_DEFUN(LAC_SYS_FORK,
[
if test "$startup_fork" != "no" ; then
    AC_CHECK_FUNCS(fork1 fork,
                    [startup_fork="yes"],
                    [startup_fork="no"])
    if test "$startup_fork" = "yes" ; then
        AC_DEFINE(HAVE_FORK_STARTUP)
        shared_startup_objs="$shared_startup_objs st_fork.\$(OFILE)"
    fi
fi
])

dnl ---------------------------------------------------------------------
dnl LAC_SYS_RSH()
dnl
dnl Determine if this system supports
dnl creating processes with the rsh program.
dnl Add st_rsh to the build if it does.
dnl ---------------------------------------------------------------------
AC_DEFUN(LAC_SYS_RSH,
[
if test "$startup_rsh" != "no" ; then
    if test "$startup_rsh" = "yes" -o "$startup_rsh" = "notset" ; then
        dnl Look for remsh or rsh
        AC_CHECK_PROGS(rsh_prog, remsh rsh,no)
        if test "$rsh_prog" != "no"; then
            AC_PATH_PROG(RSH,"$rsh_prog")
            startup_rsh="yes"
        else
            startup_rsh="no"
        fi
    else
        dnl User has specified a --with-rsh= value.
        LAC_PATH_PROG(RSH,"$startup_rsh")
        if test -n "$RSH" ; then
            startup_rsh="yes"
        else
            startup_rsh="no"
        fi
    fi
    if test "$startup_rsh" = "yes" ; then
        AC_DEFINE_UNQUOTED(NEXUS_RSH,"$RSH")
        AC_DEFINE(HAVE_RSH_STARTUP)
        shared_startup_objs="$shared_startup_objs st_rsh.\$(OFILE)"
    fi
fi
])

dnl ---------------------------------------------------------------------
dnl LAC_SYS_SS()
dnl
dnl Determine if this system supports
dnl creating processes with the secure server program.
dnl Add st_ss to the build if it does.
dnl ---------------------------------------------------------------------
AC_DEFUN(LAC_SYS_SS,
[
if test "$startup_ss" = "yes"; then
    echo "secure server startup module has been disabled"
fi
startup_ss="no"
if test "$startup_ss" != "no"; then
    startup_ss="yes"
    AC_DEFINE(HAVE_SS_STARTUP)
    shared_startup_objs="$shared_startup_objs st_ss.\$(OFILE)"
fi
])

dnl ---------------------------------------------------------------------
dnl LAC_SYS_MPL()
dnl
dnl Determine if this system supports
dnl the IBM MPL protocol (used in SP/2's).
dnl Add pr_mpl and st_mpl to the build if it does.
dnl ---------------------------------------------------------------------

AC_DEFUN(LAC_SYS_MPL,
[
dnl LAC_SYS_IBMDS and LAC_COMPILERS_* must be called before this
if test "$lac_mpl" = "yes" ; then
    AC_DEFINE(NEXUS_DONT_CATCH_SIGSEGV)
    AC_DEFINE(NEXUS_LISTENER_EXEC_HACK)
    AC_DEFINE(HAVE_MPL_PROTO)
    AC_DEFINE(HAVE_MPL_STARTUP)
    if test "$startup_ibmds" = "yes"; then
        static_proto_objs="$static_proto_objs pr_mpl.\$(OFILE)"
        static_startup_objs="$static_startup_objs st_mpl.\$(OFILE)"
    else
        shared_proto_objs="$shared_proto_objs pr_mpl.\$(OFILE)"
        shared_startup_objs="$shared_startup_objs st_mpl.\$(OFILE)"
    fi
fi
])

dnl ---------------------------------------------------------------------
dnl LAC_SYS_INX()
dnl
dnl Determine if this system supports
dnl the Intel NX protocol (used in Paragons).
dnl Add pr_inx and st_inx to the build if it does.
dnl ---------------------------------------------------------------------
AC_DEFUN(LAC_SYS_INX,
[
if test "$proto_inx" != "no" ; then
    AC_MSG_CHECKING(for NX)
    AC_CACHE_VAL(lac_cv_sys_inx,
            [
            case "$target" in
                i860-intel-osf* )
                    lac_cv_sys_inx="yes"
                ;;
                * )
                    lac_cv_sys_inx="no"
                ;;
            esac
            ])
    AC_MSG_RESULT($lac_cv_sys_inx)
    proto_inx="$lac_cv_sys_inx"
    startup_inx="$lac_cv_sys_inx"
    if test "$lac_cv_sys_inx" = "yes" ; then
        AC_DEFINE(HAVE_INX_PROTO)
        AC_DEFINE(HAVE_INX_STARTUP)
        shared_proto_objs="$shared_proto_objs pr_inx.\$(OFILE)"
        shared_startup_objs="$shared_startup_objs st_inx.\$(OFILE)"
    fi
fi
])

dnl ---------------------------------------------------------------------
dnl LAC_SYS_IBMDS()
dnl
dnl Determine if this system supports
dnl data segment cloning under AIX 3.2.5.
dnl Add st_ibmds to the full build if it does.
dnl ---------------------------------------------------------------------
AC_DEFUN(LAC_SYS_IBMDS,
[
if test "$startup_ibmds" != "no" -a "$startup_ibmds" != "notset"; then
    echo "IBM AIX dynaload (IBMDS) startup module has been disabled"
fi
startup_ibmds="no"
if test "$startup_ibmds" != "no" ; then
    AC_MSG_CHECKING(for IBMDS)
    AC_CACHE_VAL(lac_cv_sys_ibmds,
            [
            case "$target" in
                *aix3* )
                    lac_cv_sys_ibmds="yes"
                ;;
                * )
                    lac_cv_sys_ibmds="no"
                ;;
            esac
            ])
    AC_MSG_RESULT($lac_cv_sys_ibmds)
    startup_ibmds="$lac_cv_sys_ibmds"
    if test "$lac_cv_sys_ibmds" = "yes"; then
        AC_DEFINE(HAVE_IBMDS_STARTUP)
        shared_startup_objs="$shared_startup_objs st_ibmds.\$(OFILE)"
        static_startup_objs="$static_startup_objs st_ibmds2.\$(OFILE) st_ibmds4.\$(OFILE)"
    fi
fi
])

AC_DEFUN(LAC_SYS_X_ACC,
[AC_MSG_CHECKING(for X_ACC)
    AC_CACHE_VAL(lac_cv_sys_x_acc, [dnl
	AC_EGREP_CPP(zowie,
[#include <unistd.h>
#ifdef X_ACC
    zowie
#endif
], lac_cv_sys_x_acc="yes", lac_cv_sys_x_acc="no") dnl --End AC_EGREP_CPP()
]) dnl --End AC_CACHE_VAL()
    if test "$lac_cv_sys_x_acc" = "yes"; then
	AC_DEFINE(HAVE_X_ACC)
    fi
    AC_MSG_RESULT($lac_cv_sys_x_acc)
]) dnl --End AC_SYS_ACC()

AC_DEFUN(LAC_SYS_X_OK,
[AC_MSG_CHECKING(for X_OK)
AC_CACHE_VAL(lac_cv_sys_x_ok, [dnl
AC_EGREP_CPP(zowie,
[#include <unistd.h>
#ifdef X_OK
  zowie
#endif
], lac_cv_sys_x_ok="yes", lac_cv_sys_x_ok="no")
dnl
dnl --Of course gcc on the NeXT is braindead, so it doesn't include the
dnl --right header files, so we have to check special for it!!!
dnl
lac_cv_sys_x_ok="yes"
])
if test "$lac_cv_sys_x_ok" = "yes"; then
  AC_DEFINE(HAVE_X_OK)
fi
AC_MSG_RESULT($lac_cv_sys_x_ok)
])


dnl
dnl  --HP seems to think the prototype for select() should be:
dnl  --
dnl  --   int select(int, int *, int *, int *, struct timeval *);
dnl  --
dnl  --by putting the "correct" prototype in the source, it causes an
dnl  --error on any machine with a different prototype.  The strange
dnl  --thing is that the HP man pages show two examples; one using an
dnl  --int * and the other an fd_set *.  Best be safe and use what the
dnl  --compiler is expecting.
dnl  --
dnl  --IBM is even more strange.  They have the prototype for select()
dnl  --in the info pages as:
dnl  --
dnl  --   int select(int,
dnl  --		     struct sellist *,
dnl  --		     struct sellist *,
dnl  --		     struct sellist *,
dnl  --		     struct timeval *);
dnl  --
dnl  --but the header files have it as (and xlc takes):
dnl  --
dnl  --   int select(unsigned long, void *, void *, void *, struct timeval *);
dnl  --
dnl  --This is very strange indeed.
dnl
AC_DEFUN(LAC_CHECK_SELECT_PROTOTYPE,
[AC_MSG_CHECKING(what type to use for select())
    AC_CACHE_VAL(lac_cv_fd_set_cast, [dnl
	AC_TRY_COMPILE(
#include <stddef.h>
#include <sys/types.h>
#include <sys/time.h>

#ifdef _CMA_REENTRANT_CLIB_
#include <dce/cma.h>
#endif

int select(int, fd_set *, fd_set *, fd_set *, struct timeval *);
,
    fd_set read_values;
    int rc;

    rc = select(0, &read_values, NULL, NULL, NULL);
,
    lac_cv_fd_set_cast="(fd_set *)"
,
    [case "$target" in
	*-ibm-aix* )
	    lac_cv_fd_set_cast="(void *)"
	;;
	*hpux* )
	    lac_cv_fd_set_cast="(int *)"
	;;
	* )
	    lac_cv_fd_set_cast="notset"
	    echo "Please notify nexus@mcs.anl.gov that autoconf could"
	    echo "not detect the type for your select() system call."
	    echo "Please include your system type (OS, manufacturer, chip)"
	    echo "as well as a copy of /usr/include/sys/types.h and any"
	    echo "other file that is used for the select() call."
	    AC_MSG_ERROR(cannot compile without the correct information)
	;;
    esac]
	) dnl --End of AC_TRY_COMPILE()
    ]) dnl --End of AC_CACHE_VAL()

    AC_DEFINE_UNQUOTED(NEXUS_FD_SET_CAST,$lac_cv_fd_set_cast)
    AC_SUBST(NEXUS_FD_SET_CAST)
    AC_MSG_RESULT($lac_cv_fd_set_cast)
]) dnl --End of AC_CHECK_SELECT_PROTOTYPE()

AC_DEFUN(LAC_CHECK_POSIX_SIGNALS,
[found_ps_funcs="yes"
    AC_MSG_CHECKING(for posix signals)
    AC_CACHE_VAL(lac_cv_check_posix_signals, [dnl
        AC_MSG_RESULT(no, checking manually)
        AC_CHECK_FUNCS(sigaction sigemptyset sigaddset sigprocmask waitpid,
	# do nothing if it finds the functions
	,
	    found_ps_funcs="no"
	    break) dnl --End of AC_CHECK_FUNCS()
    ]) dnl --End of AC_CACHE_VAL

    if test "$found_ps_funcs" = "yes" ; then
        lac_cv_check_posix_signals="yes"
        AC_DEFINE(HAVE_POSIX_SIGNALS)
    fi
    AC_MSG_RESULT($lac_cv_check_posix_signals)
]) dnl --End of AC_CHECK_POSIX_SIGNALS()

dnl
dnl  --This is almost a direct copy of the file that John Garnett wrote
dnl  --for the CC++ installation.
dnl
AC_DEFUN(LAC_FUNC_FILE_LOCK,
[AC_MSG_CHECKING(for preferred file locking method in cache)
  AC_CACHE_VAL(lac_cv_func_file_lock, [dnl
    dnl
    dnl --Don't use AC_CHECK_FUNCS() because we don't want to have
    dnl --these functions recognized unless the program says they work.
    dnl
    AC_MSG_RESULT(no)
    AC_CHECK_FUNC(flock, AC_DEFINE(USE_FLOCK))
    AC_CHECK_FUNC(fcntl, AC_DEFINE(USE_FCNTL))
    AC_CHECK_FUNC(lockf, AC_DEFINE(USE_LOCKF))
    dnl  AC_CHECK_FUNCS(flock fcntl lockf)
    AC_MSG_CHECKING(which lock method to use)
    AC_TRY_RUN(
[#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/file.h>

char filename[32];
int fd;
FILE *output;

void cleanup(ec)
int ec;
{
        close(fd);
        unlink(filename);
        fclose(output);
        exit(ec);
}

void do_lseek(which)
char* which;
{
        while (lseek(fd, 0, SEEK_SET) == -1) {
                if (errno == EINTR) {
                        fprintf(stderr, "lseek %s failed with errno = %d\n", which, errno);
                        break;
                }
        }
}

#ifdef USE_FLOCK
int do_flock()
{
        while (flock(fd, LOCK_EX) == -1) {
                if (errno != EINTR) {
                        fprintf(stderr, "flock failed with errno = %d\n", errno);
                        return 0;
                }
        }
        return 1;
}


int do_unflock()
{
        while (flock(fd, LOCK_UN) == -1) {
                if (errno != EINTR) {
                        fprintf(stderr, "unflock failed with errno = %d\n", errno);
                        return 0;
                }
        }
        return 1;
}
#endif /* USE_FLOCK */

#ifdef USE_LOCKF
int do_lockf()
{
        do_lseek("lockf");
        while (lockf(fd, F_LOCK, 1L) == -1) {
                if (errno != EINTR) {
                        fprintf(stderr, "lockf failed with errno = %d\n", errno);
                        return 0;
                }
        }
        return 1;
}

int do_unlockf()
{
        do_lseek("unlockf");
        while (lockf(fd, F_ULOCK, 1L) == -1) {
                if (errno != EINTR) {
                        fprintf(stderr, "unlockf failed with errno = %d\n", errno);
                        return 0;
                }
        }
        return 1;
}
#endif /* USE_LOCKF */

#ifdef USE_FCNTL
int do_fcntl()
{
        struct flock theLock;

        theLock.l_type = F_WRLCK;
        theLock.l_whence = 0;
        theLock.l_start = 0;
        theLock.l_len = 0;

        while (fcntl(fd, F_SETLKW, (int) &theLock) == -1) {
                if (errno != EINTR) {
                        fprintf(stderr, "fcntl failed with errno = %d\n", errno);
                        return 0;
                }
        }
        return 1;
}

int do_unfcntl()
{
        struct flock theLock;

        theLock.l_type = F_UNLCK;
        theLock.l_whence = 0;
        theLock.l_start = 0;
        theLock.l_len = 0;

        while (fcntl(fd, F_SETLKW, (int) &theLock) == -1) {
                if (errno != EINTR) {
                        fprintf(stderr, "fcntl failed with errno = %d\n", errno);
                        return 0;
                }
        }
        return 1;
}
#endif /* USE_FCNTL */

int main(int argc, char **argv)
{
        int suffix;
        int success = 0;

        if (argc == 2) {
                suffix = atoi(argv[1]);
        } else {
                suffix = getpid();
        }
        sprintf(filename, "conftest.res", suffix);
        output = fopen(filename, "w");
        sprintf(filename, "/tmp/lock.%d", suffix);
        fd = open(filename, O_RDWR|O_CREAT|O_EXCL, 0600);
        if (fd == -1) {
                fprintf(stderr, "Open of %s failed.\n", filename);
                exit(1);
        }
#ifdef USE_LOCKF
        if (do_lockf()) {
                if (do_unlockf()) {
                    success = 1;
                    fprintf(output, "LOCKF");
                }
        }
#endif /* USE_LOCKF */
#ifdef USE_FCNTL
        if (!success && do_fcntl()) {
                if (do_unfcntl()) {
                    success = 2;
                    fprintf(output, "FCNTL");
                }
        }
#endif /* USE_FCNTL */
#ifdef USE_FLOCK
        if (!success && do_flock()) {
                if (do_unflock()) {
                    success = 3;
                    fprintf(output, "FLOCK");
                }
        }
#endif /* USE_FLOCK */
        if (!success) {
            fprintf(output, "none");
        }
        cleanup(0);
        return (!success);
}
    ],[
	lac_cv_func_file_lock=`cat conftest.res`
    ],[
	lac_cv_func_file_lock=none
	AC_MSG_RESULT(none)
    ],[
	lac_cv_func_file_lock=LOCKF
    ]) dnl --End of AC_TRY_RUN()

    if test "$lac_cv_func_file_lock" = "none" ; then
	AC_MSG_ERROR(You must have at least one locking method: flock() fcntl() lockf())
    fi
  ]) dnl --End of AC_CACHE_VAL()
  AC_MSG_RESULT($lac_cv_func_file_lock)
  AC_DEFINE_UNQUOTED(HAVE_$lac_cv_func_file_lock)
]) dnl --End of AC_FUNC_FILE_LOCK()

AC_DEFUN(LAC_FUNC_MALLOCERR,
[ AC_MSG_CHECKING(for malloc error descriptors (NeXT))
  AC_CACHE_VAL(lac_cv_func_mallocerr, [dnl
    result=0
    AC_CHECK_FUNC(malloc_debug, result=`expr $result + 1`)
    AC_CHECK_FUNC(malloc_error, result=`expr $result + 1`)
    if test "$result" = "2"; then
      lac_cv_func_mallocerr="yes"
    else
      lac_cv_func_mallocerr="no"
    fi
  ])
  AC_MSG_RESULT($lac_cv_func_mallocerr)
  if test "$lac_cv_func_mallocerr" = "yes"; then
    AC_DEFINE(HAVE_MALLOCERR)
  fi
])

AC_DEFUN(LAC_SYS_SIGNAL_WAITING,
[ AC_MSG_CHECKING(for SIGWAITING)
  AC_CACHE_VAL(lac_cv_sys_signal_waiting, [dnl
  AC_EGREP_CPP(zowie,
[#include <signal.h>
#ifdef SIGWAITING
 zowie
#endif
], lac_cv_sys_signal_waiting="yes",
lac_cv_sys_signal_waiting="no")
])
  AC_MSG_RESULT($lac_cv_sys_signal_waiting)
  if test "$lac_cv_sys_signal_waiting" = "yes"; then
    AC_DEFINE(HAVE_SYS_SIGWAITING)
  fi
])

AC_DEFUN(LAC_STRUCT_TBTABLE,
[ AC_MSG_CHECKING(for struct tbtable tracebacks)
  AC_CACHE_VAL(lac_cv_struct_tbtable, [dnl
    lac_cv_struct_tbtable="no"
    AC_CHECK_HEADER(sys/debug.h,
      AC_EGREP_HEADER(tbtable, sys/debug.h, lac_cv_struct_tbtable="yes"))
  ])
  AC_MSG_RESULT($lac_cv_struct_tbtable)
  if test "$lac_cv_struct_tbtable" = "yes"; then
    AC_DEFINE(HAVE_STRUCT_TBTABLE)
  fi
])

dnl
dnl  --This is code that was adopted from the PAC_CHECK_COMPILER()
dnl  --macro in the mpich distriubtion
dnl  --
dnl  --I would like to use this once I check it works correctly as it
dnl  --would dramically reduce the amount of C code in this file.  And
dnl  --it would make it very easy to add/delete a test.
dnl  --

dnl Not this one is never called at present time
dnl SMF Fri Aug  1, 1997
AC_DEFUN(LAC_CHECK_PROGS,
[   if test "x$check_progs_dir" = "x" ; then
        $check_progs_dir="../config/test/nexus"
    fi
    for file in $check_progs_dir/*.c ; do
	base=`basename $file .c`
	AC_MSG_CHECKING(`cat $check_progs_dir/$base.check`)
	cp $file conftest.c
	rm -f conftest.cout conftest.rout
	if eval $CC $CFLAGS -o conftest conftest.c $LIBS > conftest.cout 2>&1 ; then
	    if test -s conftest ; then
		./conftest 2>&1 1> conftest.rout
		if test -s conftest.result ; then
		    AC_MSG_RESULT(`cat conftest.res`)
		    result_unknown="no"
		else
		    AC_MSG_RESULT(unknown)
		    result_unknown="yes"
		fi
		if test $? = 0 ; then
		    # it worked
		    if test "$result_unknow" = "yes" ; then
			echo "Program appeared to work, however"
		    fi
		else
		    # it didn't
		    cat $base.msg
		    if test -s conftest.cout ; then
			echo "Output from compile step was:"
			cat conftest.cout
		    fi
		    if test -s conftest.rout ; then
			echo "Output from run step was:
			cat conftest.rout
		    fi
		fi
	    fi
	else
	    echo "Program failed to compile."
	    cat $base.msg
	    if test -s conftest.cout ; then
		echo "Output from compile step was:"
		cat conftest.cout
	    fi
	fi
    done
]) dnl --End of AC_CHECK_PROGS()


AC_DEFUN(LAC_NEXUS_ALIGN,
[
case "$target" in
    *)
        AC_DEFINE(NEXUS_ALIGN,8)
    ;;	
esac
])


AC_DEFUN(LAC_DATATYPE_SIZES,[
if test "$cross_compiling" = "no"; then
    AC_C_BIGENDIAN
    AC_CHECK_SIZEOF(short)
    AC_CHECK_SIZEOF(long)
else
    case "$target" in
      i860-intel-osf* )
        ac_cv_c_bigendian=no
        ac_cv_sizeof_short=2
        ac_cv_sizeof_long=4
        ;;
      * )
        ac_cv_c_bigendian=yes
        ac_cv_sizeof_short=2
        ac_cv_sizeof_long=4
        ;;
    esac
    echo "guessing whether byte ordering is big endian... $ac_cv_c_bigendian"
    echo "guessing size of short... $ac_cv_sizeof_short"
    echo "guessing size of long... $ac_cv_sizeof_long"
fi
])


AC_DEFUN(LAC_DATA_CONVERSION,
[
LAC_DATATYPE_SIZES

NEXUS_DC_FORMAT_32BIT_LE=0
NEXUS_DC_FORMAT_32BIT_BE=1
NEXUS_DC_FORMAT_64BIT_LE=2
NEXUS_DC_FORMAT_64BIT_BE=3
NEXUS_DC_FORMAT_CRAYC90=4
NEXUS_DC_FORMAT_JAVA=5
NEXUS_DC_FORMAT_CRAYT3E=6

case "$ac_cv_sizeof_long-$ac_cv_c_bigendian" in
    "4-no" )
	NEXUS_DC_FORMAT_LOCAL=$NEXUS_DC_FORMAT_32BIT_LE
	NEXUS_DC_FILE=nexus_dc_32le
	;;
    "4-yes" )
	NEXUS_DC_FORMAT_LOCAL=$NEXUS_DC_FORMAT_32BIT_BE
	NEXUS_DC_FILE=nexus_dc_32be
	;;
    "8-no" )
	NEXUS_DC_FORMAT_LOCAL=$NEXUS_DC_FORMAT_64BIT_LE
	NEXUS_DC_FILE=nexus_dc_64le
	;;
    "8-yes" )
	if test "$ac_cv_sizeof_short" = "2" ; then
	    NEXUS_DC_FORMAT_LOCAL=$NEXUS_DC_FORMAT_64BIT_BE
	    NEXUS_DC_FILE=nexus_dc_64be
	elif test "$ac_cv_sizeof_short" = "4" ; then
	    NEXUS_DC_FORMAT_LOCAL=$NEXUS_DC_FORMAT_CRAYT3E
	    NEXUS_DC_FILE=nexus_dc_crayt3e
	else
	    NEXUS_DC_FORMAT_LOCAL=$NEXUS_DC_FORMAT_CRAYC90
	    NEXUS_DC_FILE=nexus_dc_crayc90
	fi
	;;
esac

AC_DEFINE_UNQUOTED(NEXUS_DC_FORMAT_LOCAL,$NEXUS_DC_FORMAT_LOCAL)
AC_SUBST(NEXUS_DC_FILE)
])

