/******************************************************************************
event.c

Description:

  UNIX file descriptor toolset

CVS Information:

  $Source: /nfs/globus1/src/master/nexus_4.1/event.c,v $
  $Date: 1997/09/19 07:47:00 $
  $Revision: 1.54 $
  $State: Exp $
  $Author: tuecke $
******************************************************************************/

static char *rcsid = "$Header: /nfs/globus1/src/master/nexus_4.1/event.c,v 1.54 1997/09/19 07:47:00 tuecke Exp $";


/******************************************************************************
			     Include header files
******************************************************************************/

#include "internal.h"

#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#if defined(HAVE_DCE_CMA_UX_H) && !defined(BUILD_LITE) && defined(HAVE_THREAD_SAFE_SELECT) && defined(TARGET_ARCH_AIX)
#include <dce/cma_ux.h>
#endif


/******************************************************************************
		       Define module specific constants
******************************************************************************/


/******************************************************************************
			       Type definitions
******************************************************************************/


/******************************************************************************
		       Define module specific variables
******************************************************************************/


/******************************************************************************
			  Module specific prototypes
******************************************************************************/


#ifdef BUILD_LITE
#   ifndef HAVE_THREAD_SAFE_SELECT
#       define HAVE_THREAD_SAFE_SELECT
#   endif
#endif /* BUILD_LITE */

#undef USE_BLOCKING_SELECT
#undef USE_BLOCKING_READ
#undef USE_BLOCKING_WRITE
#undef USE_BLOCKING_WRITEV
#undef USE_THREAD_SAFE_FD_CALLS

#ifndef NEXUS_FD_SET_CAST
#ifdef NEXUS_ARCH_HPUX
#define NEXUS_FD_SET_CAST (int *)
#else
#define NEXUS_FD_SET_CAST (fd_set *)
#endif
#endif

/*
 * On an old version of HP/UX (7.something, I believe), if
 * you did a non-blocking write of a buffer that was greater
 * than 4096 bytes, the write would return a "Message too long"
 * error, instead of writing what it could and then returning.
 *
 * If you ever hit a situation like this again, set
 * NEXUS_MAX_WRITE_PACKET to the maximum number of bytes that you
 * want passed to write() at a time.  For example:
 *	#define NEXUS_MAX_WRITE_PACKET 4096
 *
 * If NEXUS_MAX_WRITE_PACKET==0, then it will always just write the
 * full buffer, no matter how long it is.
 *
 * This can also be useful if you want to break write operations
 * into blocks for, say, performance reasons.
 */
#ifndef NEXUS_MAX_WRITE_PACKET
#define NEXUS_MAX_WRITE_PACKET 0
#endif

/*
 * *_info_t
 *
 * Callback information structures for select, listener, read, write, writev
 */
typedef void (* nexusl_fd_cbf_t)(void * arg,
				 int fd);

typedef struct _select_info_t
{
    int				fd;
    nexusl_fd_cbf_t		read_callback;
    void *			read_arg;
    nexusl_fd_cbf_t		write_callback;
    void *			write_arg;
    nexusl_fd_cbf_t		except_callback;
    void *			except_arg;
} select_info_t;

typedef struct _listener_info_t
{
    unsigned short		port;
    int				fd;
    struct _listener_info_t *	next;
    struct _listener_info_t *	prev;
    void *			arg;
    nexusl_fd_cbf_t		callback;
} listener_info_t;

typedef struct _read_info_t
{
    char *			buf;
    size_t			max_nbytes;
    size_t			wait_for_nbytes;
    size_t			nbytes_read;
    void *			arg;
    void	(*callback)(void *arg,
			    int fd,
			    char *buf,
			    size_t nbytes_read,
			    char **new_buf,
			    size_t *new_max_nbytes,
			    size_t *new_wait_for_nbytes);
    void	(*error_callback)(void *arg,
				  int fd,
				  char *buf,
				  size_t nbytes,
				  int error);
} read_info_t;

typedef struct _write_info_t
{
    char *			buf;
    size_t			nbytes;
    size_t			nbytes_wrote;
    void *			arg;
    void	(*callback)(void *arg,
			    int fd,
			    char *buf,
			    size_t nbytes);
    void	(*error_callback)(void *arg,
				  int fd,
				  char *buf,
				  size_t nbytes,
				  int error);
} write_info_t;

typedef struct _writev_info_t
{
    struct iovec *	orig_iov;
    size_t		orig_iovcnt;
    struct iovec *	iov;
    size_t		iovcnt;
    size_t		nbytes_wrote;
    void *		arg;
    void	(*callback)(void *arg,
			    int fd,
			    struct iovec *iov,
			    size_t iovcnt,
			    size_t nbytes);
    void	(*error_callback)(void *arg,
				  int fd,
				  struct iovec *iov,
				  size_t iovcnt,
				  size_t nbytes,
				  int error);
} writev_info_t;


/*
 * Stuff for keeping track of file descriptors.
 *   fd_tablesize:	Maximum number of fds that system supports
 *   fd_table:		Array of pointers to select_info_t's. This is used
 *				to map from an fd to the appropriate
 *				proto or incoming data structure.
 *   n_fds_in_use:	Number of entries in fd_table that are in use
 *   *_fds:		The current fd_set that should be used for select().
 *				* = read, write, or except
 *   fd_table_modified:	Set to NEXUS_TRUE whenever the fd table
 *				changes.  This signals the
 *				handler thread which is blocked in
 *				a select that the fd_table is
 *				different than when it was called,
 *				and therefore should not be trusted.
 */

static int			fd_tablesize;
static select_info_t **		fd_table;
static int			n_fds_in_use;
static fd_set			read_fds;
static fd_set			write_fds;
static fd_set			except_fds;
static int			nexusl_fd_num_set;
static volatile nexus_bool_t	nexusl_fd_table_modified = NEXUS_FALSE;


#define nexusl_fd_read_isregistered(fd)		(FD_ISSET(fd, &read_fds))
#define nexusl_fd_write_isregistered(fd)	(FD_ISSET(fd, &write_fds))
#define nexusl_fd_except_isregistered(fd)	(FD_ISSET(fd, &except_fds))

#ifdef HAVE_SYSCONF
#   define NUM_FDS sysconf(_SC_OPEN_MAX)
#else
#   define NUM_FDS getdtablesize()
#endif


/*
 * Variables global to this module.
 */
static volatile int		nexusl_fd_select_cnt = 0;
static volatile nexus_bool_t	nexusl_fd_select_active = NEXUS_FALSE;
static volatile nexus_bool_t	nexusl_fd_select_wakeup_pending = NEXUS_FALSE;
static volatile nexus_bool_t	nexusl_fd_shutdown_called;
static listener_info_t *	listener_head;
static listener_info_t *	listener_tail;

/* pipe to myself; used to wakeup select() loop; 0 = read, 1 = write */
static int			nexusl_fd_mypipe[2];

#if !defined(BUILD_LITE)
    static nexus_bool_t		nexusl_fd_using_handler_thread;
    static int			nexusl_fd_num_handling_events;
    static nexus_thread_key_t	i_am_handler_thread_key;
#   define I_Am_Handler_Thread()					\
      (nexus_thread_getspecific(i_am_handler_thread_key) ? 1 : 0)
#endif /* !defined(BUILD_LITE) */

static nexus_bool_t defer_events;
static void (*defer_add_func)(int fd);
static void (*defer_remove_func)(int fd);

#if defined(BUILD_LITE)
#   define nexusl_fd_mutex_lock()
#   define nexusl_fd_mutex_unlock()
#   define nexusl_fd_mutex_acquired() (NEXUS_TRUE)
#   define nexusl_fd_cond_signal()
#   define nexusl_fd_cond_wait()
#else  /* defined(BUILD_LITE) */
    static nexus_mutex_t		nexusl_fd_mutex;
    static nexus_cond_t			nexusl_fd_cond;
    static volatile int			nexusl_fd_mutex_cnt = 0;
    static volatile int			nexusl_fd_cond_cnt = 0;

#   define nexusl_fd_mutex_acquired() ((nexusl_fd_mutex_cnt > 0)	\
				       ? NEXUS_TRUE			\
				       : NEXUS_FALSE)

#   define nexusl_fd_mutex_lock()			\
    {							\
        nexus_mutex_lock(&nexusl_fd_mutex);		\
        nexusl_fd_mutex_cnt++;				\
							\
	NexusAssert2((nexusl_fd_mutex_cnt < 2),		\
		     ("nexusl_fd_mutex_lock()\n"));	\
    }

#   define nexusl_fd_mutex_unlock()			\
    {							\
	NexusAssert2((nexusl_fd_mutex_acquired()),	\
		     ("nexusl_fd_mutex_unlock()\n"));	\
							\
        nexusl_fd_mutex_cnt--;				\
        nexus_mutex_unlock(&nexusl_fd_mutex);		\
    }

#   define nexusl_fd_cond_signal()			\
    {							\
	NexusAssert2((nexusl_fd_mutex_acquired()),	\
		     ("nexusl_fd_cond_signal()\n"));	\
							\
	if (nexusl_fd_cond_cnt > 0)			\
	{						\
	    nexus_cond_signal(&nexusl_fd_cond);		\
	}						\
    }

#   define nexusl_fd_cond_wait()				\
    {								\
	NexusAssert2((nexusl_fd_mutex_acquired()),		\
		     ("nexusl_fd_cond_wait()\n"));		\
								\
        nexusl_fd_mutex_cnt--;					\
        nexusl_fd_cond_cnt++;					\
	nexus_cond_wait(&nexusl_fd_cond, &nexusl_fd_mutex);	\
        nexusl_fd_cond_cnt--;					\
        nexusl_fd_mutex_cnt++;					\
    }
#endif /* (else) defined(BUILD_LITE) */

/* callbacks */
static void read_callback(void *arg, int fd);
static void write_callback(void *arg, int fd);
static void writev_callback(void *arg, int fd);
static void nexusl_fd_mypipe_callback(void *arg, int fd);


/******************************************************************************
*******************************************************************************
				Local Support Functions
*******************************************************************************
******************************************************************************/

/******************************************************************************
Function:	nexusl_fd_select_wakeup()

Description:	

		Note: the FD mutex must be acquired before calling this func

Parameters:	

Returns:	
******************************************************************************/
#if !defined(NEXUS_LITE)
static nexus_bool_t
nexusl_fd_select_wakeup()
{
    char				byte = '\0';
    int					rc;

    nexus_debug_printf(3, ("nexusl_fd_select_wakeup(): entering\n"));

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_select_wakeup()\n"));

    if (!nexusl_fd_select_active || nexusl_fd_select_wakeup_pending)
    {
	rc = NEXUS_TRUE;
	goto fn_exit;
    }

    nexus_debug_printf(5,
		       ("nexusl_fd_select_wakeup(): poking handler thread\n"));

    while ((rc = write(nexusl_fd_mypipe[1], &byte, sizeof(char))) == -1 &&
	   errno == EINTR)
    {
	/* do nothing */
    }

    if (rc > 0)
    {
	nexusl_fd_select_wakeup_pending = NEXUS_TRUE;
	rc = NEXUS_TRUE;
    }
    else
    {
	rc = NEXUS_FALSE;
    }

  fn_exit:
    nexus_debug_printf(3, ("nexusl_fd_select_wakeup(): exiting\n"));

    return rc;
}
/* nexusl_fd_select_wakeup() */
#else
#   define nexusl_fd_select_wakeup()
#endif


/******************************************************************************
Function:	nexusl_fd_table_add()

Description:	

		Note: the FD mutex must be acquired before calling this func

Parameters:	

Returns:	
******************************************************************************/
static void
nexusl_fd_table_add(int fd)
{
    select_info_t *select_info;

    nexus_debug_printf(3,
		       ("nexusl_fd_table_add(): entering, fd=%d\n", fd));

    NexusAssert2((nexusl_fd_mutex_acquired()), ("nexusl_fd_table_add()\n"));

    if (fd_table[fd])
    {
	goto fn_exit;
    }
    
    NexusMalloc(fd_table_add(),
		select_info,
		select_info_t *,
		sizeof(select_info_t));
    select_info->fd = fd;
    select_info->read_callback = NULL;
    select_info->read_arg = NULL;
    select_info->write_callback = NULL;
    select_info->write_arg = NULL;
    select_info->except_callback = NULL;
    select_info->except_arg = NULL;

    fd_table[fd] = select_info;
    n_fds_in_use++;

    /*
     * TODO: This should really be located in the calling routines...
     */
    if (defer_events)
    {
	defer_add_func(fd);
    }

  fn_exit:
    nexus_debug_printf(3, ("nexusl_fd_table_add(): exiting\n"));
}
/* fd_table_add() */


/******************************************************************************
Function:	nexusl_fd_table_remove()

Description:	free up table resources being held for a file descriptor

		Note: the FD mutex must be acquired before calling this func

Parameters:	fd		file descriptor
		select_info	pointer to the select information structure

Returns:	none
******************************************************************************/
static void
nexusl_fd_table_remove(int fd)
{
    select_info_t *			select_info;

    nexus_debug_printf(3,
		       ("nexusl_fd_table_remove(): entering, fd=%d\n", fd));

    NexusAssert2((nexusl_fd_mutex_acquired()), ("nexusl_fd_table_remove()\n"));

    select_info = fd_table[fd];
    if (select_info != NULL)
    {
	fd_table[fd] = (void *) NULL;
	n_fds_in_use--;
	nexusl_fd_table_modified = NEXUS_TRUE;

	if (defer_events)
	{
	    defer_remove_func(fd);
	}

	NexusFree(select_info);
    }

    nexus_debug_printf(3, ("nexusl_fd_table_remove(): exiting\n"));
}
/* fd_table_remove() */


/******************************************************************************
Function:	nexusl_fd_read_register()

Description:	add the specified file descriptor to the read select list and
		register an associated callback function

		Note: the FD mutex must be acquired before calling this func

Parameters:	

Returns:	
******************************************************************************/
static void
nexusl_fd_read_register(int fd,
			nexusl_fd_cbf_t callback_func,
			void * callback_arg)
{
    select_info_t *			select_info;

    nexus_debug_printf(3,
		       ("nexusl_fd_read_register(): entering, fd=%d\n", fd));

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_read_register()\n"));

    nexusl_fd_table_add(fd);
    select_info = fd_table[fd];

    NexusAssert2((!FD_ISSET(fd, &read_fds)),
		 ("nexusl_fd_read_register(): fd=%d\n", fd));

    NexusAssert2((!select_info->read_callback),
		 ("nexusl_fd_read_register(): fd=%d\n", fd));

    select_info->read_callback = callback_func;
    select_info->read_arg = callback_arg;

    FD_SET(fd, &read_fds);
    nexusl_fd_num_set++;

    if (nexusl_fd_select_active)
    {
	nexusl_fd_select_wakeup();
    }

    nexus_debug_printf(3, ("nexusl_fd_read_register(): exiting\n"));
}
/* fd_read_register() */


/******************************************************************************
Function:	nexusl_fd_read_unregister()

Description:	

		Note: the FD mutex must be acquired before calling this func

Parameters:	

Returns:	
******************************************************************************/
static void
nexusl_fd_read_unregister(int fd)
{
    select_info_t *			select_info;

    nexus_debug_printf(3,
		       ("nexusl_fd_read_unregister(): entering, fd=%d\n", fd));

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_read_unregister()\n"));

    nexusl_fd_table_add(fd);
    select_info = fd_table[fd];

    NexusAssert2((FD_ISSET(fd, &read_fds)),
		 ("nexusl_fd_read_unregister(): fd=%d\n", fd));

    NexusAssert2((select_info->read_callback),
		 ("nexusl_fd_read_unregister(): fd=%d\n", fd));

    select_info->read_callback = NULL;
    select_info->read_arg = NULL;

    FD_CLR(fd, &read_fds);
    nexusl_fd_num_set--;
    nexusl_fd_table_modified = NEXUS_TRUE;
    
    nexus_debug_printf(3, ("nexusl_fd_read_unregister(): exiting\n"));
}
/* fd_read_unregister() */


/******************************************************************************
Function:	nexusl_fd_write_register()

Description:	add the specified file descriptor to the write select list and
		register an associated callback function

		Note: the FD mutex must be acquired before calling this func

Parameters:	

Returns:	
******************************************************************************/
static void
nexusl_fd_write_register(int fd,
			 nexusl_fd_cbf_t callback_func,
			 void * callback_arg)
{
    select_info_t *			select_info;

    nexus_debug_printf(3,
		       ("nexusl_fd_write_register(): entering, fd=%d\n", fd));

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_write_register()\n"));

    nexusl_fd_table_add(fd);
    select_info = fd_table[fd];

    NexusAssert2((!FD_ISSET(fd, &write_fds)),
		 ("nexusl_fd_write_register(): fd=%d\n", fd));

    NexusAssert2((!select_info->write_callback),
		 ("nexusl_fd_write_register(): fd=%d\n", fd));

    FD_SET(fd, &write_fds);
    select_info->write_callback = callback_func;
    select_info->write_arg = callback_arg;

    FD_SET(fd, &write_fds);
    nexusl_fd_num_set++;

    if (nexusl_fd_select_active)
    {
	nexusl_fd_select_wakeup();
    }

    nexus_debug_printf(3, ("nexusl_fd_write_register(): exiting\n"));
}
/* fd_write_register() */


/******************************************************************************
Function:	nexusl_fd_write_unregister()

Description:	

		Note: the FD mutex must be acquired before calling this func

Parameters:	

Returns:	
******************************************************************************/
static void
nexusl_fd_write_unregister(int fd)
{
    select_info_t *			select_info;

    nexus_debug_printf(3, ("nexusl_fd_write_unregister(): entering, fd=%d\n",
			   fd));

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_write_unregister()\n"));

    nexusl_fd_table_add(fd);
    select_info = fd_table[fd];

    NexusAssert2((FD_ISSET(fd, &write_fds)),
		 ("nexusl_fd_write_unregister(): fd=%d\n", fd));

    NexusAssert2((select_info->write_callback),
		 ("nexusl_fd_write_unregister(): fd=%d\n", fd));

    select_info->write_callback = NULL;
    select_info->write_arg = NULL;

    FD_CLR(fd, &write_fds);
    nexusl_fd_num_set--;
    nexusl_fd_table_modified = NEXUS_TRUE;
    
    nexus_debug_printf(3, ("nexusl_fd_write_unregister(): exiting\n"));
}
/* fd_write_unregister() */


/******************************************************************************
Function:	nexusl_fd_except_register()

Description:	add the specified file descriptor to the exception select list
		and register an associated callback function

		Note: the FD mutex must be acquired before calling this func

Parameters:	

Returns:	
******************************************************************************/
static void
nexusl_fd_except_register(int fd,
			  nexusl_fd_cbf_t callback_func,
			  void * callback_arg)
{
    select_info_t *			select_info;

    nexus_debug_printf(3,
		       ("nexusl_fd_except_register(): entering, fd=%d\n", fd));

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_except_register()\n"));

    nexusl_fd_table_add(fd);
    select_info = fd_table[fd];

    NexusAssert2((!FD_ISSET(fd, &except_fds)),
		 ("nexusl_fd_except_register(): fd=%d\n", fd));

    NexusAssert2((!select_info->except_callback),
		 ("nexusl_fd_except_register(): fd=%d\n", fd));

    select_info->except_callback = callback_func;
    select_info->except_arg = callback_arg;

    FD_SET(fd, &except_fds);
    nexusl_fd_num_set++;

    if (nexusl_fd_select_active)
    {
	nexusl_fd_select_wakeup();
    }

    nexus_debug_printf(3, ("nexusl_fd_except_register(): exiting\n"));
}
/* fd_except_register() */


/******************************************************************************
Function:	nexusl_fd_except_unregister()

Description:	

		Note: the FD mutex must be acquired before calling this func

Parameters:	

Returns:	
******************************************************************************/
static void
nexusl_fd_except_unregister(int fd)
{
    select_info_t *			select_info;

    nexus_debug_printf(3, ("nexusl_fd_except_unregister(): entering, fd=%d\n",
			   fd));

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_except_unregister()\n"));

    nexusl_fd_table_add(fd);
    select_info = fd_table[fd];

    NexusAssert2((FD_ISSET(fd, &except_fds)),
		 ("nexusl_fd_except_unregister(): fd=%d\n", fd));

    NexusAssert2((select_info->except_callback),
		 ("nexusl_fd_except_unregister(): fd=%d\n", fd));

    select_info->except_callback = NULL;
    select_info->except_arg = NULL;

    FD_CLR(fd, &except_fds);
    nexusl_fd_num_set--;
    nexusl_fd_table_modified = NEXUS_TRUE;
    
    nexus_debug_printf(3, ("nexusl_fd_except_unregister(): exiting\n"));
}
/* fd_except_unregister() */


/******************************************************************************
Function:	nexusl_fd_setup_nonblocking()

Description:	Set the passed fd to be non-blocking so that the event driver
		can use it properly.

		TODO: Do we need the F_SETFD to set the close-on-exec bit?

		Note: the FD mutex must be acquired before calling this func

Parameters:	

Returns:	
******************************************************************************/
static int
nexusl_fd_setup_nonblocking(int fd)
{
    int save_errno;
    int flags;
    int rc = 0;

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_setup_nonblocking()\n"));

    if ((flags = fcntl(fd, F_GETFL, 0)) < 0)
    {
	save_errno = errno;
	rc = -save_errno;
	goto fn_exit;
    }

#   ifdef TARGET_ARCH_HPUX
    {
	flags |= O_NONBLOCK;
    }
#   else
    {
	flags |= O_NDELAY;
    }
#   endif    

    if (fcntl(fd, F_SETFL, flags) == -1)
    {
	save_errno = errno;
	rc = -save_errno;
	goto fn_exit;
    }

    nexusl_fd_table_add(fd);

  fn_exit:
    return(rc);
}
/* nexusl_fd_setup_nonblocking() */


/******************************************************************************
Function:	nexusl_fd_setup_blocking()

Description:	Set the passed fd to be blocking.

		Note: the FD mutex must be acquired before calling this func

		Note: The fd cannot be used with the event driver after this.

Parameters:	

Returns:	
******************************************************************************/
static int
nexusl_fd_setup_blocking(int fd)
{
    int save_errno;
    int flags;
    int flag_mask;
    int rc = 0;

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_setup_blocking()\n"));

    if ((flags = fcntl(fd, F_GETFL, 0)) < 0)
    {
	save_errno = errno;
	rc = -save_errno;
	goto fn_exit;
    }

    flag_mask = -1;
#   if defined(TARGET_ARCH_HPUX)
    {
	flag_mask ^= O_NONBLOCK;
    }
#   else
    {
	flag_mask ^= O_NDELAY;
    }
#   endif
    flags &= flag_mask;
    
    if (fcntl(fd, F_SETFL, flags) == -1)
    {
	save_errno = errno;
	rc = -save_errno;
	goto fn_exit;
    }

  fn_exit:
    return(rc);
}
/* nexusl_fd_setup_blocking() */


/******************************************************************************
Function:	nexusl_fd_except_setup_socket()

Description:	Set the passed fd to be non-blocking so that the event driver
		can use it properly.  Also set the socket for TCP_NODELAY
		and SO_KEEPALIVE.

		TODO: do we need the F_SETFD to set the close-on-exec bit?

		Note: the FD mutex must be acquired before calling this func

Parameters:	none

Returns:	0 if successful, -errno if an error occurs
******************************************************************************/
static int
nexusl_fd_setup_socket(int fd)
{
    int one = 1;
    int save_errno;
    int rc = 0;

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_setup_socket()\n"));

#   ifdef TCP_NODELAY
    {
	if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (char *) &one,
		       sizeof(one)) < 0)
	{
	    save_errno = errno;
	    rc = -save_errno;
	    goto setup_abort;
	}
    }
#   endif
    
#   ifdef SO_KEEPALIVE
    {
	if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *) &one,
		       sizeof(one)) < 0)
	{
	    save_errno = errno;
	    rc = -save_errno;
	    goto setup_abort;
	}
    }
#   endif
    
#   if defined(HAVE_RIO_HACKS) && defined(SO_SNDBUF)
    {
	/* That's 1/2 Mbytes + 16 Kbyte. */
	sock_buffsize = 512 * 1024 + 16384;
	
	if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *) &sock_buffsize,
		       sizeof(sock_buffsize)) < 0)
	{
	    save_errno = errno;
	    rc = -save_errno;
	    goto setup_abort;
	}
    }
#   endif

#   if defined(HAVE_RIO_HACKS) && defined(SO_RCVBUF)
    {
	/* That's 1/2 Mbytes + 16 Kbyte. */
	sock_buffsize = 512 * 1024 + 16384;

	if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *) &sock_buffsize,
		       sizeof(sock_buffsize)) < 0)
	{
	    save_errno = errno;
	    rc = -save_errno;
	    goto setup_abort;
	}
    }
#   endif
    
    rc = nexusl_fd_setup_nonblocking(fd);

  setup_abort:
    return(rc);
}
/* nexusl_fd_setup_socket() */


/******************************************************************************
Function:	nexusl_fd_close()

Description:	close the specified FD

		Note: the FD mutex must be acquired before calling this func

Parameters:	fd			file descriptor to close

Returns:	0			success
		-value			errno reported by close()
******************************************************************************/
static int
nexusl_fd_close(int fd)
{
    nexus_bool_t done = NEXUS_FALSE;
    int save_errno;
    int rc = 0;

    nexus_debug_printf(3, ("nexusl_fd_close(): entering, fd=%d\n", fd));

    NexusAssert2((nexusl_fd_mutex_acquired()),
		 ("nexusl_fd_close()\n"));

    /*
     * If callbacks are registered for this file descriptor, unregister them
     */
    if (nexusl_fd_read_isregistered(fd))
    {
	nexusl_fd_read_unregister(fd);
    }

    if (nexusl_fd_write_isregistered(fd))
    {
	nexusl_fd_write_unregister(fd);
    }

    if (nexusl_fd_except_isregistered(fd))
    {
	nexusl_fd_except_unregister(fd);
    }

    /*
     * Remove this FD from the FD table
     */
#   if 0
    {
	/* Disabled to reduce number of mallocs */
	nexusl_fd_table_remove(fd);
    }
#   endif

    /*
     * Force the handler thread to return from the select() before attempting
     * to close() listeners.  This is necessary because we may be in the
     * select() and some machines like the SGI running IRIX 6.4 hang if close()
     * is called on a FD which is also being monitored by select().
     */
    if (nexusl_fd_select_active && nexusl_fd_table_modified)
    {
	unsigned			cnt = nexusl_fd_select_cnt;
	
	nexusl_fd_select_wakeup();

	while(cnt == nexusl_fd_select_cnt)
	{
	    nexusl_fd_cond_wait();
	}
    }

    /*
     * Close() the FD
     */
    while (!done)
    {
        if (close(fd))
        {
	    save_errno = errno;
	    if (save_errno != EINTR)
	    {
		done = NEXUS_TRUE;
	        rc = -save_errno;
	    }
        }
	else
	{
	    done = NEXUS_TRUE;
	}
    }


    nexus_debug_printf(3, ("nexusl_fd_close(): exiting\n"));

    return(rc);
}
/* nexusl_fd_close() */


/******************************************************************************
Function:	nexusl_fd_listener_close()

Description:	close the specified listener

		Note: the FD mutex must be acquired before calling this func

Parameters:	cur_listener		pointer to the listener info structure

Returns:	0			success
		TODO:			check other error codes
******************************************************************************/
/*
 * nexusl_fd_listener_close()
 *
 * stop listening on the passed listener
 */
static int
nexusl_fd_listener_close(listener_info_t *cur_listener)
{
    if (cur_listener == listener_head)
    {
	listener_head = cur_listener->next;
    }
    if (cur_listener == listener_tail)
    {
	listener_tail = cur_listener->prev;
    }

    if (cur_listener->prev)
    {
	cur_listener->prev->next = cur_listener->next;
    }
    if (cur_listener->next)
    {
	cur_listener->next->prev = cur_listener->prev;
    }

    nexusl_fd_read_unregister(cur_listener->fd);
    nexusl_fd_close(cur_listener->fd);
    NexusFree(cur_listener);

    return(0);
}
/* nexusl_fd_listener_close() */


/******************************************************************************
*******************************************************************************
				 API Functions
*******************************************************************************
******************************************************************************/

/*
 * nexus_fd_init()
 *
 *  Init all the global variables and setup mutexes.
 */
int nexus_fd_init(int *argc, char ***argv)
{
    int i;
    
    nexus_debug_printf(3, ("nexusl_fd_init(): entering\n"));

    nexusl_fd_shutdown_called = NEXUS_FALSE;
    listener_head = listener_tail = NULL;
#   if !defined(BUILD_LITE)
    {
	nexus_mutex_init(&nexusl_fd_mutex, (nexus_mutexattr_t *) NULL);
	nexus_cond_init(&nexusl_fd_cond, (nexus_condattr_t *) NULL);

	nexusl_fd_using_handler_thread = NEXUS_FALSE;
	nexusl_fd_num_handling_events = 0;
	nexus_thread_key_create(&i_am_handler_thread_key, NULL);
    }
#   endif

    nexusl_fd_mutex_lock();

    /* Initialize the fd tables and fd_set's */
    fd_tablesize = NUM_FDS;
    NexusMalloc(nexus_fd_init(),
		fd_table,
		select_info_t **,
		(sizeof(select_info_t *) * fd_tablesize));
    for (i = 0; i < fd_tablesize; i++)
    {
	fd_table[i] = (select_info_t *) NULL;
    }
    nexusl_fd_table_modified = NEXUS_FALSE;
    FD_ZERO(&read_fds);
    FD_ZERO(&write_fds);
    FD_ZERO(&except_fds);
    nexusl_fd_num_set = 0;
    
    /*
     * Create a pipe to myself, so that I can wake up the thread that is
     * blocked on a select().
     */
    if (pipe(nexusl_fd_mypipe) != 0)
    {
	return(-1);
    }
    nexusl_fd_setup_nonblocking(nexusl_fd_mypipe[0]);
    nexusl_fd_read_register(nexusl_fd_mypipe[0],
			    nexusl_fd_mypipe_callback,
			    NULL);

    defer_events = NEXUS_FALSE;
    defer_add_func = defer_remove_func = NULL;

    nexusl_fd_mutex_unlock();

    nexus_debug_printf(3, ("nexusl_fd_init(): exiting\n"));
    
    return(0);
} /* nexus_fd_init() */


/*
 * nexus_fd_set_handler_thread()
 */
void nexus_fd_set_handler_thread(nexus_bool_t i_am_handler_thread)
{
#ifndef BUILD_LITE
    nexus_thread_setspecific(i_am_handler_thread_key,
			     (void *) i_am_handler_thread);
    nexusl_fd_using_handler_thread = NEXUS_TRUE;
#endif
} /* nexus_fd_set_handler_thread() */


/*
 * nexus_fd_tablesize()
 *
 * Return the file descriptor table size
 */
int nexus_fd_tablesize()
{
    return(fd_tablesize);
} /* nexus_fd_tablesize() */


/*
 * nexus_fd_shutdown()
 *
 * Make sure the select() wakes up and set the nexusl_fd_shutdown_called flag
 * to TRUE
 */
int
nexus_fd_shutdown(void)
{
    listener_info_t *			cur_listener;
    listener_info_t *			next_listener;
    int					fd;

    nexus_debug_printf(3, ("nexus_fd_shutdown(): entering\n"));

    nexusl_fd_mutex_lock();

    nexusl_fd_shutdown_called = NEXUS_TRUE;

    /*
     * Wakeup the handler thread from the select(), close the pipe to ourself,
     * and shutdown all listeners.
     */
    if (nexusl_fd_select_active)
    {
	nexusl_fd_select_wakeup();
    }

    nexusl_fd_close(nexusl_fd_mypipe[0]);
    nexusl_fd_close(nexusl_fd_mypipe[1]);

    for (cur_listener = listener_head;
	 cur_listener;
	 cur_listener = next_listener)
    {
	next_listener = cur_listener->next;
	nexusl_fd_listener_close(cur_listener);
    }

    /*
     * free up table resources
     */
    for (fd = 0; fd < fd_tablesize; fd++)
    {
	nexusl_fd_table_remove(fd);
    }
    NexusFree(fd_table);

    nexusl_fd_mutex_unlock();
    nexus_debug_printf(3, ("nexus_fd_shutdown(): exiting\n"));
    
    return(0);
} /* nexus_fd_shutdown() */


/*
 * nexus_fd_open()
 *
 * Open the fd in non-blocking mode
 */
int
nexus_fd_open(char *path,
	      int flags,
	      int mode,
	      int * fd_p)
{
    int					fd;
    int					real_flags;
    nexus_bool_t			done = NEXUS_FALSE;
    int					rc = 0;
    int					save_errno;

    nexus_debug_printf(3, ("nexus_fd_open(): entering\n"));

    nexusl_fd_mutex_lock();
	
    real_flags = flags | O_NDELAY | ~O_SYNC;

    while (!done)
    {
        fd = open(path, real_flags, mode);
	save_errno = errno;
	
        if (fd < 0)
        {
	    if (save_errno != EINTR)
	    {
	        rc = -save_errno;
		goto abort;
	    }
	}
	else
	{
	    done = NEXUS_TRUE;
	}
    }

    nexusl_fd_table_add(fd);

    *fd_p = fd;

  abort:
    nexusl_fd_mutex_unlock();
    nexus_debug_printf(3, ("nexus_fd_open(): exiting\n"));
    return(rc);
} /* nexus_fd_open() */


/*
 * nexus_fd_setup_nonblocking()
 */
int nexus_fd_setup_nonblocking(int fd)
{
    int rc;
    nexus_debug_printf(3, ("nexus_fd_setup_nonblocking(): entering\n"));
    nexusl_fd_mutex_lock();
    rc = nexusl_fd_setup_nonblocking(fd);
    nexusl_fd_mutex_unlock();
    nexus_debug_printf(3, ("nexus_fd_setup_nonblocking(): exiting\n"));
    return(rc);
} /* nexus_fd_setup_nonblocking() */


/*
 * nexus_fd_setup_blocking()
 */
int nexus_fd_setup_blocking(int fd)
{
    int rc;
    nexus_debug_printf(3, ("nexus_fd_setup_blocking(): entering\n"));
    nexusl_fd_mutex_lock();
    rc = nexusl_fd_setup_blocking(fd);
    nexusl_fd_mutex_unlock();
    nexus_debug_printf(3, ("nexus_fd_setup_blocking(): exiting\n"));
    return(rc);
} /* nexus_fd_setup_blocking() */


/*
 * nexus_fd_setup_socket()
 */
int nexus_fd_setup_socket(int fd)
{
    int rc;
    nexus_debug_printf(3, ("nexus_fd_setup_socket(): entering\n"));
    nexusl_fd_mutex_lock();
    rc = nexusl_fd_setup_socket(fd);
    nexusl_fd_mutex_unlock();
    nexus_debug_printf(3, ("nexus_fd_setup_socket(): exiting\n"));
    return(rc);
} /* nexus_fd_setup_socket() */


/*
 * nexus_fd_close()
 *
 * Close the passed fd.
 */
int nexus_fd_close(int fd)
{
    int rc;

    nexus_debug_printf(3, ("nexus_fd_close(): entering\n"));
#ifndef USE_THREAD_SAFE_FD_CALLS
    nexusl_fd_mutex_lock();
#endif

    rc = nexusl_fd_close(fd);

#ifndef USE_THREAD_SAFE_FD_CALLS
    nexusl_fd_mutex_unlock();
#endif
    nexus_debug_printf(3, ("nexus_fd_close(): exiting\n"));
    return(rc);
} /* nexus_fd_close() */


/*
 * nexus_fd_register_for_select()
 *
 * Register the passed funcs to be called when the fd is able to
 * to read/write on it.
 */
int nexus_fd_register_for_select(int fd,
			         void (*read_callback_func)(void *arg,
							    int fd),
			         void (*write_callback_func)(void *arg,
							     int fd),
			         void (*except_callback_func)(void *arg,
				 			      int fd),
			         void *callback_arg)
{
    nexus_debug_printf(3, ("nexus_fd_register_for_select(): entering\n"));
    nexusl_fd_mutex_lock();

    if (read_callback_func != NULL)
    {
	nexusl_fd_read_register(fd, read_callback_func, callback_arg);
    }

    if (write_callback_func != NULL)
    {
	nexusl_fd_write_register(fd, write_callback_func, callback_arg);
    }

    if (except_callback_func != NULL)
    {
	nexusl_fd_except_register(fd, except_callback_func, callback_arg);
    }
			 
    nexusl_fd_mutex_unlock();
    nexus_debug_printf(3, ("nexus_fd_register_for_select(): exiting\n"));
    return(0);
} /* nexus_fd_register_for_select() */


/*
 * nexus_fd_register_for_read()
 *
 * read on fd, filling buf with nbytes of data, calling the proper callback
 * when done
 *
 * This function can do one of two things:
 *   1. Call select() until this fd is used and then call the appropriate
 *        callback, or
 *   2. Create a new thread that does a blocking read on the fd.
 */
int nexus_fd_register_for_read(int fd,
			       char *buf,
			       size_t max_nbytes,
			       size_t wait_for_nbytes,
			       void (*callback_func)(
				                  void *arg,
						  int fd,
						  char *buf,
						  size_t nbytes,
						  char **new_buf,
						  size_t *new_max_nbytes,
						  size_t *new_wait_for_nbytes),
			       void (*error_callback_func)(void *arg,
							   int fd,
							   char *buf,
							   size_t nbytes,
							   int error),
			       void *callback_arg)
{
    read_info_t *read_info;

    nexus_debug_printf(3, ("nexus_fd_register_for_read(): entering, "
			   "fd=%d, max=%d, min=%d\n",
			   fd, max_nbytes, wait_for_nbytes));

    NexusMalloc(nexus_fd_register_for_read(),
		read_info,
		read_info_t *,
		sizeof(read_info_t));
    
    read_info->buf = buf;
    read_info->max_nbytes = max_nbytes;
    read_info->wait_for_nbytes = wait_for_nbytes;
    read_info->nbytes_read = 0;
    read_info->arg = callback_arg;
    read_info->callback = callback_func;
    read_info->error_callback = error_callback_func;

    nexusl_fd_mutex_lock();
    nexusl_fd_read_register(fd,
			    read_callback,
			    (void *) read_info);
    nexusl_fd_mutex_unlock();

    nexus_debug_printf(3, ("nexus_fd_register_for_read(): exiting\n"));
    return(0);
} /* nexus_fd_register_for_read() */


/*
 * read_callback()
 *
 * Callback to do work for nexus_fd_register_for_read()
 */
static void read_callback(void *arg, int fd)
{
    read_info_t *read_info = (read_info_t *) arg;
    long n_read;
    int save_errno;
    nexus_byte_t done;

    nexus_debug_printf(5,("read_callback(): entering\n"));
    for (done = NEXUS_FALSE; !done; )
    {
#ifndef USE_THREAD_SAFE_FD_CALLS
	nexusl_fd_mutex_lock();
#endif

	nexus_debug_printf(5,("read_callback(): calling read, fd=%i, buf=%lx, size=%i\n", fd, (read_info->buf + read_info->nbytes_read), (read_info->max_nbytes - read_info->nbytes_read)));
	n_read = read(fd,
		      (read_info->buf + read_info->nbytes_read),
		      (read_info->max_nbytes - read_info->nbytes_read));
	save_errno = errno;
	nexus_debug_printf(5,("read_callback(): read returned n_read=%li\n", n_read));

#ifdef BUILD_DEBUG
	if (NexusDebug(7))
	{
	    int i;
	    unsigned char *start;
	    
	    start = (unsigned char *)(read_info->buf
	    			      + read_info->nbytes_read);

	    nexus_printf("Received %d bytes of data on fd %d\n",
	    		 n_read,
			 fd);

	    nexus_stdio_lock();
	    printf("First %d bytes:  ", NEXUS_MIN(n_read, 100));
	    for (i = 0; i < NEXUS_MIN(n_read, 100); i++)
	    {
		printf("%2x  ", *(start + i));
	    }
	    printf("\n"); fflush(stdout);
	    nexus_stdio_unlock();
	}
#endif

#ifndef USE_THREAD_SAFE_FD_CALLS
	nexusl_fd_mutex_unlock();
#endif
	
	/*
	 * n_read: is > 0 if it successfully read some bytes
	 *         is < 0 on error -- need to check errno
	 *         is 0 on EOF
	 */
	
	if (n_read > 0)
	{
	    read_info->nbytes_read += n_read;
	    if (read_info->nbytes_read >= read_info->wait_for_nbytes)
	    {
		char *new_buf = (char *) NULL;
		size_t new_max_nbytes = 0;
		size_t new_wait_for_nbytes = 0;
		(*read_info->callback)(read_info->arg,
				       fd,
				       read_info->buf,
				       read_info->nbytes_read,
				       &new_buf,
				       &new_max_nbytes,
				       &new_wait_for_nbytes);
		if (   new_buf
		       && (new_max_nbytes > 0)
		       && (new_wait_for_nbytes > 0))
		{
		    read_info->buf = new_buf;
		    read_info->max_nbytes = new_max_nbytes;
		    read_info->wait_for_nbytes = new_wait_for_nbytes;
		    read_info->nbytes_read = 0;
		}
		else
		{
		    NexusFree(read_info);
		    done = NEXUS_TRUE;
		}
	    }
	    else
	    {
		nexusl_fd_mutex_lock();
		nexusl_fd_read_register(fd,
					read_callback,
					(void *)read_info);
		nexusl_fd_mutex_unlock();
	        done = NEXUS_TRUE;
	    }
	}
	else if (n_read == 0)
	{
	    (*read_info->error_callback)(read_info->arg,
					 fd,
					 read_info->buf,
					 read_info->nbytes_read,
					 0);
	    NexusFree(read_info);
	    done = NEXUS_TRUE;
	}
	else /* n_read < 0 */
	{

	    nexus_debug_printf(3, ("read_callback(): ERROR, errno=%d, fd=%d\n",
			save_errno, fd));
	    
	    if (save_errno == EINTR)
	    {
		/* Try again */
	    }
	    else if (save_errno == EAGAIN || save_errno == EWOULDBLOCK)
	    {
		/* We've read all we can for now.  So repost the read. */
		nexusl_fd_mutex_lock();
		nexusl_fd_read_register(fd,
					read_callback,
					(void *) read_info);
		nexusl_fd_mutex_unlock();
		done = NEXUS_TRUE;
	    }
	    else
	    {
		(*read_info->error_callback)(read_info->arg,
					     fd,
					     read_info->buf,
					     read_info->nbytes_read,
					     save_errno);
		NexusFree(read_info);
		done = NEXUS_TRUE;
	    }
	}
    }

    nexus_debug_printf(5,("read_callback(): exiting\n"));
} /* read_callback */


/*
 * nexus_fd_register_for_write()
 *
 * This function can do one of two things:
 *   1. Call select() until this fd is ready, and then call the appropriate
 *        callback, or
 *   2. Create a new thread that does a blocking write on the fd.
 */
int nexus_fd_register_for_write(int fd,
				char *buf,
				size_t nbytes,
				void (*callback_func)(void *arg,
						      int fd,
						      char *buf,
						      size_t nbytes),
				void (*error_callback_func)(void *arg,
							    int fd,
							    char *buf,
							    size_t nbytes,
							    int error),
				void *callback_arg)
{
    write_info_t *write_info;
    
    nexus_debug_printf(3,
		       ("nexus_fd_register_for_write(): entering, fd=%d\n",
			fd));

    NexusMalloc(nexus_fd_register_for_write(),
		write_info,
		write_info_t *,
		sizeof(write_info_t));
    write_info->buf = buf;
    write_info->nbytes = nbytes;
    write_info->nbytes_wrote = 0;
    write_info->arg = callback_arg;
    write_info->callback = callback_func;
    write_info->error_callback = error_callback_func;

    nexusl_fd_mutex_lock();
    nexusl_fd_write_register(fd,
			     write_callback,
			     (void *) write_info);
    nexusl_fd_mutex_unlock();

    nexus_debug_printf(3, ("nexus_fd_register_for_write(): exiting\n"));
    return(0);
} /* nexus_fd_register_for_write() */


/*
 * write_callback()
 *
 * Callback to do work for nexus_fd_register_for_write()
 */
static void write_callback(void *arg, int fd)
{
    write_info_t *write_info = (write_info_t *) arg;
    long n_wrote;
    int save_errno;
    nexus_bool_t done;
    size_t nbytes;
    char *buf;

    for (done = NEXUS_FALSE; !done; )
    {
	buf = (write_info->buf + write_info->nbytes_wrote);
	nbytes = (write_info->nbytes - write_info->nbytes_wrote);
#if NEXUS_MAX_WRITE_PACKET > 0
	nbytes = MIN(NEXUS_MAX_WRITE_PACKET, nbytes);
#endif
	
#ifndef USE_THREAD_SAFE_FD_CALLS
	nexusl_fd_mutex_lock();
#endif
	
	n_wrote = write(fd, buf, nbytes);
	save_errno = errno;

#ifdef BUILD_DEBUG
	if (NexusDebug(7))
	{
	    int i;

	    nexus_printf("Sent %d bytes of data on fd %d\n",
	    		 n_wrote,
			 fd);

	    nexus_printf("First %d bytes:  ", NEXUS_MIN(n_wrote, 100));
	    nexus_stdio_lock();
	    for (i = 0; i < NEXUS_MIN(n_wrote, 100); i++)
	    {
		printf("%2x  ", *((unsigned char *)(buf + i)));
	    }
	    printf("\n"); fflush(stdout);
	    nexus_stdio_unlock();
	}
#endif
	
#ifndef USE_THREAD_SAFE_FD_CALLS
	nexusl_fd_mutex_unlock();
#endif
	
	/*
	 * n_wrote: is > 0 on success -- number of bytes written
	 *          is < 0 on error -- need to check errno
	 *          is 0 (SysV) or (-1 && errno==EWOULDBLOCK) (BSD)
	 *              if the write would block without writing anything
	 */

	if (n_wrote > 0)
	{
	    write_info->nbytes_wrote += n_wrote;
	    if (write_info->nbytes_wrote >= write_info->nbytes)
	    {
		(*write_info->callback)(write_info->arg,
					fd,
					write_info->buf,
					write_info->nbytes_wrote);
		NexusFree(write_info);
		done = NEXUS_TRUE;
	    }
	}
	else if (   (n_wrote == 0)
		 || (n_wrote < 0 && save_errno == EWOULDBLOCK))
	{
	    nexusl_fd_mutex_lock();
	    nexusl_fd_write_register(fd,
				     write_callback,
				     (void *) write_info);
	    nexusl_fd_mutex_unlock();
	    done = NEXUS_TRUE;
	}
	else /* n_wrote < 0 */
	{
	    if (save_errno == EINTR)
	    {
		/* Try again */
	    }
	    else
	    {
		(*write_info->error_callback)(write_info->arg,
					      fd,
					      write_info->buf,
					      write_info->nbytes_wrote,
					      save_errno);
		NexusFree(write_info);
		done = NEXUS_TRUE;
	    }
	}
    }
} /* write_callback() */


/*
 * nexus_fd_write_one_nonblocking()
 *
 * Try to write the one character, c, to the fd.
 * But do not block on the write.  This function should
 * return immediately whether the write succeeds or fails.
 */
int nexus_fd_write_one_nonblocking(int fd, char c)
{
    long n_wrote;
    int save_errno;
    nexus_bool_t done;
    size_t nbytes;
    char *buf;
    int rc;

    nexus_debug_printf(3, ("nexus_fd_write_one_nonblocking(): entering\n"));
#ifndef USE_THREAD_SAFE_FD_CALLS
    nexusl_fd_mutex_lock();
#endif
	
retry:
    n_wrote = write(fd, &c, 1);
    save_errno = errno;
    if ((n_wrote < 0) && (save_errno) == EINTR)
    {
	goto retry;
    }
    rc = (n_wrote > 0 ? 1 : 0);
    
#ifndef USE_THREAD_SAFE_FD_CALLS
    nexusl_fd_mutex_unlock();
#endif
    nexus_debug_printf(3, ("nexus_fd_write_one_nonblocking(): exiting\n"));
    return(rc);
} /* write_callback() */


/*
 * nexus_fd_register_for_writev()
 *
 * This function can do one of two things:
 *   1. Call select() until this fd is used and then call the appropriate
 *        callback, or
 *   2. Create a new thread that does a blocking write on the fd.
 */
int nexus_fd_register_for_writev(int fd,
				 struct iovec *iov,
				 size_t iovcnt,
				 void callback_func(void *arg,
						    int fd,
						    struct iovec *iov,
						    size_t iovcnt,
						    size_t nbytes),
				 void error_callback_func(void *arg,
							  int fd,
							  struct iovec *iov,
							  size_t iovcnt,
							  size_t nbytes,
							  int error),
				 void *callback_arg)
{
    writev_info_t *writev_info;

    nexus_debug_printf(3, ("nexus_fd_register_for_writev(): entering\n"));

    NexusMalloc(nexus_fd_register_for_write(),
		writev_info,
		writev_info_t *,
		sizeof(writev_info_t));
    writev_info->orig_iov = iov;
    writev_info->orig_iovcnt = iovcnt;
    NexusMalloc(nexus_fd_register_for_write(),
		writev_info->iov,
		struct iovec *,
		(sizeof(struct iovec) * iovcnt));
    writev_info->iovcnt = iovcnt;
    writev_info->arg = callback_arg;
    writev_info->callback = callback_func;
    writev_info->error_callback = error_callback_func;

    nexusl_fd_mutex_lock();
    nexusl_fd_write_register(fd,
			     writev_callback,
			     (void *) writev_info);
    nexusl_fd_mutex_unlock();

    nexus_debug_printf(3, ("nexus_fd_register_for_writev(): exiting\n"));
    return(0);
} /* nexus_fd_register_for_writev() */


/*
 * writev_callback()
 *
 * Callback to do work for nexus_fd_register_for_writev()
 */
static void writev_callback(void *arg, int fd)
{
    writev_info_t *writev_info = (writev_info_t *) arg;
    long n_wrote;
    int save_errno;
    nexus_bool_t done;

    for (done = NEXUS_FALSE; !done; )
    {
#ifndef USE_THREAD_SAFE_FD_CALLS
	nexusl_fd_mutex_lock();
#endif
	
	n_wrote = writev(fd,
			 writev_info->iov,
			 writev_info->iovcnt);
	save_errno = errno;
	
#ifndef USE_THREAD_SAFE_FD_CALLS
	nexusl_fd_mutex_unlock();
#endif
	
	/*
	 * n_wrote: is > 0 on success -- number of bytes written
	 *          is < 0 on error -- need to check errno
	 *          is 0 (SysV) or (-1 && errno==EWOULDBLOCK) (BSD)
	 *              if the write would block without writing anything
	 */

	if (n_wrote > 0)
	{
	    writev_info->nbytes_wrote += n_wrote;
	    while (n_wrote > 0)
	    {
		if (n_wrote >= writev_info->iov->iov_len)
		{
		    n_wrote -= writev_info->iov->iov_len;
		    writev_info->iov++;
		    writev_info->iovcnt--;
		}
		else
		{
		    writev_info->iov->iov_base =
			(void *) (((nexus_byte_t *) writev_info->iov->iov_base)
			          + n_wrote);
		    writev_info->iov->iov_len -= n_wrote;
		    n_wrote = 0;
		}
	    }
	    if (writev_info->iovcnt <= 0)
	    {
		(*writev_info->callback)(writev_info->arg,
					 fd,
					 writev_info->orig_iov,
					 writev_info->orig_iovcnt,
					 writev_info->nbytes_wrote);
		NexusFree(writev_info);
		done = NEXUS_TRUE;
	    }
	}
	else if (   (n_wrote == 0)
		 || (n_wrote < 0 && save_errno == EWOULDBLOCK))
	{
	    nexusl_fd_mutex_lock();
	    nexusl_fd_write_register(fd,
				     writev_callback,
				     (void *) writev_info);
	    nexusl_fd_mutex_unlock();
	    done = NEXUS_TRUE;
	}
	else /* n_wrote < 0 */
	{
	    if (save_errno == EINTR)
	    {
		/* Try again */
	    }
	    else
	    {
		(*writev_info->error_callback)(writev_info->arg,
					       fd,
					       writev_info->orig_iov,
					       writev_info->orig_iovcnt,
					       writev_info->nbytes_wrote,
					       save_errno);
		NexusFree(writev_info);
		done = NEXUS_TRUE;
	    }
	}
    }
} /* writev_callback() */


/*
 * nexus_fd_unregister()
 */
int nexus_fd_unregister(int fd,
			void callback_func(void *arg, int fd))
{
    select_info_t *select_info;
    nexus_debug_printf(3, ("nexus_fd_unregister(): entering\n"));
    nexusl_fd_mutex_lock();
    select_info = fd_table[fd];
    if (select_info)
    {
	if (callback_func)
	{
	    /* TODO: Do something better with callback_func() */
	    if (select_info->read_arg)
	    {
		(*callback_func)(select_info->read_arg, fd);
	    }
	    if (select_info->write_arg)
	    {
		(*callback_func)(select_info->write_arg, fd);
	    }
	    if (select_info->except_arg)
	    {
		(*callback_func)(select_info->except_arg, fd);
	    }
	}

	if (nexusl_fd_read_isregistered(fd))
	{
	    nexusl_fd_read_unregister(fd);
	}

	if (nexusl_fd_write_isregistered(fd))
	{
	    nexusl_fd_write_unregister(fd);
	}

	if (nexusl_fd_except_isregistered(fd))
	{
	    nexusl_fd_except_unregister(fd);
	}
    }
    nexusl_fd_mutex_unlock();
    nexus_debug_printf(3, ("nexus_fd_unregister(): exiting\n"));
    return(0);
} /* nexus_fd_unregister() */



/******************************************************************************
Function:	nexus_fd_handle_events()

Description:	Perform a select() on the registered fds and call the proper
		callbacks.

Parameters:	poll_type	NEXUS_FD_POLL_BLOCKING_ALL
				NEXUS_FD_POLL_BLOCKING_ONCE
				NEXUS_FD_POLL_BLOCKING_UNTIL_SHUTDOWN
				NEXUS_FD_POLL_NONBLOCKING_ALL
				NEXUS_FD_POLL_NONBLOCKING_ONCE

		msg_handled	pointer to boolean variable which will be set
				if one or messages are processed, and cleared
				otherwise

Returns:	error code (XXX: currently not set)
******************************************************************************/

int nexus_fd_handle_events(int poll_type,
			   int *message_handled)
{
    fd_set				read_set;
    fd_set				write_set;
    fd_set				except_set;
    struct timeval			timeout;
    nexus_bool_t			done;
    nexus_bool_t			use_timeout;
    int					n_ready;
    int					n_checked;
    int					fd;
    int					select_errno;
    select_info_t *			select_info;
    int					rc = 0;
    void *				arg;
    nexusl_fd_cbf_t			callback;

    nexus_debug_printf(5,
		       ("nexus_fd_handle_events(): entering, poll_type=%i\n",
			poll_type));

    nexusl_fd_mutex_lock();

#   if !defined(BUILD_LITE)
    {
	if (nexusl_fd_using_handler_thread && !I_Am_Handler_Thread())
	{
	    nexusl_fd_mutex_unlock();
	    if (message_handled)
	    {
		*message_handled = 0;
	    }

	    return(0);
	}

	if (!nexusl_fd_using_handler_thread
	    && nexusl_fd_num_handling_events > 0
	    && (poll_type == NEXUS_FD_POLL_BLOCKING_ALL
		|| poll_type == NEXUS_FD_POLL_BLOCKING_ONCE
		|| poll_type == NEXUS_FD_POLL_BLOCKING_UNTIL_SHUTDOWN))
	{
	    nexusl_fd_mutex_unlock();
	    if (message_handled)
	    {
		*message_handled = 0;
	    }

	    return(0);
	}

	nexusl_fd_num_handling_events++;
    }
#   endif /* !BUILD_LITE */

    done = NEXUS_FALSE;
    while (!done && !nexusl_fd_shutdown_called)
    {
	if (   (nexusl_fd_num_set
		<= 0)
	    && (poll_type != NEXUS_FD_POLL_BLOCKING_UNTIL_SHUTDOWN) )
	{
	    done = NEXUS_TRUE;
	    continue;
	}

	read_set = read_fds;
	write_set = write_fds;
	except_set = except_fds;

	switch (poll_type)
	{
	  case NEXUS_FD_POLL_BLOCKING_ALL:
#           if defined(HAVE_THREAD_SAFE_SELECT)
	    {
	        use_timeout = NEXUS_FALSE;
		poll_type = NEXUS_FD_POLL_NONBLOCKING_ALL;
	    }
#	    else
	    {
		use_timeout = NEXUS_TRUE;
		timeout.tv_sec = 0L;
		timeout.tv_usec = 0L;
	    }
#           endif
	    break;

	  case NEXUS_FD_POLL_BLOCKING_ONCE:
	  case NEXUS_FD_POLL_BLOCKING_UNTIL_SHUTDOWN:
#           if defined(HAVE_THREAD_SAFE_SELECT)
	    {
	        use_timeout = NEXUS_FALSE;
	    }
#	    else
	    {
		use_timeout = NEXUS_TRUE;
		timeout.tv_sec = 0L;
		timeout.tv_usec = 0L;
	    }
#           endif
	    break;

	  case NEXUS_FD_POLL_NONBLOCKING_ALL:
	  case NEXUS_FD_POLL_NONBLOCKING_ONCE:
	    use_timeout = NEXUS_TRUE;
	    timeout.tv_sec = 0L;
	    timeout.tv_usec = 0L;
	    break;
	}


	/*
	 * If we were using a timeout on select() then we must release the
	 * FD mutex so that other threads can do some work
	 */
	if (!use_timeout)
	{
	    nexusl_fd_select_active = NEXUS_TRUE;
	    nexusl_fd_mutex_unlock();
	}

	nexus_debug_printf(1,("nexus_fd_handle_events(): Calling select()\n"));
	n_ready = select(fd_tablesize,
			 NEXUS_FD_SET_CAST &read_set,
			 NEXUS_FD_SET_CAST &write_set,
			 NEXUS_FD_SET_CAST &except_set,
			 (use_timeout ? &timeout : NULL));
	select_errno = errno;
	nexus_debug_printf(1,
			   ("nexus_fd_handle_events(): select() returned\n"));

	/*
	 * If we were using a timeout on select() then we must reacquire the
	 * FD mutex and verify that our state hasn't changed
	 */
	if (!use_timeout)
	{
	    nexusl_fd_mutex_lock();
	    nexusl_fd_select_active = NEXUS_FALSE;

	    /*
	     * Increase the select() counter and signal any waiting threads
	     * that the current select has completed.  The select() counter is
	     * used by routines like nexusl_fd_close() to determine when it's
	     * file descriptor has been removed from the select list.
	     */
	    nexusl_fd_select_cnt++;
	    nexusl_fd_cond_signal();
	
	    /*
	     * If the FD table has been modified in a way which invalidates the
	     * information returned by select(), then reset the modification
	     * flag and retry the select()
	     */
	    if (nexusl_fd_table_modified)
	    {
		nexusl_fd_table_modified = NEXUS_FALSE;
		continue;
	    }

	    /*
	     * If shutdown has been called, then just bail out, ignoring any
	     * pending communication requests
	     */
	    if (nexusl_fd_shutdown_called)
	    {
		break;
	    }
	}

	if (n_ready < 0)
	{
	    if (select_errno == EINTR)
	    {
		continue;
	    }
	    else
	    {
		rc = -1;
		goto handle_abort;
	    }
	}
	else if (n_ready > 0)
	{
	    if (message_handled)
	    {
		*message_handled = 1;
	    }
    
	    for (n_checked = 0, fd = 0;
		 n_checked < n_ready;
		 fd++)
	    {
		if (FD_ISSET(fd, &read_set))
		{
		    select_info = fd_table[fd];
		    callback = select_info->read_callback;
		    arg = select_info->read_arg;
		    nexus_debug_printf(5,
				       ("nexus_fd_handle_events(): "
					"read, fd=%d\n", fd));
		    /* TODO: fix unregister so that it stops setting the table
		       modified flag when called from here */
		    nexusl_fd_read_unregister(fd);

		    nexusl_fd_mutex_unlock();
		    (*callback)(arg, fd);
		    nexusl_fd_mutex_lock();

		    n_checked++;
		    if (nexusl_fd_table_modified)
		    {
			break;
		    }
		}
		if (FD_ISSET(fd, &write_set))
		{
		    nexus_debug_printf(5,
				       ("nexus_fd_handle_events(): "
					"write, fd=%d\n", fd));
		    select_info = fd_table[fd];
		    callback = select_info->write_callback;
		    arg = select_info->write_arg;
		    /* TODO: fix unregister so that it stops setting the table
		       modified flag when called from here */
		    nexusl_fd_write_unregister(fd);

		    nexusl_fd_mutex_unlock();
		    (*callback)(arg, fd);
		    nexusl_fd_mutex_lock();

		    n_checked++;
		    if (nexusl_fd_table_modified)
		    {
			break;
		    }
		}
		if (FD_ISSET(fd, &except_set))
		{
		    nexus_debug_printf(5,
				       ("nexus_fd_handle_events(): "
					"except, fd=%d\n", fd));
		    select_info = fd_table[fd];
		    callback = select_info->except_callback;
		    arg = select_info->except_arg;
		    /* TODO: fix unregister so that it stops setting the table
		       modified flag when called from here */
		    nexusl_fd_except_unregister(fd);

		    nexusl_fd_mutex_unlock();
		    (*callback)(arg, fd);
		    nexusl_fd_mutex_lock();

		    n_checked++;
		    if (nexusl_fd_table_modified)
		    {
			break;
		    }
		}
	    } /* for */
	} /* endif */

	if (   poll_type == NEXUS_FD_POLL_NONBLOCKING_ONCE
	    || poll_type == NEXUS_FD_POLL_BLOCKING_ONCE
	    || (   (n_ready == 0)
		&& (   (poll_type == NEXUS_FD_POLL_NONBLOCKING_ALL)
		    || (poll_type == NEXUS_FD_POLL_BLOCKING_ALL) ) ) )
	{
	    done = NEXUS_TRUE;
	}
    }

handle_abort:
#   ifndef BUILD_LITE
    {
	nexusl_fd_num_handling_events--;
    }
#   endif /* BUILD_LITE */
    nexusl_fd_mutex_unlock();
    nexus_debug_printf(5, ("nexus_fd_handle_events(): exiting\n"));
    return(0);
} /* nexus_fd_handle_events */


/*
 * nexusl_fd_listener_accept()
 *
 *  a fd that we are listening to has someone trying to connect.  The first
 *  argument points to the listener information.  After accepting the
 *  connection, call the function registered by the listener with its
 *  argument.
 */
static void nexusl_fd_listener_accept(void *arg, int fd)
{
    listener_info_t *listener = (listener_info_t *) arg;
    struct sockaddr addr;
    int addrlen;
    int new_fd;
    int rc;
    int save_errno;
    void (*user_callback)(void *, int);
    void *user_arg;

    nexus_debug_printf(1, ("nexusl_fd_listener_accept(): enter, fd=%d, "
			   "arg=0x%08x\n", fd, (unsigned) arg));

    nexusl_fd_mutex_lock();

    addrlen = sizeof(struct sockaddr);

    new_fd = accept(fd, &addr, &addrlen);

    if (new_fd < 0)
    {
	/* TODO: We should probably deal with this error better */
	nexus_debug_printf(2, ("nexusl_fd_listener_accept(): "
			       "accept() failed\n"));
	return;
    }
    if ((rc = nexusl_fd_setup_nonblocking(new_fd)) != 0)
    {
	/* TODO: We should probably deal with this error better */
	nexus_debug_printf(2, ("nexusl_fd_listener_accept(): "
			       "nexusl_fd_setup_nonblocking() failed\n"));

	close(new_fd);
	return;
    }

    if ((rc = nexusl_fd_setup_socket(new_fd)) != 0)
    {
	/* TODO: We should probably deal with this error better */
	nexus_debug_printf(2, ("nexusl_fd_listener_accept(): "
			       "nexusl_fd_setup_socket() failed\n"));
	close(new_fd);
	return;
    }

    user_callback = listener->callback;
    user_arg = listener->arg;
    
    nexusl_fd_read_register(listener->fd,
			    nexusl_fd_listener_accept,
			    (void *) listener);

    nexusl_fd_mutex_unlock();

    nexus_debug_printf(3, ("nexusl_fd_listener_accept(): "
			   "calling user handler, fd=%d, arg=%d\n",
			   new_fd, (unsigned) user_arg));
    (*user_callback)(user_arg, new_fd);
    nexus_debug_printf(3, ("nexusl_fd_listener_accept(): "
			   "returning from user handler\n"));

    nexus_debug_printf(1, ("nexusl_fd_listener_accept(): exit\n"));
}
/* nexusl_fd_listener_accept() */
    

/*
 * nexus_fd_create_listener()
 *
 * Create a listener on port that calls callback_func() when someone
 * tries to connect.
 */
int nexus_fd_create_listener(unsigned short *port,
			     int backlog,
			     void (*callback_func)(void *arg,
						   int fd),
			     void *callback_arg)
{
    struct sockaddr_in my_addr;
    int listen_fd;
    int len;
    int save_errno;
    listener_info_t *listener;

    nexus_debug_printf(3, ("nexus_fd_create_listener(): entering\n"));

    len = sizeof(my_addr);
    memset(&my_addr, '\0', len);

    my_addr.sin_addr.s_addr = INADDR_ANY;
    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons(*port);

    nexusl_fd_mutex_lock();
    
    listen_fd = socket(AF_INET, SOCK_STREAM, 0);
    if (listen_fd < 0)
    {
	save_errno = errno;
	nexusl_fd_mutex_unlock();
	return(save_errno);
    }

#if defined(SO_REUSEADDR) && !defined(TARGET_ARCH_SUNOS41) && !defined(TARGET_ARCH_NEXT040)
    /*
     * Set the port so it can be reused immediately if this
     * process dies.
     *
     * Under SunOS4.1 and NeXTStep, there apparently is a bug in this.
     * With this option set, a bind() will succeed on a port even if
     * that port is still in active use by another process.
     */
    if (*port != 0)
    {
	int one = 1;

	if (setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one)) < 0)
	{
	    save_errno = errno;
	    nexus_warning("setup_listener(): setsockopt SO_REUSEADDR failed: %s\n", _nx_md_system_error_string(save_errno));
	}
    }
#endif /* SO_REUSEADDR */
    
    if (bind(listen_fd, (struct sockaddr *)&my_addr, len) < 0)
    {
	save_errno = errno;
	close(listen_fd);
	nexusl_fd_mutex_unlock();
        return(save_errno);
    }

    if (listen(listen_fd, (backlog < 0 ? SOMAXCONN : backlog)) < 0)
    {
	save_errno = errno;
	close(listen_fd);
	nexusl_fd_mutex_unlock();
	return(save_errno);
    }

    if (getsockname(listen_fd, (struct sockaddr *)&my_addr, &len) < 0)
    {
	save_errno = errno;
	close(listen_fd);
	nexusl_fd_mutex_unlock();
	return(save_errno);
    }

    *port = ntohs(my_addr.sin_port);

    NexusMalloc(nexus_fd_create_listener(),
		listener,
		listener_info_t *,
		sizeof(listener_info_t));
    listener->port = *port;
    listener->fd = listen_fd;
    listener->callback = callback_func;
    listener->arg = callback_arg;
    listener->next = NULL;

    if (listener_head)
    {
	listener->prev = listener_tail;
	listener_tail->next = listener;
	listener_tail = listener;
    }
    else
    {
	listener_head = listener_tail = listener;
	listener->prev = NULL;
    }

    nexusl_fd_setup_nonblocking(listen_fd);
    nexusl_fd_read_register(listen_fd,
			    nexusl_fd_listener_accept,
			    (void *) listener);

    nexusl_fd_mutex_unlock();
    nexus_debug_printf(3, ("nexus_fd_create_listener(): exiting\n"));
    return(0);
} /* nexus_fd_create_listener() */


/*
 * nexus_fd_close_listener()
 *
 *  stop listening at the passed port
 */
int nexus_fd_close_listener(unsigned short port)
{
    listener_info_t *cur_listener;
    listener_info_t *prev_listener;
    int rc;

    nexus_debug_printf(3, ("nexus_fd_close_listener(): entering\n"));
    nexusl_fd_mutex_lock();

    cur_listener = listener_head;
    while(cur_listener && cur_listener->port != port)
    {
	cur_listener = cur_listener->next;
    }
    
    if (cur_listener)
    {
	rc = nexusl_fd_listener_close(cur_listener);
    }
    else
    {
	rc = -1;
    }
    
    nexusl_fd_mutex_unlock();
    nexus_debug_printf(3, ("nexus_fd_close_listener(): exiting, returning %i\n", rc));
    return(rc);
} /* nexus_fd_close_listener() */


/*
 * nexus_fd_connect()
 *
 *  connect to host/port and return the new descriptor in *fd
 */
int nexus_fd_connect(char *host, unsigned short port, int *fd)
{
    struct sockaddr_in his_addr, use_his_addr;
    nexus_bool_t connect_succeeded;
    struct hostent *hp;
    int save_errno;
#ifdef HAVE_GETHOSTBYNAME_R_5
    struct hostent hp2;
    char hp_tsdbuffer[500];
    int hp_errnop;
#elif HAVE_GETHOSTBYNAME_R_3
    struct hostent hp2;
    struct hostent_data hp_data;
#endif
    
    nexus_debug_printf(3, ("nexus_fd_connect(): entering\n"));

#ifdef HAVE_GETHOSTBYNAME_R_5
    hp = gethostbyname_r(host, &hp2, hp_tsdbuffer, 500, &hp_errnop);
#elif defined(HAVE_GET_HOSTBYNAME_R_3)
    rc = gethostbyname_r(host, &hp2, &hp_data);
    if (rc == 0)
    {
	hp = &hp2;
    }
    else
    {
	hp = NULL;
    }
#else
    /*
     * On SunOS 4.1.X, gethostbyname() calls sprintf() :(
     */
    nexus_stdio_lock();
    hp = gethostbyname(host);
    nexus_stdio_unlock();
#endif

    memset(&his_addr, '\0', sizeof(his_addr));
    his_addr.sin_port = htons(port);

    if (hp == NULL)
    {
	/*
	 * gethostbyname() on many machines does the right thing for IP
	 * addresses (e.g., "140.221.7.13").  But on some machines
	 * (e.g., SunOS 4.1.x) it doesn't.  So hack it in this case.
	 */
	if (isdigit(host[0]))
	{
	    unsigned long i1, i2, i3, i4;
	    int rc;

	    nexus_stdio_lock();
	    rc = sscanf(host, "%lu.%lu.%lu.%lu", &i1, &i2, &i3, &i4);
	    nexus_stdio_unlock();

	    if (rc != 4)
	    {
		return(-1);
	    }

	    his_addr.sin_addr.s_addr = (  (i1 << 24)
					  | (i2 << 16)
					  | (i3 << 8)
					  | i4);
	    his_addr.sin_family = AF_INET;
	}
	else
	{
	    return(-1);
	}
    }
    else
    {
	memcpy(&his_addr.sin_addr, hp->h_addr, hp->h_length);
	his_addr.sin_family = hp->h_addrtype;
    }

#ifndef USE_THREAD_SAFE_FD_CALLS
    nexusl_fd_mutex_lock();
#endif
    
    connect_succeeded = NEXUS_FALSE;
    while (!connect_succeeded)
    {
	if ((*fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	    save_errno = errno;
#ifndef USE_THREAD_SAFE_FD_CALLS
	    nexusl_fd_mutex_unlock();
#endif
	    return(-save_errno);
	}

	use_his_addr = his_addr;

	if (connect(*fd, (struct sockaddr *)&use_his_addr,
		    sizeof(use_his_addr)) == 0)
	{
	    connect_succeeded = NEXUS_TRUE;
	}
	else
	{
	    save_errno = errno;
	    if (save_errno == EINPROGRESS)
	    {
		/*
		 * man connect: EINPROGRESS:
		 *   The socket is non-blocking and the connection cannot be
		 *   completed immediately.  It is possible to select(2) for
		 *   completion by selecting the socket for writing.
		 * So this connect has, for all practical purposes, succeeded.
		 */
		connect_succeeded = NEXUS_TRUE;
	    }
	    else if (save_errno == EINTR)
	    {
#ifdef TARGET_ARCH_SUNOS41
		/*
		 * With SunOS 4.1.3 (at least when used
		 * with FSU pthreads), connect() has
		 * an annoying little bug.  It may return EINTR, but
		 * still succeed.  When using round robin scheduling,
		 * where SIGALRM is used to timeslice between threads,
		 * this bug actually occurs with surprising frequency.
		 *
		 * My first attempted solution was to call getpeername(),
		 * with the idea that if it succeeds then the connect()
		 * succeeded, and if it failed then the connect() failed.
		 * This didn't work; getpeername() would sometime fail
		 * yet the connect() would still succeed.
		 *
		 * But it appears that selecting the file descriptor
		 * for writing causes it to wait until things are sane.
		 * (This idea comes from EINPROGRESS.)  And it also
		 * appears that when connect() returns EINTR it
		 * nonetheless always succeeds.
		 *
		 * Note that sometimes the select() returns EINTR,
		 * so we need to retry the select() in that case.
		 * And the EBADF check is just for good measure, as
		 * this would seem to catch the (as yet unobserved)
		 * case where the interrupted connect() actually
		 * failed.
		 *
		 * -Steve Tuecke
		 */
		fd_set my_fd_set;
		int select_rc;
		int select_loop_done = NEXUS_FALSE;
		while (!select_loop_done)
		{
		    FD_ZERO(&my_fd_set);
		    FD_SET(*fd, &my_fd_set);
		    select_rc = select((*fd)+1,
				       NEXUS_FD_SET_CAST NULL,
				       NEXUS_FD_SET_CAST &my_fd_set,
				       NEXUS_FD_SET_CAST NULL,
				       NULL);
		    if (select_rc == 1)
		    {
			select_loop_done = NEXUS_TRUE;
			connect_succeeded = NEXUS_TRUE;
		    }
		    else
		    {
			save_errno = errno;
			if (save_errno == EBADF)
			{
			    close(*fd);
			    select_loop_done = NEXUS_TRUE;
			}
			else if (save_errno != EINTR)
			{
			    /*
			     * connect() returned EINTR,
			     * but selecting on the socket failed.
			     * Guess its time to try another hack
			     * around this connect() bug.
			     */
#ifndef USE_THREAD_SAFE_FD_CALLS
			    nexusl_fd_mutex_unlock();
#endif
			    return(-1);
			}
		    }
		}
#else /* TARGET_ARCH_SUNOS41 */
		/*
		 * Do nothing.  Just try again.
		 */
		close(*fd);
#endif
	    }
	    else if (save_errno == ETIMEDOUT)
	    {
		close(*fd);
#ifndef BUILD_LITE
		/*
		 * Might as well give other threads a chance to run before
		 * trying again.
		 */
	        nexus_thread_yield();
#endif
	    }
	    else
	    {
		close(*fd);
#ifndef USE_THREAD_SAFE_FD_CALLS
		nexusl_fd_mutex_unlock();
#endif
		return(-2);
	    }
	}
    }

#ifndef USE_THREAD_SAFE_FD_CALLS
    nexusl_fd_mutex_unlock();
#endif

    if (nexus_fd_setup_nonblocking(*fd) != 0)
    {
	close(*fd);
	return(-3);
    }
    if (nexus_fd_setup_socket(*fd) != 0)
    {
	close(*fd);
	return(-4);
    }

    nexus_debug_printf(3, ("nexus_fd_connect(): exiting\n"));
    return(0);
} /* nexus_fd_connect() */


/*
 * nexus_fd_defer()
 *
 *  let another event driver handle the detection of events.  The two
 *  functions passed here are called whenever a new fd is registered to be
 *  select() or read() from
 */
int nexus_fd_defer(void (*add_fd)(int fd),
		   void (*remove_fd)(int fd))
{
    nexusl_fd_mutex_lock();
    defer_events = NEXUS_TRUE;
    defer_add_func = add_fd;
    defer_remove_func = remove_fd;
    nexusl_fd_mutex_unlock();

    /* TODO: Defer any currently registered fds */
    
    return(0);
} /* nexus_fd_defer() */


/*
 * nexus_fd_defer_callback()
 *
 *  when another event driver detects an event on a fd, we need to call the
 *  proper callbacks
 */
int nexus_fd_defer_callback(int fd)
{
    select_info_t *select_info;

    if (fd < 0)
    {
	/* TODO: Change this so that it calls a select */
	return(-1);
    }

    nexusl_fd_mutex_lock();
    select_info = fd_table[fd];
    nexusl_fd_mutex_unlock();

    if (!select_info)
    {
	return(-1);
    }
    
    if (select_info->read_callback)
    {
	(*select_info->read_callback)(select_info->read_arg, fd);
    }
    if (select_info->write_callback)
    {
	(*select_info->write_callback)(select_info->write_arg, fd);
    }
    if (select_info->except_callback)
    {
	(*select_info->except_callback)(select_info->except_arg, fd);
    }

    return(0);
} /* nexus_fd_defer_callback() */


/*
 * nexusl_fd_mypipe_callback()
 *
 *  This is the function that gets called when the pipe to self gets signaled
 *  during shutdown.  We don't really want to do anything except to guarantee
 *  that select() will wake after nexusl_fd_shutdown_called is set to true.
 */
static void
nexusl_fd_mypipe_callback(void *arg, int fd)
{
    char				buf;
    int					done = 0;

    
    while (!done)
    {
	done = read(fd, &buf, sizeof(buf));
	if (done == -1)
	{
	    const int			errno_save = errno;

	    if (errno_save == EINTR)
	    {
		done = 0;
	    }
	    else if (errno_save != EAGAIN && errno_save != EWOULDBLOCK)
	    {
		/* XXX: badness has happened; do something about it */
	    }
	}
    }

    nexusl_fd_mutex_lock();

    nexusl_fd_select_wakeup_pending = NEXUS_FALSE;
    nexusl_fd_read_register(fd, nexusl_fd_mypipe_callback, NULL);

    nexusl_fd_mutex_unlock();

    return;
} /* nexusl_fd_mypipe_callback() */
