/*
 * initwrap.c
 *
 * This wraps the nexus initialization routine, _nx_init_nexus().
 * Under AIX, this must be statically linked into an application
 * so that libnexus can be dynamically linked with no
 * unresolved symbols.
 */

static char *rcsid = "$Header: /nfs/globus1/src/master/nexus_4.1/initwrap.c,v 1.34 1997/09/02 21:25:30 stepanek Exp $";

#include "internal.h"

#ifdef HAVE_TCP_PROTO
extern void *_nx_pr_tcp_info(void);
#define NEXUS_USE_TCP_PROTO	{"protocols", "tcp", _nx_pr_tcp_info},
#else
#define NEXUS_USE_TCP_PROTO
#endif

#ifdef HAVE_UDP_PROTO
extern void *_nx_pr_udp_info(void);
#define NEXUS_USE_UDP_PROTO	{"protocols", "udp", _nx_pr_udp_info},
#else
#define NEXUS_USE_UDP_PROTO
#endif

#ifdef HAVE_MPL_PROTO
extern void *_nx_pr_mpl_info(void);
#define NEXUS_USE_MPL_PROTO	{"protocols", "mpl", _nx_pr_mpl_info},
#else
#define NEXUS_USE_MPL_PROTO
#endif

#ifdef HAVE_MPI_PROTO
extern void *_nx_pr_mpi_info(void);
#define NEXUS_USE_MPI_PROTO	{"protocols", "mpi", _nx_pr_mpi_info},
#else
#define NEXUS_USE_MPI_PROTO
#endif

#ifdef HAVE_INX_PROTO
extern void *_nx_pr_inx_info(void);
#define NEXUS_USE_INX_PROTO	{"protocols", "inx", _nx_pr_inx_info},
#else
#define NEXUS_USE_INX_PROTO
#endif

#ifdef HAVE_SHM_PROTO
extern void *_nx_pr_shm_info(void);
#define NEXUS_USE_SHM_PROTO	{"protocols", "shm", _nx_pr_shm_info},
#else
#define NEXUS_USE_SHM_PROTO
#endif

#ifdef HAVE_SS_STARTUP
extern void *_nx_st_ss_info(void);
#define NEXUS_USE_SS_STARTUP	{"startups", "ss", _nx_st_ss_info},
#else
#define NEXUS_USE_SS_STARTUP
#endif

#ifdef HAVE_RSH_STARTUP
extern void *_nx_st_rsh_info(void);
#define NEXUS_USE_RSH_STARTUP	{"startups", "rsh", _nx_st_rsh_info},
#else
#define NEXUS_USE_RSH_STARTUP
#endif

#ifdef HAVE_FORK_STARTUP
extern void *_nx_st_fork_info(void);
#define NEXUS_USE_FORK_STARTUP	{"startups", "fork", _nx_st_fork_info},
#else
#define NEXUS_USE_FORK_STARTUP
#endif

#ifdef HAVE_MPL_STARTUP
extern void *_nx_st_mpl_info(void);
#define NEXUS_USE_MPL_STARTUP	{"startups", "mpl", _nx_st_mpl_info},
#else
#define NEXUS_USE_MPL_STARTUP
#endif

#ifdef HAVE_MPI_STARTUP
extern void *_nx_st_mpi_info(void);
#define NEXUS_USE_MPI_STARTUP	{"startups", "mpi", _nx_st_mpi_info},
#else
#define NEXUS_USE_MPI_STARTUP
#endif

#ifdef HAVE_INX_STARTUP
extern void *_nx_st_inx_info(void);
#define NEXUS_USE_INX_STARTUP	{"startups", "inx", _nx_st_inx_info},
#else
#define NEXUS_USE_INX_STARTUP
#endif

#ifdef HAVE_IBMDS_STARTUP
extern void *_nx_st_ibmds_info(void);
#define NEXUS_USE_IBMDS_STARTUP	{"startups", "ibmds", _nx_st_ibmds_info},
#else
#define NEXUS_USE_IBMDS_STARTUP
#endif

#ifdef HAVE_SOLDL_STARTUP
extern void *_nx_st_soldl_info(void);
#define NEXUS_USE_SOLDL_STARTUP	{"startups", "soldl", _nx_st_soldl_info},
#else
#define NEXUS_USE_SOLDL_STARTUP
#endif


#if (defined HAVE_QOS) && (defined BUILD_RESOURCE)
extern void *nexus_overseer_qos_info(void);
#define NEXUS_USE_QOS_OVERSEER {"overseers", "qos", nexus_overseer_qos_info},
#else
#define NEXUS_USE_QOS_OVERSEER
#endif


extern void *_nx_rdb_file_info(void);


static nexus_module_list_t default_module_list[] =
{
    NEXUS_USE_SOLDL_STARTUP
    NEXUS_USE_IBMDS_STARTUP
    NEXUS_USE_MPL_STARTUP
    NEXUS_USE_MPI_STARTUP
    NEXUS_USE_INX_STARTUP
    NEXUS_USE_FORK_STARTUP
    NEXUS_USE_SS_STARTUP
    NEXUS_USE_RSH_STARTUP
    NEXUS_USE_SHM_PROTO
    NEXUS_USE_MPL_PROTO
    NEXUS_USE_MPI_PROTO
    NEXUS_USE_INX_PROTO
    NEXUS_USE_TCP_PROTO
    NEXUS_USE_QOS_OVERSEER
    NEXUS_USE_UDP_PROTO
    {"rdb", "file", _nx_rdb_file_info},
    {NULL, NULL, NULL},
};


/*
 * nexus_init()
 */
void nexus_init(int *argc,
		char ***argv,
		char *args_env_variable,
		char *package_designator,
		int (*package_args_init_func)(int *argc, char ***argv),
		void (*usage_message_func)(void),
		int (*new_process_params_func)(char *buf, int size),
		nexus_module_list_t module_list[],
		nexus_node_t **nodes,
		int *n_nodes)
{
    _nx_NexusBoot = NexusBoot;
    _nx_NexusExit = NexusExit;
    _nx_NexusAcquiredAsNode = NexusAcquiredAsNode;
    _nx_init_nexus(argc,
		   argv,
		   args_env_variable,
		   package_designator,
		   package_args_init_func,
		   usage_message_func,
		   new_process_params_func,
		   (module_list ? module_list : default_module_list),
		   nodes,
		   n_nodes);
} /* nexus_init() */
