#include "mn_Malloc.h"

/*
 *	Project:	High Speed Network ASR
 *	Organization:	The Aerospace Corporation
 *	Programmer:	Robert Lindell
 *	File:		$RCSfile: mn_Malloc.c,v $
 *	Date:		$Date: 1996/10/07 04:43:10 $
 *	Version:	$Revision: 1.3 $
 *
 *      Modified by Cheryl DeMatteis, converted to 'C' code
 */

static const char rcsid[] = "$Id: mn_Malloc.c,v 1.3 1996/10/07 04:43:10 tuecke Exp $";

void construct_mn_Malloc( Malloc** M, caddr_t buffer, unsigned size )
{
   int i;

/*   *M =  (Malloc*) malloc( sizeof( Malloc ));*/
   (*M)->buf = buffer;
   (*M)->len = size;
   (*M)->top = (*M)->buf;
   (*M)->realloc_srchlen = 4;	/* 4 should be plenty, -1 =>'s whole list */
   (*M)->pagesz = 0;
   (*M)->pagebucket = 0;
   for (i = 0;i < NBUCKETS; i++)
      (*M)->nextf[i] = 0;
}

void destruct_mn_Malloc( Malloc* M )
{
   free( M );
}

caddr_t mn_sbrk( Malloc* M, int x )
{
   caddr_t oldtop;

	if (x < 0)
		return((caddr_t) -1);
	if (x == 0)
		return( M->top );
	if (x > (  M->len - ( M->top - M->buf )))
		return((caddr_t) -1);
	oldtop = M->top;
	M->top += x;
	return (oldtop);
}


caddr_t get_mn_buffer( Malloc* M ) 
{
	return( M->buf );
}

void set_mn_buffer( Malloc* M, caddr_t x )
{
	M->buf = x;
	M->top = M->buf;
}

unsigned mn_get_length( Malloc* M )
{
	return( M->len );
}

void mn_set_length( Malloc* M, unsigned x )
{
	M->len = x;
}


/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that: (1) source distributions retain this entire copyright
 * notice and comment, and (2) distributions including binaries display
 * the following acknowledgement:  ``This product includes software
 * developed by the University of California, Berkeley and its contributors''
 * in the documentation or other materials provided with the distribution
 * and in all advertising materials mentioning features or use of this
 * software. Neither the name of the University nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)malloc.c	5.9 (Berkeley) 6/1/90";
#endif /* LIBC_SCCS and not lint */

/*
 * malloc.c (Caltech) 2/21/82
 * Chris Kingsley, kingsley@cit-20.
 *
 * This is a very fast storage allocator.  It allocates blocks of a small 
 * number of different sizes, and keeps free lists of each size.  Blocks that
 * don't exactly fit are passed up to the next larger size.  In this 
 * implementation, the available sizes are 2^n-4 (or 2^n-10) bytes long.
 * This is designed for use in a virtual memory environment.
 */

#include <sys/types.h>

#ifndef	NULL
#define	NULL 0
#endif	/* NULL */

/*
 * The overhead on a block is at least 4 bytes.  When free, this space
 * contains a pointer to the next free block, and the bottom two bits must
 * be zero.  When in use, the first byte is set to MAGIC, and the second
 * byte is the size index.  The remaining bytes are for alignment.
 * If range checking is enabled then a second word holds the size of the
 * requested block, less 1, rounded up to a multiple of sizeof(RMAGIC).
 * The order of elements is critical: ov_magic must overlay the low order
 * bits of ov_next, and ov_magic can not be a valid ov_next bit pattern.
 */
union	overhead {
	union	overhead *ov_next;	/* when free */
	struct {
		u_char	ovu_magic;	/* magic number */
		u_char	ovu_index;	/* bucket # */
#ifdef RCHECK
		u_short	ovu_rmagic;	/* range magic number */
		u_int	ovu_size;	/* actual block size */
#endif
	} ovu;
#define	ov_magic	ovu.ovu_magic
#define	ov_index	ovu.ovu_index
#define	ov_rmagic	ovu.ovu_rmagic
#define	ov_size		ovu.ovu_size
};

#define	MAGIC		0xef		/* magic # on accounting info */
#define RMAGIC		0x5555		/* magic # on range info */

#ifdef RCHECK
#define	RSLOP		sizeof (u_short)
#else
#define	RSLOP		0
#endif

#ifdef MSTATS
/*
 * nmalloc[i] is the difference between the number of mallocs and frees
 * for a given block size.
 */
static	u_int nmalloc[NBUCKETS];
#include <stdio.h>
#endif

#if defined(DEBUG) || defined(RCHECK)
#define	ASSERT(p)   if (!(p)) botch(#p)
#include <stdio.h>
static
botch( char *s)
{
	fprintf(stderr, "\r\nassertion botched: %s\r\n", s);
 	(void) fflush(stderr);		/* just in case user buffered it */
	*((int *)-1) = 0; /* force fault; abort will probably fail if malloc's dead */
	abort();
}
#else
#define	ASSERT(p)
#endif

char* mn_malloc( Malloc* M, unsigned nbytes )
{
  	register union overhead *op;
  	register int bucket, n;
	register unsigned amt;

	/*
	 * First time malloc is called, setup page size and
	 * align break pointer so all data will be page aligned.
	 */
	if ( M->pagesz == 0) {
		M->pagesz = n = getpagesize();
		op = (union overhead *)mn_sbrk( M, 0 );
  		n = n - sizeof (*op) - ((int)op & (n - 1));
		if (n < 0)
			n += M->pagesz;
  		if (n) {
  			if (mn_sbrk( M, n ) == (char *)-1)
				return (NULL);
		}
		bucket = 0;
		amt = 8;
		while ( M->pagesz > amt ) {
			amt <<= 1;
			bucket++;
		}
		M->pagebucket = bucket;
	}
	/*
	 * Convert amount of memory requested into closest block size
	 * stored in hash buckets which satisfies request.
	 * Account for space used per block for accounting.
	 */
	if (nbytes <= (n = M->pagesz - sizeof (*op) - RSLOP)) {
#ifndef RCHECK
		amt = 8;	/* size of first bucket */
		bucket = 0;
#else
		amt = 16;	/* size of first bucket */
		bucket = 1;
#endif
		n = -(sizeof (*op) + RSLOP);
	} else {
		amt = M->pagesz;
		bucket = M->pagebucket;
	}
	while (nbytes > amt + n) {
		amt <<= 1;
		if (amt == 0)
			return (NULL);
		bucket++;
	}
	/*
	 * If nothing in hash bucket right now,
	 * request more memory from the system.
	 */
  	if ((op = M->nextf[bucket]) == NULL) {
  		mn_morecore( M, bucket );
  		if ((op = M->nextf[bucket]) == NULL)
  			return (NULL);
	}
	/* remove from linked list */
  	M->nextf[bucket] = op->ov_next;
	op->ov_magic = MAGIC;
	op->ov_index = bucket;
#ifdef MSTATS
  	nmalloc[bucket]++;
#endif
#ifdef RCHECK
	/*
	 * Record allocated size of block and
	 * bound space with magic numbers.
	 */
	op->ov_size = (nbytes + RSLOP - 1) & ~(RSLOP - 1);
	op->ov_rmagic = RMAGIC;
  	*(u_short *)((caddr_t)(op + 1) + op->ov_size) = RMAGIC;
#endif
  	return ((char *)(op + 1));
}

/* added by DJ Delorie */
char* mn_calloc( Malloc* M, unsigned size, unsigned nelem )
{
  char *rv = mn_malloc( M, size*nelem );
  if (rv)
    memset(rv, 0, size*nelem);
  return rv;
}

void mn_free( Malloc* M, char *cp )
{   
  	register int size;
	register union overhead *op;

  	if (cp == NULL)
  		return;
	op = (union overhead *)((caddr_t)cp - sizeof (union overhead));
#ifdef DEBUG
  	ASSERT(op->ov_magic == MAGIC);		/* make sure it was in use */
#else
	if (op->ov_magic != MAGIC)
		return;				/* sanity */
#endif
#ifdef RCHECK
  	ASSERT(op->ov_rmagic == RMAGIC);
	ASSERT(*(u_short *)((caddr_t)(op + 1) + op->ov_size) == RMAGIC);
#endif
  	size = op->ov_index;
  	ASSERT(size < NBUCKETS);
	op->ov_next = M->nextf[size];	/* also clobbers ov_magic */
  	M->nextf[size] = op;
#ifdef MSTATS
  	nmalloc[size]--;
#endif
}

/*
 * When a program attempts "storage compaction" as mentioned in the
 * old malloc man page, it realloc's an already freed block.  Usually
 * this is the last block it freed; occasionally it might be farther
 * back.  We have to search all the free lists for the block in order
 * to determine its bucket: 1st we make one pass thru the lists
 * checking only the first block in each; if that fails we search
 * ``realloc_srchlen'' blocks in each list for a match (the variable
 * is extern so the caller can modify it).  If that fails we just copy
 * however many bytes was given to realloc() and hope it's not huge.
 */

char* mn_realloc( Malloc* M, char *cp, unsigned nbytes )
{   
  	register u_int onb;
	register int i;
	union overhead *op;
  	char *res;
	int was_alloced = 0;

  	if (cp == NULL)
  		return ( mn_malloc( M, nbytes ));
	op = (union overhead *)((caddr_t)cp - sizeof (union overhead));
	if (op->ov_magic == MAGIC) {
		was_alloced++;
		i = op->ov_index;
	} else {
		/*
		 * Already free, doing "compaction".
		 *
		 * Search for the old block of memory on the
		 * free list.  First, check the most common
		 * case (last element free'd), then (this failing)
		 * the last ``realloc_srchlen'' items free'd.
		 * If all lookups fail, then assume the size of
		 * the memory block being realloc'd is the
		 * largest possible (so that all "nbytes" of new
		 * memory are copied into).  Note that this could cause
		 * a memory fault if the old area was tiny, and the moon
		 * is gibbous.  However, that is very unlikely.
		 */
		if ((i = mn_findbucket( M, op, 1 )) < 0 &&
		    (i = mn_findbucket( M, op, M->realloc_srchlen )) < 0)
			i = NBUCKETS;
	}
	onb = 1 << (i + 3);
	if (onb < M->pagesz)
		onb -= sizeof (*op) + RSLOP;
	else
		onb += M->pagesz - sizeof (*op) - RSLOP;
	/* avoid the copy if same size block */
	if (was_alloced) {
		if (i) {
			i = 1 << (i + 2);
			if (i < M->pagesz)
				i -= sizeof (*op) + RSLOP;
			else
				i += M->pagesz - sizeof (*op) - RSLOP;
		}
		if (nbytes <= onb && nbytes > i) {
#ifdef RCHECK
			op->ov_size = (nbytes + RSLOP - 1) & ~(RSLOP - 1);
			*(u_short *)((caddr_t)(op + 1) + op->ov_size) = RMAGIC;
#endif
			return(cp);
		} else
			mn_free( M, cp );
	}
  	if ((res = mn_malloc( M, nbytes )) == NULL)
  		return (NULL);
  	if (cp != res)		/* common optimization if "compacting" */
		memcpy(res, cp, (nbytes < onb) ? nbytes : onb);
  	return (res);
}

/*
 * Allocate more memory to the indicated bucket.
 */
void mn_morecore( Malloc* M, int bucket )
{
  	register union overhead *op;
	register int sz;		/* size of desired block */
  	int amt;			/* amount to allocate */
  	int nblks;			/* how many blocks we get */

	/*
	 * sbrk_size <= 0 only for big, FLUFFY, requests (about
	 * 2^30 bytes on a VAX, I think) or for a negative arg.
	 */
	sz = 1 << (bucket + 3);
#ifdef DEBUG
	ASSERT(sz > 0);
#else
	if (sz <= 0)
		return;
#endif
	if (sz < M->pagesz ) {
		amt = M->pagesz;
  		nblks = amt / sz;
	} else {
		amt = sz + M->pagesz;
		nblks = 1;
	}
	op = (union overhead *)mn_sbrk( M, amt );
	/* no more room! */
  	if ((int)op == -1)
  		return;
	/*
	 * Add new memory allocated to that on
	 * free list for this hash bucket.
	 */
  	M->nextf[bucket] = op;
  	while (--nblks > 0) {
		op->ov_next = (union overhead *)((caddr_t)op + sz);
		op = (union overhead *)((caddr_t)op + sz);
  	}
	op->ov_next = 0;
}

/*
 * Search ``srchlen'' elements of each free list for a block whose
 * header starts at ``freep''.  If srchlen is -1 search the whole list.
 * Return bucket number, or -1 if not found.
 */
int mn_findbucket( Malloc* M, union overhead *freep, int srchlen )
{
	register union overhead *p;
	register int i, j;

	for (i = 0; i < NBUCKETS; i++) {
		j = 0;
		for (p = M->nextf[i]; p && j != srchlen; p = p->ov_next) {
			if (p == freep)
				return (i);
			j++;
		}
	}
	return (-1);
}

#ifdef MSTATS
/*
 * mstats - print out statistics about malloc
 * 
 * Prints two lines of numbers, one showing the length of the free list
 * for each size category, the second showing the number of mallocs -
 * frees for each size category.
 */
void mn_mstats( Malloc* M )
{
  	register int i, j;
  	register union overhead *p;
  	int totfree = 0,
  	totused = 0;

  	fprintf(stderr, "Memory allocation statistics\nfree:\t");
  	for (i = 0; i < NBUCKETS; i++) {
  		for (j = 0, p = M->nextf[i]; p; p = p->ov_next, j++)
  			;
  		fprintf(stderr, " %d", j);
  		totfree += j * (1 << (i + 3));
  	}
  	fprintf(stderr, "\nused:\t");
  	for (i = 0; i < NBUCKETS; i++) {
  		fprintf(stderr, " %d", nmalloc[i]);
  		totused += nmalloc[i] * (1 << (i + 3));
  	}
  	fprintf(stderr, "\n\tTotal in use: %d, total free: %d\n",
	    totused, totfree);
}
#endif


