/*
 * mp_mpl.h
 *
 * IBM SP MPL protocol configuration for pr_mp.c
 * 
 * rcsid = "$Header: /nfs/globus1/src/master/nexus_4.1/mp_mpl.h,v 1.31 1997/07/22 18:06:39 ludban Exp $"
 */

#include "mpproto.h"

#define NEXUS_MP_PROTO_MPL
#define NEXUS_PROTO_TYPE_MP NEXUS_PROTO_TYPE_MPL
#define NEXUS_PROTO_NAME_MP "mpl"
#define MP_PROTOCOL_INFO _nx_pr_mpl_info
#define MP_STARTUP_INFO _nx_st_mpl_info

/*
 * MP_PROTO_IS_THREAD_SAFE
 *
 * It is advantageous if an MP module be thread safe, since this
 * allows you to simply create a handler thread which blocks on a
 * receive, and automatically wakes up and handles the message when
 * one arrives.
 *
 * In order for an MP module to be considered "thread safe", the
 * following must be true:
 *   1) One thread can be doing a blocking receive, while other threads
 *	are simultaneously doing sends.
 *   2) A thread can send a message to its own node, which will
 *	be received by the blocking receive in the handler thread.
 *
 * If these conditions are true, then #define MP_PROTO_IS_THREAD_SAFE
 */
/*
#define MP_PROTO_IS_THREAD_SAFE
*/


/*
 * mp_destination_t
 *
 * This encapsulates the information about a node to which we
 * may send messages.
 */
typedef int mp_destination_t;

/* Desribe contents of mp_destination_t for nexus_mi_proto_t manipulation */
#define MI_INT_COUNT 1
#define MI_STR_COUNT 0

/*
 * mp_send_status_t
 * mp_receive_status_t
 */
typedef int mp_send_status_t;
typedef int mp_receive_status_t;

#define MPL_MSG_TYPE 42

/*
 * MP_DEFAULT_STORAGE_SIZE
 *
 * This is the default buffer size that will be used on sends and
 * receives to avoid extraneous mallocs and locks.
 */
#define MP_DEFAULT_STORAGE_SIZE 32736

/*
 * MP_BIG_MESSAGE_MIN_SIZE
 *
 * Any message >= to this size will be sent using a two message
 * protocol.  A first, small message is sent with the size,
 * followed by a second message with the actual data.
 */
#define MP_BIG_MESSAGE_MIN_SIZE 1500

/*
 * MP_INITIALIZE()
 *
 * Do any initialization required before calling any other MP_ functions.
 */
#define MP_INITIALIZE()

/*
 * MP_INIT_NODE_INFO(mp_destination_t Mynode, int N_nodes)
 *
 * Set Mynode to the mp_destionation_t info for my node.
 * Set N_nodes the the number of nodes.
 */
#define MP_INIT_NODE_INFO(My_node, N_nodes) \
{ \
    mpc_environ(&(N_nodes), &(My_node)); \
}

/*
 * MP_IS_MASTER_NODE(mp_destination_t dest, nexus_bool_t result)
 */
#define MP_IS_MASTER_NODE(dest, result) \
{ \
    if ((dest) == 0) \
	result = NEXUS_TRUE; \
    else \
	result = NEXUS_FALSE; \
}

/*
 * MP_SET_DESTINATION(mp_destination_t dest, int node_num)
 */
#define MP_SET_DESTINATION(dest, node_num) \
    (dest) = (node_num)

/*
 * MP_POST_RECEIVE(Func,
 *		   nexus_byte_t *Buf,
 *		   unsigned long Buf_size,
 *		   mp_receive_status_t Status)
 *
 * Post a receive into Buf with a maximum size of Buf_size.
 * MP_RECEIVE_STATUS() and MP_RECEIVE_WAIT() should be used
 * with Status to determine completion of the receive.
 */
static int mpl_source;
static int mpl_type;
#define MP_POST_RECEIVE(Func, Buf, Buf_size, Status) \
{ \
    mpl_source = DONTCARE; \
    mpl_type = MPL_MSG_TYPE; \
    if (mpc_recv((char *) (Buf), \
		 (int) (Buf_size), \
		 &mpl_source, \
		 &mpl_type, \
		 &(Status)) != 0) \
    { \
	nexus_fatal("MP_POST_RECEIVE(): mpc_recv() failed, mperrno=%d\n", mperrno); \
    } \
}


/*
 * MP_RECEIVE_STATUS(Func,
 *                   mp_receive_status_t Status,
 *                   nexus_bool_t Done)
 *
 * See if the receive described by Status has completed.  If so, then
 * set Done=NEXUS_TRUE. else set Done=NEXUS_FALSE.
 */
#define MP_RECEIVE_STATUS(Func, Status, Done) \
{ \
    int mpl_rc; \
    if((mpl_rc = mpc_status(Status)) >= 0) \
    { \
	(Done) = NEXUS_TRUE; \
    } \
    else if(mpl_rc == -1) \
    { \
	(Done) = NEXUS_FALSE; \
    } \
    else \
    { \
	nexus_fatal("MP_RECEIVE_STATUS(): mpc_status() failed, mperrno=%d\n", mperrno); \
    } \
}


/*
 * MP_RECEIVE_WAIT(Func,
 *                 mp_receive_status_t Status,
 *                 nexus_bool_t Done)
 *
 * Wait for the receive described by Status to complete.
 * If it completes successfully, then
 * set Done=NEXUS_TRUE. else set Done=NEXUS_FALSE.
 */
#define MP_RECEIVE_WAIT(Func, Status, Done) \
{ \
    size_t mpl_n_bytes; \
    if(mpc_wait(&(Status), &mpl_n_bytes) == 0) \
    { \
	(Done) = NEXUS_TRUE; \
    } \
    else \
    { \
	nexus_fatal("MP_RECEIVE_WAIT(): mpc_wait() failed, mperrno=%d\n", mperrno); \
    } \
}


/*
 * MP_RECEIVE_CANCEL(mp_receive_status_t Status)
 *
 * Cancel a previously posted receive.
 */
#define MP_RECEIVE_CANCEL(Status)

/*
 * MP_SEND(mp_destination_t Dest,
 *         nexus_byte_t *Buf,
 *         unsigned long Buf_size,
 *         mp_send_status_t Status)
 *
 * Send Buf, of size Buf_size, to Dest.
 * The completion of the send will be checked by calling
 * MP_SEND_STATUS() using the Status argument.
 */
#define MP_SEND(Dest, Buf, Buf_size, Status) \
{ \
    if (mpc_send((char *) (Buf), \
		 (int) (Buf_size), \
		 (Dest), \
		 MPL_MSG_TYPE, \
		 &(Status)) != 0) \
    { \
        nexus_fatal("MP_SEND(): mpc_send() failed, mperrno=%d\n", mperrno); \
    } \
}


/*
 * MP_SEND_STATUS(mp_send_status_t Status, nexus_bool_t Done)
 *
 * See if the send described by Status has completed.
 * If so, set Done=NEXUS_TRUE, else set Done=NEXUS_FALSE.
 */
#define MP_SEND_STATUS(Status, Done) \
{ \
    if (mpc_status(Status) >= 0) \
	(Done) = NEXUS_TRUE; \
    else \
	(Done) = NEXUS_FALSE; \
}


/*
 * MP_COMPARE_DESTINATIONS(mp_destination_t D1, mp_destination_t D2,
 *			  nexus_bool_t Result)
 *
 * Set Result==NEXUS_TRUE if mp_destination_t's D1 and D2 are the same.
 */
#define MP_COMPARE_DESTINATIONS(D1, D2, Result) \
    Result = (((D1) == (D2)) ? NEXUS_TRUE : NEXUS_FALSE);


/*
 * MP_COPY_DESTINATION(mp_destination_t To, mp_destination_t From)
 *
 * Copy mp_destination_t 'From' to 'To'.
 */
#define MP_COPY_DESTINATION(To, From) \
    (To) = (From)


/*
 * MP_FREE_DESTINATION(mp_destination_t Dest)
 *
 * Free up any memory in 'Dest' which was malloced.
 */
#define MP_FREE_DESTINATION(Dest)


/*
 * MP_HASH_DESTINATION(mp_destination_t Dest, int Hash)
 *
 * Hash the destination information into a value
 * between 0 and PROTO_TABLE_SIZE.
 */
#define MP_HASH_DESTINATION(Dest, Hash) \
    (Hash) = ((Dest) % PROTO_TABLE_SIZE)


/*
 * MP_INIT_DESTINATION(mp_destination_t Dest)
 *
 * Initialize the destination.
 */
#define MP_INIT_DESTINATION(Dest)


/*
 * MP_NODE_EXIT()
 *
 * If a terminating process needs to do something to map itself out
 * of the set of processes, this macro should do it.
 */
#define MP_NODE_EXIT()


/*
 * MP_ABORT()
 *
 * Special statements to abort all processes.
 */
#define MP_ABORT() \
{ \
    mpc_stopall(1); \
}


/*
 * MP_GET_MY_MI_PROTO_SIZE(int Size)
 *
 * Fillin Size with the number of bytes I need to store my_node
 * into the mi_proto byte array.
 */
#define MP_GET_MY_MI_PROTO_SIZE(Size) (Size) = sizeof(int)


/*
 * MP_GET_MY_MI_PROTO(nexus_byte_t *Array)
 *
 * Fillin my_node into the mi_proto byte array
 */
#define MP_GET_MY_MI_PROTO(Array) \
{ \
    memcpy((Array), &my_node, sizeof(int)); \
}


/* 
 * MP_CONSTRUCT_FROM_MI_PROTO(mp_destination_t Dest,
 *			      nexus_mi_proto_t *Mi_proto,
 *			      nexus_byte_t *   Array)
 *
 * Copy the needed elements from Array to Dest.
 */
#define MP_CONSTRUCT_FROM_MI_PROTO(Dest, Mi_proto, Array) \
{ \
    memcpy(&(Dest), (Array), sizeof(int)); \
}


/* 
 * MP_BROADCAST_COMMAND(int Command)
 *
 * Send the command out to everybody.
 *
 * If this macro is not defined, then pr_mp.c will use
 * MP_SEND() to send the message to everybody.
 */
#define MP_BROADCAST_COMMAND(Command) \
{ \
    nexus_byte_t mpl_buf[1]; \
    int mpl_i; \
    nexus_bool_t mpl_send_done; \
    int mpl_message_id; \
    mpl_buf[0] = (nexus_byte_t) (Command); \
    for (mpl_i = 0; mpl_i < n_nodes; mpl_i++) \
    { \
	if (mpl_i != my_node) \
	{ \
	    mpc_send((char *) mpl_buf, \
		     1, \
		     mpl_i, \
		     MPL_MSG_TYPE, \
		     &mpl_message_id); \
	    if (mpc_status(mpl_message_id) >= 0) \
		mpl_send_done = NEXUS_TRUE; \
	    else \
		mpl_send_done = NEXUS_FALSE; \
	    while (!mpl_send_done) \
	    { \
		if (!using_handler_thread) \
		{ \
		    receive_messages(NON_BLOCKING); \
		} \
		if (mpc_status(mpl_message_id) >= 0) \
		    mpl_send_done = NEXUS_TRUE; \
		else \
		    mpl_send_done = NEXUS_FALSE; \
	    } \
	} \
    } \
}
