/*
 * nexus_resource.h
 *
 * rcsid = "$Header"
 *
 * Nexus Resources
 */

#ifndef _NEXUS_INCLUDE_NEXUS_RESOURCE_H
#define _NEXUS_INCLUDE_NEXUS_RESOURCE_H

#include <netinet/in.h>

#ifndef BUILD_RESOURCE

typedef enum {
#define NEXUS_RESOURCE_NTYPES	4
        RESOURCE_SOCKET = 0,
        RESOURCE_PROCESS = 1,
        RESOURCE_THREAD = 2,
        RESOURCE_MEMORY = 3
} nexus_resource_type;

typedef struct _nexus_resource_name_t {
	int			id; 
	nexus_resource_type	type;
} nexus_resource_name_t;

#define nexus_resource_macro_sizeof_name(Count) \
	(Count * nexus_sizeof_int(2))

typedef enum {
#define NEXUS_RUSAGE_N_OPS   3
        RUSAGE_CREATE,
        RUSAGE_DESTROY,
        RUSAGE_MODIFY
} nexus_rusage_op;

typedef struct _nexus_rusage_obj {
	nexus_resource_name_t*	who;
        nexus_rusage_op 	op;
} nexus_rusage_obj;

#endif  /* BUILD_RESOURCE */


typedef struct {
        const nexus_resource_name_t	*who;
        nexus_rusage_op 		op;
        int 				pid;
        int 				fd;
        struct sockaddr_in 		addr;
        int 				protocol;
} nexus_rusage_socket;

typedef struct {
        const nexus_resource_name_t	*who;
        nexus_rusage_op 		op;
        int 				pid;
} nexus_rusage_process;

typedef struct {
        const nexus_resource_name_t 	*who;
        nexus_rusage_op 		op;
        int 				pid;
  /* An int tid is sufficient for Solaris and IRIX pthreads but not
   * for DCE/pthreads.  In general, a tid may be an arbitrary data
   * type.  This will work for now.
   */
        int 				tid;
} nexus_rusage_thread;

typedef unsigned long nexus_rusage_memory_t;

typedef struct {
        const nexus_resource_name_t	*who;
        nexus_rusage_op 		op;
        nexus_rusage_memory_t 		*handle;
} nexus_rusage_memory;


typedef union {
	nexus_rusage_socket	socket;
	nexus_rusage_process	process;
	nexus_rusage_thread	thread;
	nexus_rusage_memory	memory;
} nexus_rusage_info;


extern int nexus_resource_name(nexus_resource_name_t*, nexus_resource_type);
extern int nexus_resource_name_cmp(const nexus_resource_name_t*, 
				   const nexus_resource_name_t*); 
extern int nexus_resource_name_cp( nexus_resource_name_t*, const nexus_resource_name_t* ); 

#ifndef USE_MACROS
extern int nexus_resource_sizeof_name(int count);
#else
#define nexus_resource_sizeof_name(count) \
	nexus_resource_macro_sizeof_name(count)
#endif

extern void nexus_resource_name_put(nexus_buffer_t*, nexus_resource_name_t*);
extern void nexus_resource_name_get(nexus_buffer_t*, nexus_resource_name_t*);

extern int nexus_sizeof_rusage(nexus_rusage_obj*, int);
extern void nexus_rusage_put(nexus_buffer_t*, nexus_rusage_obj*);
extern void nexus_rusage_get(nexus_buffer_t*, nexus_rusage_obj*);

#endif /* _NEXUS_INCLUDE_NEXUS_RESOURCE_H */

