/*
 * Nexus
 * Authors:     James Stepanek
 *              Aerospace
 *
 * qoslib.c	- QoSlib module
 */

static char *rcsid = "$Header: /nfs/globus1/src/master/nexus_4.1/qoslib.c,v 1.1 1997/09/02 21:25:41 stepanek Exp $";

#include "internal.h"

#include <sys/socket.h>
#include <netinet/in.h>

#include "nexus_resource.h"

#include "qoslib.h"

static int _sizeof_socket(qos_request_obj*, int);
static int _sizeof_process(qos_request_obj*, int);
static int _sizeof_thread(qos_request_obj*, int);
static int _sizeof_memory(qos_request_obj*, int);

static void _put_socket(nexus_buffer_t*,  qos_request_obj*);
static void _put_process(nexus_buffer_t*,  qos_request_obj*);
static void _put_thread(nexus_buffer_t*,  qos_request_obj*);
static void _put_memory(nexus_buffer_t*,  qos_request_obj*);

static void _get_socket(nexus_buffer_t*,  qos_request_obj*);
static void _get_process(nexus_buffer_t*,  qos_request_obj*);
static void _get_thread(nexus_buffer_t*,  qos_request_obj*);
static void _get_memory(nexus_buffer_t*,  qos_request_obj*);

static struct _request_funcs {
  int (*size)	(qos_request_obj*, int);
  void (*put)	(nexus_buffer_t*, qos_request_obj*);
  void (*get)	(nexus_buffer_t*, qos_request_obj*);
} request_funcs[NEXUS_RESOURCE_NTYPES] = {
  /* size */		/* put */	/* get */
  {_sizeof_socket, 	_put_socket, 	_get_socket},	/* SOCKET */
  {_sizeof_process, 	_put_process, 	_get_process},	/* PROCESS */
  {_sizeof_thread, 	_put_thread, 	_get_thread},	/* THREAD */
  {_sizeof_memory, 	_put_memory, 	_get_memory}	/* MEMORY */
};


/*
 * qos_sizeof_request()
 */
int qos_sizeof_request(qos_request_obj* request_obj, int count) {
  if (request_funcs[request_obj->who->type].size) 
    return ((request_funcs[request_obj->who->type].size)(request_obj, count));
  return (0);
}

/*
 * qos_request_put()
 */
void qos_request_put( nexus_buffer_t *buffer, qos_request_obj *request_obj )
{
  if  ( request_funcs[request_obj->who->type].put )
    (request_funcs[request_obj->who->type].put)( buffer, request_obj );
}

/*
 * qos_request_get()
 */
void qos_request_get( nexus_buffer_t *buffer, qos_request_obj *request_obj )
{
  if  ( request_funcs[request_obj->who->type].get )
    (request_funcs[request_obj->who->type].get)( buffer, request_obj );
}


static int _sizeof_socket(qos_request_obj *request_obj, int count)
{
  return (count * (
		   nexus_sizeof_byte(sizeof(qos_callback *))
#ifdef QOS_USE_RSVP
		   + nexus_sizeof_byte(sizeof(qos_rspec *))
		   + nexus_sizeof_byte(sizeof(qos_tspec))
#endif
		   )
	  );
}

static void _put_socket(nexus_buffer_t *buffer, qos_request_obj *request_obj)
{
  nexus_put_u_long( buffer, (unsigned long*)&(((qos_request_socket *)request_obj)->proc), 1 );
#ifdef QOS_USE_RSVP
  nexus_printf("qoslib: Error: rspec/tspec _put_socket not implemented\n");
#endif
}

static void _get_socket(nexus_buffer_t *buffer, qos_request_obj *request_obj)
{
  nexus_get_u_long( buffer, (unsigned long*)&(((qos_request_socket *)request_obj)->proc), 1 );
#ifdef QOS_USE_RSVP
  nexus_printf("qoslib: Error: rspec/tspec _get_socket not implemented\n");
#endif
}


static int _sizeof_process(qos_request_obj *request_obj, int count)
{
  return (count * (
		   nexus_sizeof_byte(sizeof(qos_callback *))
#ifdef QOS_USE_POSIX_SCHED
		   + nexus_sizeof_byte( sizeof(qos_sched_class_t) )
		   + nexus_sizeof_byte( sizeof(qos+sched_priority_t) )
#endif
		   )
	  );
}

static void _put_process(nexus_buffer_t *buffer, qos_request_obj *request_obj)
{
  nexus_put_u_long( buffer, (unsigned long*)&(((qos_request_process *)request_obj)->proc), 1 );
#ifdef QOS_USE_POSIX_SCHED
  nexus_put_byte( buffer, (nexus_byte_t*)&(((qos_request_process *)request_obj)->class), sizeof(qos_sched_class_t) );
  nexus_put_byte( buffer, (nexus_byte_t*)&(((qos_request_process *)request_obj)->priority), sizeof(qos_sched_priority_t) );
#endif
}

static void _get_process(nexus_buffer_t *buffer, qos_request_obj *request_obj)
{
  nexus_get_u_long( buffer, (unsigned long*)&(((qos_request_process *)request_obj)->proc), 1 );
#ifdef QOS_USE_POSIX_SCHED
  nexus_get_byte( buffer, (nexus_byte_t*)&(((qos_request_process *)request_obj)->class), sizeof(qos_sched_class_t) );
  nexus_get_byte( buffer, (nexus_byte_t*)&(((qos_request_process *)request_obj)->priority), sizeof(qos_sched_priority_t) );
#endif
}


static int _sizeof_thread(qos_request_obj *request_obj, int count)
{
  return (count * (
		   nexus_sizeof_byte(sizeof(qos_callback *))
#ifdef QOS_USE_POSIX_SCHED
		   + nexus_sizeof_byte( sizeof(qos_sched_class_t) )
		   + nexus_sizeof_byte( sizeof(qos+sched_priority_t) )
#endif
		   )
	  );
}

static void _put_thread(nexus_buffer_t *buffer, qos_request_obj *request_obj)
{
  nexus_put_u_long( buffer, (unsigned long*)&(((qos_request_thread *)request_obj)->proc), 1 );
#ifdef QOS_USE_POSIX_SCHED
  nexus_put_byte( buffer, (nexus_byte_t*)&(((qos_request_thread *)request_obj)->class), sizeof(qos_sched_class_t) );
  nexus_put_byte( buffer, (nexus_byte_t*)&(((qos_request_thread *)request_obj)->priority), sizeof(qos_sched_priority_t) );
#endif
}

static void _get_thread(nexus_buffer_t *buffer, qos_request_obj *request_obj)
{
  nexus_get_u_long( buffer, (unsigned long*)&(((qos_request_thread *)request_obj)->proc), 1 );
#ifdef QOS_USE_POSIX_SCHED
  nexus_get_byte( buffer, (nexus_byte_t*)&(((qos_request_thread *)request_obj)->class), sizeof(qos_sched_class_t) );
  nexus_get_byte( buffer, (nexus_byte_t*)&(((qos_request_thread *)request_obj)->priority), sizeof(qos_sched_priority_t) );
#endif
}


static int _sizeof_memory(qos_request_obj *request_obj, int count)
{
  return (count * (
		   nexus_sizeof_byte( sizeof(qos_callback *) )
		   + nexus_sizeof_byte( sizeof(nexus_bool_t) )
		   )
	  );
}

static void _put_memory(nexus_buffer_t *buffer, qos_request_obj *request_obj)
{
  nexus_put_u_long( buffer, (unsigned long*)&(((qos_request_memory *)request_obj)->proc), 1 );
  nexus_put_byte( buffer, (nexus_byte_t*)&(((qos_request_memory *)request_obj)->lock), sizeof(nexus_bool_t) );
}

static void _get_memory(nexus_buffer_t *buffer, qos_request_obj *request_obj)
{
  nexus_get_u_long( buffer, (unsigned long*)&(((qos_request_memory *)request_obj)->proc), 1 );
  nexus_get_byte( buffer, (nexus_byte_t*)&(((qos_request_memory *)request_obj)->lock), sizeof(nexus_bool_t) );
}
