/*
 * qoslib.h
 *
 * rcsid = "$Header"
 *
 * QoS Request Definitions
 */

#ifndef _NEXUS_INCLUDE_QOSLIB_H
#define _NEXUS_INCLUDE_QOSLIB_H


/* QoSMan Operatoins */

#define QOS_RESOURCE_OP		0x00
#define QOS_LIST_OP		0x01
#define QOS_DETACH_OP		0x02
#define QOS_TERMINATE_OP	0x03

/*  QosLib Return Codes  */

#define QOS_FAILURE      -1
#define QOS_SUCCESS       0
#define QOS_NOT_ATTACHED  1


typedef int qos_status_obj;

typedef void (*qos_callback)(nexus_resource_name_t *, qos_status_obj *);

typedef struct {
        const nexus_resource_name_t 	*who;
        qos_callback 			*proc;
} qos_request_obj;


typedef struct {
        const nexus_resource_name_t 	*who;
        qos_callback 			*proc;
#ifdef QOS_USE_RSVP
	qos_rspec 			*rspec;
  	qos_tspec 			*tspec;
#endif  /* QOS_USE_RSVP */
} qos_request_socket;


typedef struct {
        const nexus_resource_name_t 	*who;
        qos_callback 			*proc;
#ifdef QOS_USE_POSIX_SCHED
        qos_sched_class_t 		class;
        qos_sched_priority_t 		priority;
#endif /* QOS_USE_POSIX_SCHED */
} qos_request_process;

typedef struct {
        const nexus_resource_name_t 	*who;
        qos_callback 			*proc;
#ifdef QOS_USE_POSIX_SCHED
        qos_sched_class_t 		class;
        qos_sched_priority_t 		priority;
#endif /* QOS_USE_POSIX_SCHED */
} qos_request_thread;


typedef struct {
        const nexus_resource_name_t 	*who;
        qos_callback 			*proc;
        nexus_bool_t	 		lock;
} qos_request_memory;


typedef union {
	qos_request_socket	socket;
	qos_request_process	process;
	qos_request_thread	thread;
	qos_request_memory	memory;
} qos_request_info;


extern int qos_request(qos_request_obj *);

extern int qos_sizeof_request(qos_request_obj*, int);
extern void qos_request_put(nexus_buffer_t*, qos_request_obj*);
extern void qos_request_get(nexus_buffer_t*, qos_request_obj*);

extern nexus_resource_name_t *nexus_context_name( void );
extern nexus_resource_name_t *nexus_thread_name( void );
extern nexus_resource_name_t *nexus_memory_name( void );
extern nexus_resource_name_t *nexus_socket_name( nexus_startpoint_t * );

extern void qos_list_qosman( void );

#endif /* _NEXUS_INCLUDE_QOSLIB_H */
