/*
 * st_mp.c
 */

#include "internal.h"
/*
 * Some preprocessors (Convex Exemplar) have problems with
 * #including a define.  Need to change this (and another
 * one in pr_mp.c) to manually include the proper mp_*.h
 * file.
 */
#include NEXUS_MP_INCLUDE

#ifndef NEXUS_MP_NODE_NAME
#define NEXUS_MP_NODE_NAME "unknown-" NEXUS_PROTO_NAME_MP
#endif

/*
 * Hack to pass argc and argv to MPI_Init().
 * Implementations which require valid argv and
 * argc pointers (such as Convex Exemplar) must
 * declare the _ptr variables as extern, and set
 * them before calling nexus_init().
 */
static int		hack_argc = 0;
static char **		hack_argv = (char **)NULL;
int *			hack_argc_ptr = &hack_argc;
char ***		hack_argv_ptr = &hack_argv;

static mp_destination_t	my_node;
static int		n_nodes;
static nexus_bool_t	nodes_started = NEXUS_FALSE;
static char		*master_name = NULL;
static char		**HACK = NULL;

extern char *		_nx_session_string;
extern int		_nx_session_string_length;

/* startup func prototypes */
static nexus_bool_t	st_mp_preinit(nexus_startup_node_t **nl_head,
				      nexus_startup_node_t **nl_tail);
static char *		st_mp_get_master_node_name(void);
static void     	st_mp_init(int *argc, char ***argv);
static int		st_mp_start_node(nexus_startup_node_t *node,
					 nexus_startpoint_t *reply_sp,
					 int first_checkin_number);

/* startup funcs table */
static nexus_startup_funcs_t st_mp_funcs =
{
    st_mp_preinit,
    st_mp_get_master_node_name,
    st_mp_init,			
    NULL,			/* st_shutdown          */
    NULL,			/* st_abort             */
    st_mp_start_node,
    NULL			/* st_mp_start_context	*/
};


/*
 * _nx_st_???_info()
 *
 * Return the function table for this module.
 */

void *MP_STARTUP_INFO(void)
{
    return ((void *) (&st_mp_funcs));
}
/* _nx_st_???_info() */

/*
 * st_mp_preinit()
 */

static nexus_bool_t st_mp_preinit(nexus_startup_node_t **nl_head,
				   nexus_startup_node_t **nl_tail)
{
    size_t		n_bytes, n_bytes_out;
    nexus_bool_t	ignore_command_line_args;    
    mp_receive_status_t	status;
    nexus_bool_t	done;
    nexus_bool_t	result;

    MP_INITIALIZE();

    *nl_head = *nl_tail = (nexus_startup_node_t *) NULL;
    
    /* Needed? some way to make sure this protocol exists */    
    MP_INIT_NODE_INFO(my_node, n_nodes);

    /*
     * If there is only one node in the computation, it should NOT
     * ignore the command line parameters.
     */
    if (n_nodes == 1)
	return (NEXUS_FALSE);

    MP_IS_MASTER_NODE(my_node, result);
    if (result == NEXUS_TRUE)
    {
	/*
	 * This is the master node.
	 * Setup a nexus_startup_node_t structure for the rest
	 * of the MP nodes.  That will cause st_mp_start_node
	 * to be called during Nexus initialization.
	 */
	nexus_startup_node_t *initial_node;
	
	NexusMalloc(st_mp_preinit(),
		    initial_node,
		    nexus_startup_node_t *,
		    sizeof(nexus_startup_node_t));
	
	initial_node->name = _nx_copy_string(NEXUS_MP_NODE_NAME);
	HACK = &(initial_node->name);
	initial_node->number = 1;
	initial_node->count = n_nodes - 1;
	initial_node->count_other_nodes = 0;
	initial_node->startup_module = _nx_copy_string(NEXUS_PROTO_NAME_MP);
	initial_node->startup_funcs = &st_mp_funcs;
	initial_node->next = (nexus_startup_node_t *) NULL;
	initial_node->directory_path = (char *) NULL;
	initial_node->executable_path = (char *) NULL;
	initial_node->st_ptr = NULL;

	*nl_head = *nl_tail = initial_node;

	ignore_command_line_args = NEXUS_FALSE;
    }
    else
    {
	/*
	 * This is a non-master node.
	 * Wait for a message from the master node that contains
	 * _nx_my_process_params for my node.
	 * That message is sent from st_mp_start_node.
	 */
	n_bytes = NEXUS_MAX_COMMAND_LENGTH;
	NexusMalloc(st_mp_preinit(),
		    _nx_my_process_params,
		    char *,
		    n_bytes);
	
	MP_POST_RECEIVE(st_mp_preinit(),
			_nx_my_process_params,
			n_bytes,
			status);

	MP_RECEIVE_WAIT(st_mp_preinit(),
			status,
			done);

	nodes_started = NEXUS_TRUE;
	ignore_command_line_args = NEXUS_TRUE;
    }

    return(ignore_command_line_args);
}
/* st_mp_preinit() */


/*
 * st_mp_init()
 */

static void st_mp_init(int *argc, char ***argv)
{
    int arg_num;
    
    /* session name */
    if ((arg_num = nexus_find_argument(argc, argv,
				       "nx_mp_session",
				       2)) >= 0)
    {
        _nx_session_string = _nx_copy_string((*argv)[arg_num+1]);
        nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    else
    {
	_nx_session_string = _nx_md_get_unique_session_string();
    }
    _nx_session_string_length = strlen(_nx_session_string);

    /* master name */
    if ((arg_num = nexus_find_argument(argc, argv, "nx_master_name", 2)) >= 0)
    {
	master_name = _nx_copy_string((*argv)[arg_num+1]);
	if (HACK)
	{
	    *HACK = master_name;
	}
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    else
    {
	master_name = _nx_copy_string(NEXUS_MP_NODE_NAME);
    }
}
/* st_mp_init() */

/*
 * st_mp_get_master_node_name()
 */

static char *st_mp_get_master_node_name(void)
{
    return(master_name);
}
/* st_mp_get_master_node_name() */


/*
 * st_mp_start_node()
 */

static int st_mp_start_node(nexus_startup_node_t *node,
			    nexus_startpoint_t *reply_sp,
			    int first_checkin_number)
{
    int i;
    char cmd[NEXUS_MAX_COMMAND_LENGTH], cmd2[NEXUS_MAX_COMMAND_LENGTH];
    int cmd2_len;
    int msglen;

    mp_destination_t	dest;
    mp_send_status_t	status;
    nexus_bool_t	done;
    
    /*
     * Cannot start dynamic nodes, but must call NexusAcquiredAsNode
     * on requested node
     */
    if (nodes_started == NEXUS_TRUE)
    {
	nexus_fatal("st_mp_start_node(): ERROR: Cannot acquire nodes "
		    "at this time when using " NEXUS_PROTO_NAME_MP "\n");
    }
    else
    {
	/* Start the nodes on the machines listed in arg_host_list */
	for (i = 0; i < node->count; i++)
	{
	    /* Get all arguments that need to go to this process */
	    _nx_new_process_params(cmd2, NEXUS_MAX_COMMAND_LENGTH);
	    cmd2_len = strlen(cmd2);
	    _nx_proto_get_creator_sp_params(cmd2 + cmd2_len,
					    NEXUS_MAX_COMMAND_LENGTH-cmd2_len,
					    reply_sp);
	    
	    nexus_stdio_lock();
	    sprintf(cmd, "%s -nx_node %s %d y -nx_node_id %d -nx_checkin %d "
		    "%s -nx_mp_session "
		    "%s -nx_master_name %s %s",
		    _nx_get_package_id_start(),
		    node->name,
		    node->number + i,
		    node->id + i,
		    first_checkin_number + i,
		    cmd2,
		    _nx_session_string,
		    st_mp_get_master_node_name(),
		    _nx_get_package_id_end() );
	    nexus_stdio_unlock();
	    
	    msglen = strlen(cmd) + 1;

	    MP_INIT_DESTINATION(dest);
	    MP_SET_DESTINATION(dest, node->number + i);

	    MP_SEND(dest,
		    cmd,
		    msglen,
		    status);

	    do {
		MP_SEND_STATUS(status, done);
	    } while (done == NEXUS_FALSE);

	    MP_FREE_DESTINATION(dest);
	    
	    nodes_started = NEXUS_TRUE;
	}
    }

    return(0); /* NEXUS_SUCCESS?? */
}
/* _nx_start_nodes() */
