/*
 * dc_config.h
 */

/*
 * Converting denormals allows for more precision,
 * but some machines have done away with them to gain
 * speed.  Converting on these machines would only
 * waste CPU cycles.
 *
 * TODO:
 *   On the other end of the scale, the T3E is dumping
 *   core with an FP exception every time it sees a
 *   denormal.  We need to find the compiler switch to get
 *   compliance, or check every float and double that
 *   is received and zero out the denormals.  DEC's Alpha
 *   compiler's option -ieee fixes the problem on Ptera,
 *   so the problem is in the Cray compiler and not the chip.
 *
 * If this is 0, all conversions between IEEE and CRAY
 * that fall in the IEEE denormal range will be set to 0.
 */
#define ALLOW_DENORMALS		0

/*
 * Define all the data types.
 */
#define	S8	5000
#define	U8	5001
#define	S16LE	5002
#define S16BE	5003
#define U16LE	5004
#define U16BE	5005
#define S32LE	5006
#define S32BE	5007
#define U32LE	5008
#define U32BE	5009
#define S64LE	5010
#define S64BE	5011
#define U64LE	5012
#define U64BE	5013
#define IEEE32LE 5014
#define IEEE32BE 5015
#define IEEE64LE 5016
#define IEEE64BE 5017
#define CRAY64BE 5018

/*
 * Specification of native data type formats:
 *   [SU]<# bits>[LB]E
 *    -Signed/Unsigned
 *    -Little/Big Endian
 *   IEEE = floating point format
 *   CRAY = floating point format
 *   INT32_T = data type that is 32 bit integer
 *   INT64_T = data type that is 64 bit integer
 */
#if	NEXUS_DC_FORMAT_LOCAL == NEXUS_DC_FORMAT_32BIT_LE
#  undef  BIG_ENDIAN
#  define NATIVE_CHAR	S8
#  define NATIVE_UCHAR	U8
#  define NATIVE_SHORT	S16LE
#  define NATIVE_USHORT	U16LE
#  define NATIVE_INT	S32LE
#  define NATIVE_UINT	U32LE
#  define NATIVE_LONG	S32LE
#  define NATIVE_ULONG	U32LE
#  define NATIVE_FLOAT  IEEE32LE
#  define NATIVE_DOUBLE IEEE64LE
#  define INT32_T	int
#  undef  INT64_T
#elif	NEXUS_DC_FORMAT_LOCAL == NEXUS_DC_FORMAT_32BIT_BE
#  define BIG_ENDIAN
#  define NATIVE_CHAR	S8
#  define NATIVE_UCHAR	U8
#  define NATIVE_SHORT	S16BE
#  define NATIVE_USHORT	U16BE
#  define NATIVE_INT	S32BE
#  define NATIVE_UINT	U32BE
#  define NATIVE_LONG	S32BE
#  define NATIVE_ULONG	U32BE
#  define NATIVE_FLOAT  IEEE32BE
#  define NATIVE_DOUBLE IEEE64BE
#  define INT32_T	int
#  undef  INT64_T
#elif	NEXUS_DC_FORMAT_LOCAL == NEXUS_DC_FORMAT_64BIT_LE
#  undef  BIG_ENDIAN
#  define NATIVE_CHAR	S8
#  define NATIVE_UCHAR	U8
#  define NATIVE_SHORT	S16LE
#  define NATIVE_USHORT	U16LE
#  define NATIVE_INT	S32LE
#  define NATIVE_UINT	U32LE
#  define NATIVE_LONG	S64LE
#  define NATIVE_ULONG	U64LE
#  define NATIVE_FLOAT  IEEE32LE
#  define NATIVE_DOUBLE IEEE64LE
#  define INT32_T	int
#  define INT64_T	long
#elif	NEXUS_DC_FORMAT_LOCAL == NEXUS_DC_FORMAT_64BIT_BE
#  define BIG_ENDIAN
#  define NATIVE_CHAR	S8
#  define NATIVE_UCHAR	U8
#  define NATIVE_SHORT	S16BE
#  define NATIVE_USHORT	U16BE
#  define NATIVE_INT	S32BE
#  define NATIVE_UINT	U32BE
#  define NATIVE_LONG	S64BE
#  define NATIVE_ULONG	U64BE
#  define NATIVE_FLOAT  IEEE32BE
#  define NATIVE_DOUBLE IEEE64BE
#  define INT32_T	int
#  define INT64_T	long
#elif	NEXUS_DC_FORMAT_LOCAL == NEXUS_DC_FORMAT_CRAYC90
#  define BIG_ENDIAN
#  define NATIVE_CHAR	S8
#  define NATIVE_UCHAR	U8
#  define NATIVE_SHORT	S64BE
#  define NATIVE_USHORT	U64BE
#  define NATIVE_INT	S64BE
#  define NATIVE_UINT	U64BE
#  define NATIVE_LONG	S64BE
#  define NATIVE_ULONG	U64BE
#  define NATIVE_FLOAT  CRAY64BE
#  define NATIVE_DOUBLE CRAY64BE
#  undef  INT32_T
#  define INT64_T	long
#elif	NEXUS_DC_FORMAT_LOCAL == NEXUS_DC_FORMAT_CRAYT3E
#  define BIG_ENDIAN
#  define NATIVE_CHAR	S8
#  define NATIVE_UCHAR	U8
#  define NATIVE_SHORT	S32BE
#  define NATIVE_USHORT	U32BE
#  define NATIVE_INT	S64BE
#  define NATIVE_UINT	U64BE
#  define NATIVE_LONG	S64BE
#  define NATIVE_ULONG	U64BE
#  define NATIVE_FLOAT  IEEE32BE
#  define NATIVE_DOUBLE IEEE64BE
#  define INT32_T	short
#  define INT64_T	long
#else
   /* error -- unsupported format! */
#endif

/*
 * Define offsets of bytes in words.
 *
 * For example:
 *   #define BYTE_16_0	1
 * means the least significant byte
 * of a 16 bit word is at offset 1.
 */
#ifdef BIG_ENDIAN

/* */
#  define BYTE_16_0	1
#  define BYTE_16_1	0
/* */
#  define BYTE_32_0	3
#  define BYTE_32_1	2
#  define BYTE_32_2	1
#  define BYTE_32_3	0
/* */
#  define BYTE_64_0	7
#  define BYTE_64_1	6
#  define BYTE_64_2	5
#  define BYTE_64_3	4
#  define BYTE_64_4	3
#  define BYTE_64_5	2
#  define BYTE_64_6	1
#  define BYTE_64_7	0

#else /* !BIG_ENDIAN */

/* */
#  define BYTE_16_0	0
#  define BYTE_16_1	1
/* */
#  define BYTE_32_0	0
#  define BYTE_32_1	1
#  define BYTE_32_2	2
#  define BYTE_32_3	3
/* */
#  define BYTE_64_0	0
#  define BYTE_64_1	1
#  define BYTE_64_2	2
#  define BYTE_64_3	3
#  define BYTE_64_4	4
#  define BYTE_64_5	5
#  define BYTE_64_6	6
#  define BYTE_64_7	7

#endif /* BIG_ENDIAN */

/***************************************
 * Supported dc_formats and type sizes *
 ***************************************/

/*
 * supported dc_formats:
 *   32LE, 64LE,
 *   32BE, 64BE,
 *   JAVA,
 *   CRAYC90, CRAYT3E
 */

/*
 * supported type formats:
 *   char:
 *     U8, S16BE
 *   uchar:
 *     U8, S16BE
 *   short:
 *     S16LE, S16BE, S32BE, S64BE
 *   ushort:
 *     U16LE, U16BE, S16BE, U32BE, U64BE
 *   int:
 *     S32LE, S32BE, S64BE
 *   uint:
 *     U32LE, U32BE, S32BE, S64BE
 *   long:
 *     S32LE, S32BE, S64LE, S64BE
 *   ulong:
 *     U32LE, U32BE, U64LE, U64BE, S64BE
 *   float:
 *     ieee32, cray64
 *   double:
 *     ieee64, cray64
 */

/*
 * FILLVAL()
 *
 * Macro to generate a byte with all bits set
 * the same as sign bit of the input byte.
 * First two methods will avoid a branch if the
 * behavior of >> on signed chars is known.
 *
 * (b) must be used exactly once, since the
 * macro is used f = FILLVAL(dst[x] = src[y]).
 */
#ifdef SIGNED_RSHIFT
#  define FILLVAL(b)	((nexus_byte_t)(((signed char)(b)) >> 7))
#else
#  ifdef UNSIGNED_RSHIFT
#    define FILLVAL(b)	(0 - ((b) >> 7))
#  else
     /* default */
#    define FILLVAL(b)	(((b) & 0x80) ? 0xff : 0x00)
#  endif
#endif

/********************************************
 * Lists of sources for all types and sizes *
 ********************************************/

/* source char formats: */
#define SOURCE_CHAR_S8 \
          NEXUS_DC_FORMAT_32BIT_LE: \
     case NEXUS_DC_FORMAT_32BIT_BE: \
     case NEXUS_DC_FORMAT_64BIT_LE: \
     case NEXUS_DC_FORMAT_64BIT_BE: \
     case NEXUS_DC_FORMAT_CRAYC90: \
     case NEXUS_DC_FORMAT_CRAYT3E
#define SOURCE_CHAR_S16 \
          NEXUS_DC_FORMAT_JAVA

/* source uchar formats */
#define SOURCE_UCHAR_U8 \
          NEXUS_DC_FORMAT_32BIT_LE: \
     case NEXUS_DC_FORMAT_32BIT_BE: \
     case NEXUS_DC_FORMAT_64BIT_LE: \
     case NEXUS_DC_FORMAT_64BIT_BE: \
     case NEXUS_DC_FORMAT_CRAYC90: \
     case NEXUS_DC_FORMAT_CRAYT3E
#define SOURCE_UCHAR_S16 \
	  NEXUS_DC_FORMAT_JAVA

/* source short formats */
#define SOURCE_SHORT_S16LE \
          NEXUS_DC_FORMAT_32BIT_LE: \
     case NEXUS_DC_FORMAT_64BIT_LE
#define SOURCE_SHORT_S16BE \
          NEXUS_DC_FORMAT_32BIT_BE: \
     case NEXUS_DC_FORMAT_64BIT_BE: \
     case NEXUS_DC_FORMAT_JAVA
#define SOURCE_SHORT_S32BE \
	  NEXUS_DC_FORMAT_CRAYT3E
#define SOURCE_SHORT_S64BE \
	  NEXUS_DC_FORMAT_CRAYC90

/* source ushort formats */
#define SOURCE_USHORT_U16LE \
	  NEXUS_DC_FORMAT_32BIT_LE: \
     case NEXUS_DC_FORMAT_64BIT_LE
#define SOURCE_USHORT_U16BE \
	  NEXUS_DC_FORMAT_32BIT_BE: \
     case NEXUS_DC_FORMAT_64BIT_BE
#define SOURCE_USHORT_S16BE \
	  NEXUS_DC_FORMAT_JAVA
#define SOURCE_USHORT_U32BE \
	  NEXUS_DC_FORMAT_CRAYT3E
#define SOURCE_USHORT_U64BE \
	  NEXUS_DC_FORMAT_CRAYC90

/* source int formats */
#define SOURCE_INT_S32LE \
	  NEXUS_DC_FORMAT_32BIT_LE: \
     case NEXUS_DC_FORMAT_64BIT_LE
#define SOURCE_INT_S32BE \
	  NEXUS_DC_FORMAT_32BIT_BE: \
     case NEXUS_DC_FORMAT_64BIT_BE: \
     case NEXUS_DC_FORMAT_JAVA
#define SOURCE_INT_S64BE \
	  NEXUS_DC_FORMAT_CRAYC90: \
     case NEXUS_DC_FORMAT_CRAYT3E
	  
/* source uint formats */
#define SOURCE_UINT_U32LE \
	  NEXUS_DC_FORMAT_32BIT_LE: \
     case NEXUS_DC_FORMAT_64BIT_LE
#define SOURCE_UINT_U32BE \
	  NEXUS_DC_FORMAT_32BIT_BE: \
     case NEXUS_DC_FORMAT_64BIT_BE
#define SOURCE_UINT_S32BE \
	  NEXUS_DC_FORMAT_JAVA
#define SOURCE_UINT_U64BE \
	  NEXUS_DC_FORMAT_CRAYC90: \
     case NEXUS_DC_FORMAT_CRAYT3E

/* source long formats */
#define SOURCE_LONG_S32LE \
	  NEXUS_DC_FORMAT_32BIT_LE
#define SOURCE_LONG_S32BE \
	  NEXUS_DC_FORMAT_32BIT_BE
#define SOURCE_LONG_S64LE \
	  NEXUS_DC_FORMAT_64BIT_LE
#define SOURCE_LONG_S64BE \
	  NEXUS_DC_FORMAT_64BIT_BE: \
     case NEXUS_DC_FORMAT_JAVA: \
     case NEXUS_DC_FORMAT_CRAYC90: \
     case NEXUS_DC_FORMAT_CRAYT3E
	  
/* source ulong formats */
#define SOURCE_ULONG_U32LE \
	  NEXUS_DC_FORMAT_32BIT_LE
#define SOURCE_ULONG_U32BE \
	  NEXUS_DC_FORMAT_32BIT_BE
#define SOURCE_ULONG_U64LE \
	  NEXUS_DC_FORMAT_64BIT_LE
#define SOURCE_ULONG_U64BE \
	  NEXUS_DC_FORMAT_64BIT_BE: \
     case NEXUS_DC_FORMAT_CRAYC90: \
     case NEXUS_DC_FORMAT_CRAYT3E
#define SOURCE_ULONG_S64BE \
	  NEXUS_DC_FORMAT_JAVA

/* source float formats */
#define SOURCE_FLOAT_IEEE32LE \
	  NEXUS_DC_FORMAT_32BIT_LE: \
     case NEXUS_DC_FORMAT_64BIT_LE
#define SOURCE_FLOAT_IEEE32BE \
	  NEXUS_DC_FORMAT_32BIT_BE: \
     case NEXUS_DC_FORMAT_64BIT_BE: \
     case NEXUS_DC_FORMAT_JAVA: \
     case NEXUS_DC_FORMAT_CRAYT3E
#define SOURCE_FLOAT_CRAY64BE \
	  NEXUS_DC_FORMAT_CRAYC90
	  
/* source double formats */
#define SOURCE_DOUBLE_IEEE64LE \
	  NEXUS_DC_FORMAT_32BIT_LE: \
     case NEXUS_DC_FORMAT_64BIT_LE
#define SOURCE_DOUBLE_IEEE64BE \
	  NEXUS_DC_FORMAT_32BIT_BE: \
     case NEXUS_DC_FORMAT_64BIT_BE: \
     case NEXUS_DC_FORMAT_JAVA: \
     case NEXUS_DC_FORMAT_CRAYT3E
#define SOURCE_DOUBLE_CRAY64BE \
	  NEXUS_DC_FORMAT_CRAYC90
	  
