/*
 * nexus_dc.c
 */

/*
 * Don't change the order of #includes or
 * the macros will not expand correctly!
 */

/* 1) nexus_dc.h */
#include "nexus_dc.h"

/* 2) dc_config.h */
#include "dc_config.h"

/* 3) dc_get_macros.h */
#include "dc_get_macros.h"

/* 4) dc_chk_macros.h */
#include "dc_chk_macros.h"

/*
 * nexus_dc_is_native_....()
 *
 * Return NEXUS_TRUE if the format is the same as
 * the native format, or NEXUS_FALSE if conversion
 * is required.
 */

int nexus_dc_is_native_char(int format)
{
    if (format == NEXUS_DC_FORMAT_JAVA)
	return (NEXUS_FALSE);
    else
	return (NEXUS_TRUE);
}

int nexus_dc_is_native_u_char(int format)
{
    if (format == NEXUS_DC_FORMAT_JAVA)
	return (NEXUS_FALSE);
    else
	return (NEXUS_TRUE);	    
}

int nexus_dc_is_native_short(int format)
{
    switch (format)
    {
#if   NATIVE_SHORT == S16LE
    case SOURCE_SHORT_S16LE:

#elif NATIVE_SHORT == S16BE	
    case SOURCE_SHORT_S16BE:
	
#elif NATIVE_SHORT == S32BE
    case SOURCE_SHORT_S32BE:
	
#elif NATIVE_SHORT == S64BE
    case SOURCE_SHORT_S64BE:
	
#endif
	return (NEXUS_TRUE);
    default:
	return (NEXUS_FALSE);
    }
}

int nexus_dc_is_native_u_short(int format)
{
    switch (format)
    {
#if   NATIVE_USHORT == U16LE
    case SOURCE_USHORT_U16LE:
	
#elif NATIVE_USHORT == U16BE
    case SOURCE_USHORT_U16BE:
	
#elif NATIVE_USHORT == U32BE
    case SOURCE_USHORT_U32BE:
	
#elif NATIVE_USHORT == U64BE
    case SOURCE_USHORT_U64BE:
	
#endif
	return (NEXUS_TRUE);
    default:
	return (NEXUS_FALSE);
    }
}

int nexus_dc_is_native_int(int format)
{
    switch (format)
    {
#if   NATIVE_INT == S32LE
    case SOURCE_INT_S32LE:
	
#elif NATIVE_INT == S32BE
    case SOURCE_INT_S32BE:
	
#elif NATIVE_INT == S64BE
    case SOURCE_INT_S64BE:
	
#endif
	return (NEXUS_TRUE);
    default:
	return (NEXUS_FALSE);
    }
}

int nexus_dc_is_native_u_int(int format)
{
    switch (format)
    {
#if   NATIVE_UINT == U32LE
    case SOURCE_UINT_U32LE:
	
#elif NATIVE_UINT == U32BE
    case SOURCE_UINT_U32BE:
	
#elif NATIVE_UINT == U64BE
    case SOURCE_UINT_U64BE:
	
#endif
	return (NEXUS_TRUE);
    default:
	return (NEXUS_FALSE);
    }
}

int nexus_dc_is_native_long(int format)
{
    switch (format)
    {
#if   NATIVE_LONG == S32LE
    case SOURCE_LONG_S32LE:
	
#elif NATIVE_LONG == S32BE
    case SOURCE_LONG_S32BE:
	
#elif NATIVE_LONG == S64LE
    case SOURCE_LONG_S64LE:
	
#elif NATIVE_LONG == S64BE
    case SOURCE_LONG_S64BE:
	
#endif
	return (NEXUS_TRUE);
    default:
	return (NEXUS_FALSE);
    }
}

int nexus_dc_is_native_u_long(int format)
{
    switch (format)
    {
#if   NATIVE_ULONG == U32LE
    case SOURCE_ULONG_U32LE:
	
#elif NATIVE_ULONG == U32BE
    case SOURCE_ULONG_U32BE:
	
#elif NATIVE_ULONG == U64LE
    case SOURCE_ULONG_U64LE:
	
#elif NATIVE_ULONG == U64BE
    case SOURCE_ULONG_U64BE:
	
#endif
	return (NEXUS_TRUE);
    default:
	return (NEXUS_FALSE);
    }
}

int nexus_dc_is_native_float(int format)
{
    switch (format)
    {
#if   NATIVE_FLOAT == IEEE32LE
    case SOURCE_FLOAT_IEEE32LE:

#elif NATIVE_FLOAT == IEEE32BE
    case SOURCE_FLOAT_IEEE32BE:

#elif NATIVE_FLOAT == CRAY64BE
    case SOURCE_FLOAT_CRAY64BE:
	
#endif
	return (NEXUS_TRUE);
    default:
	return (NEXUS_FALSE);
    }
}

int nexus_dc_is_native_double(int format)
{
    switch (format)
    {
#if   NATIVE_DOUBLE == IEEE64LE
    case SOURCE_DOUBLE_IEEE64LE:
	
#elif NATIVE_DOUBLE == IEEE64BE
    case SOURCE_DOUBLE_IEEE64BE:

#elif NATIVE_DOUBLE == CRAY64BE
    case SOURCE_DOUBLE_CRAY64BE:

#endif
	return (NEXUS_TRUE);
    default:
	return (NEXUS_FALSE);
    }
}

/*
 * nexus_dc_get_...()
 *
 */

void nexus_dc_get_char(nexus_byte_t **buffer,
		       char *data,
		       unsigned long count,
		       int format)
{
    DECLARE_GET_VARS;
    
    if (format == NEXUS_DC_FORMAT_JAVA)
    { /* {} req'd */
	GET_16BE_TO_8();
    }
    else
    {
	GET_8_TO_8();
    }
}

void nexus_dc_get_u_char(nexus_byte_t **buffer,
			 unsigned char *data,
			 unsigned long count,
			 int format)
{
    DECLARE_GET_VARS;

    if (format == NEXUS_DC_FORMAT_JAVA)
    { /* {} req'd */
	GET_16BE_TO_8();
    }
    else
    {
	GET_8_TO_8();
    }
}

void nexus_dc_get_short(nexus_byte_t **buffer,
			short *data,
			unsigned long count,
			int format)
{
    DECLARE_GET_VARS;
    
    switch (format)
    {
    case SOURCE_SHORT_S16LE:
#if   NATIVE_SHORT == S16LE
	GET_16_TO_16();
#elif NATIVE_SHORT == S16BE
	GET_16_TO_16_SWAP();
#elif NATIVE_SHORT == S32BE
	GET_S16LE_TO_S32();
#elif NATIVE_SHORT == S64BE
	GET_S16LE_TO_S64();
#endif
	break;
	
    case SOURCE_SHORT_S16BE:
#if   NATIVE_SHORT == S16LE
	GET_16_TO_16_SWAP();
#elif NATIVE_SHORT == S16BE
	GET_16_TO_16();
#elif NATIVE_SHORT == S32BE
	GET_S16BE_TO_S32();
#elif NATIVE_SHORT == S64BE
	GET_S16BE_TO_S64();
#endif
	break;
	
    case SOURCE_SHORT_S32BE:
#if   (NATIVE_SHORT == S16LE) || (NATIVE_SHORT == S16BE)
	GET_32BE_TO_16();
#elif NATIVE_SHORT == S32BE
	GET_32_TO_32();
#elif NATIVE_SHORT == S64BE
	GET_S32BE_TO_S64();
#endif
	break;
	
    case SOURCE_SHORT_S64BE:
#if   (NATIVE_SHORT == S16LE) || (NATIVE_SHORT == S16BE)
	GET_64BE_TO_16();
#elif NATIVE_SHORT == S32BE
	GET_64BE_TO_32();
#elif NATIVE_SHORT == S64BE
	GET_64_TO_64();
#endif
	break;
    }
}

void nexus_dc_get_u_short(nexus_byte_t **buffer,
			  unsigned short *data,
			  unsigned long count,
			  int format)
{
    DECLARE_GET_VARS;

    switch (format)
    {
    case SOURCE_USHORT_U16LE:
#if   NATIVE_USHORT == U16LE
	GET_16_TO_16();
#elif NATIVE_USHORT == U16BE
	GET_16_TO_16_SWAP();
#elif NATIVE_USHORT == U32BE
	GET_16LE_TO_U32();
#elif NATIVE_USHORT == U64BE
	GET_16LE_TO_U64();
#endif
	break;
	
    case SOURCE_USHORT_U16BE:
    case SOURCE_USHORT_S16BE:
#if   NATIVE_USHORT == U16LE
	GET_16_TO_16_SWAP();
#elif NATIVE_USHORT == U16BE
	GET_16_TO_16();
#elif NATIVE_USHORT == U32BE
	GET_16BE_TO_U32();
#elif NATIVE_USHORT == U64BE
	GET_16BE_TO_U64();
#endif
	break;

    case SOURCE_USHORT_U32BE:
#if   (NATIVE_USHORT == U16LE) || (NATIVE_USHORT == U16BE)
	GET_32BE_TO_16();
#elif NATIVE_USHORT == U32BE
	GET_32_TO_32();
#elif NATIVE_USHORT == U64BE
	GET_32BE_TO_U64();
#endif
	break;

    case SOURCE_USHORT_U64BE:
#if   (NATIVE_USHORT == U16LE) || (NATIVE_USHORT == U16BE)
	GET_64BE_TO_16();
#elif NATIVE_USHORT == U32BE
	GET_64BE_TO_32();
#elif NATIVE_USHORT == U64BE
	GET_64_TO_64();
#endif
	break;
    }
}

void nexus_dc_get_int(nexus_byte_t **buffer,
		      int *data,
		      unsigned long count,
		      int format)
{
    DECLARE_GET_VARS;

    switch (format)
    {
    case SOURCE_INT_S32LE:
#if   NATIVE_INT == S32LE
	GET_32_TO_32();
#elif NATIVE_INT == S32BE
	GET_32_TO_32_SWAP();
#elif NATIVE_INT == S64BE
	GET_S32LE_TO_S64();
#endif
	break;
	
    case SOURCE_INT_S32BE:
#if   NATIVE_INT == S32LE
	GET_32_TO_32_SWAP();
#elif NATIVE_INT == S32BE
	GET_32_TO_32();
#elif NATIVE_INT == S64BE
	GET_S32BE_TO_S64();
#endif
	break;
	
    case SOURCE_INT_S64BE:
#if   (NATIVE_INT == S32LE) || (NATIVE_INT == S32BE)
	GET_64BE_TO_32();
#elif NATIVE_INT == S64BE
	GET_64_TO_64();
#endif
	break;
    }
}

void nexus_dc_get_u_int(nexus_byte_t **buffer,
			unsigned int *data,
			unsigned long count,
			int format)
{
    DECLARE_GET_VARS;

    switch (format)
    {
    case SOURCE_UINT_U32LE:
#if   NATIVE_UINT == U32LE
	GET_32_TO_32();
#elif NATIVE_UINT == U32BE
	GET_32_TO_32_SWAP();
#elif NATIVE_UINT == U64BE
	GET_32LE_TO_U64();
#endif
	break;
	
    case SOURCE_UINT_U32BE:
    case SOURCE_UINT_S32BE:
#if   NATIVE_UINT == U32LE
	GET_32_TO_32_SWAP();
#elif NATIVE_UINT == U32BE
	GET_32_TO_32();
#elif NATIVE_UINT == U64BE
	GET_32BE_TO_U64();
#endif
	break;

    case SOURCE_UINT_U64BE:
#if   (NATIVE_UINT == U32LE) || (NATIVE_UINT == U32BE)
	GET_64BE_TO_32();
#elif NATIVE_UINT == U64BE
	GET_64_TO_64();
#endif
	break;
    }
}

void nexus_dc_get_long(nexus_byte_t **buffer,
		       long *data,
		       unsigned long count,
		       int format)
{
    DECLARE_GET_VARS;

    switch (format)
    {
    case SOURCE_LONG_S32LE:
#if   NATIVE_LONG == S32LE
	GET_32_TO_32();
#elif NATIVE_LONG == S32BE
	GET_32_TO_32_SWAP();
#elif (NATIVE_LONG == S64LE) || (NATIVE_LONG == S64BE)
	GET_S32LE_TO_S64();
#endif
	break;

    case SOURCE_LONG_S32BE:
#if   NATIVE_LONG == S32LE
	GET_32_TO_32_SWAP();
#elif NATIVE_LONG == S32BE
	GET_32_TO_32();
#elif (NATIVE_LONG == S64LE) || (NATIVE_LONG == S64BE)
	GET_S32BE_TO_S64();
#endif
	break;

    case SOURCE_LONG_S64LE:
#if   (NATIVE_LONG == S32LE) || (NATIVE_LONG == S32BE)
	GET_64LE_TO_32();
#elif NATIVE_LONG == S64LE
	GET_64_TO_64();
#elif NATIVE_LONG == S64BE
	GET_64_TO_64_SWAP();
#endif
	break;

    case SOURCE_LONG_S64BE:
#if   (NATIVE_LONG == S32LE) || (NATIVE_LONG == S32BE)
	GET_64BE_TO_32();
#elif NATIVE_LONG == S64LE
	GET_64_TO_64_SWAP();
#elif NATIVE_LONG == S64BE
	GET_64_TO_64();
#endif
	break;
    }
}

void nexus_dc_get_u_long(nexus_byte_t **buffer,
			 unsigned long *data,
			 unsigned long count,
			 int format)
{
    DECLARE_GET_VARS;

    switch (format)
    {
    case SOURCE_ULONG_U32LE:
#if   NATIVE_ULONG == U32LE
	GET_32_TO_32();
#elif NATIVE_ULONG == U32BE
	GET_32_TO_32_SWAP();
#elif (NATIVE_ULONG == U64LE) || (NATIVE_ULONG == U64BE)
	GET_32LE_TO_U64();
#endif
	break;

    case SOURCE_ULONG_U32BE:
#if   NATIVE_ULONG == U32LE
	GET_32_TO_32_SWAP();
#elif NATIVE_ULONG == U32BE
	GET_32_TO_32();
#elif (NATIVE_ULONG == U64LE) || (NATIVE_ULONG == U64BE)
	GET_32BE_TO_U64();
#endif
	break;

    case SOURCE_ULONG_U64LE:
#if   (NATIVE_ULONG == U32LE) || (NATIVE_ULONG == U32BE)
	GET_64LE_TO_32();
#elif NATIVE_ULONG == U64LE
	GET_64_TO_64();
#elif NATIVE_ULONG == U64BE
	GET_64_TO_64_SWAP();
#endif
	break;

    case SOURCE_ULONG_U64BE:
    case SOURCE_ULONG_S64BE:
#if   (NATIVE_ULONG == U32LE) || (NATIVE_ULONG == U32BE)
	GET_64BE_TO_32();
#elif NATIVE_ULONG == U64LE
	GET_64_TO_64_SWAP();
#elif NATIVE_ULONG == U64BE
	GET_64_TO_64();
#endif
	break;
    }
}

void nexus_dc_get_float(nexus_byte_t **buffer,
			float *data,
			unsigned long count,
			int format)
{
    DECLARE_GET_VARS;

    switch (format)
    {
    case SOURCE_FLOAT_IEEE32LE:
#if   NATIVE_FLOAT == IEEE32LE
	GET_32_TO_32();
#elif NATIVE_FLOAT == IEEE32BE
	GET_32_TO_32_SWAP();
#elif NATIVE_FLOAT == CRAY64BE
	GET_IEEE32LE_TO_CRAY64();
#endif
	break;

    case SOURCE_FLOAT_IEEE32BE:
#if   NATIVE_FLOAT == IEEE32LE
	GET_32_TO_32_SWAP();
#elif NATIVE_FLOAT == IEEE32BE
	GET_32_TO_32();
#elif NATIVE_FLOAT == CRAY64BE
	GET_IEEE32BE_TO_CRAY64();
#endif
	break;

    case SOURCE_FLOAT_CRAY64BE:
#if   (NATIVE_FLOAT == IEEE32LE) || (NATIVE_FLOAT == IEEE32BE)
	GET_CRAY64BE_TO_IEEE32();
#elif NATIVE_FLOAT == CRAY64BE
	GET_64_TO_64();
#endif
	break;
    }
}

void nexus_dc_get_double(nexus_byte_t **buffer,
			 double *data,
			 unsigned long count,
			 int format)
{
    DECLARE_GET_VARS;

    switch (format)
    {
    case SOURCE_DOUBLE_IEEE64LE:
#if   NATIVE_DOUBLE == IEEE64LE
	GET_64_TO_64();
#elif NATIVE_DOUBLE == IEEE64BE
	GET_64_TO_64_SWAP();
#elif NATIVE_DOUBLE == CRAY64BE
	GET_IEEE64LE_TO_CRAY64();
#endif
	break;

    case SOURCE_DOUBLE_IEEE64BE:
#if   NATIVE_DOUBLE == IEEE64LE
	GET_64_TO_64_SWAP();
#elif NATIVE_DOUBLE == IEEE64BE
	GET_64_TO_64();
#elif NATIVE_DOUBLE == CRAY64BE
	GET_IEEE64BE_TO_CRAY64();
#endif
	break;

    case SOURCE_DOUBLE_CRAY64BE:
#if   ((NATIVE_DOUBLE == IEEE64LE) || (NATIVE_DOUBLE == IEEE64BE))
#  if   defined (INT64_T) /* prefer 64 bit version */
	GET_CRAY64BE_TO_IEEE64_64();
#  elif defined (INT32_T)
	GET_CRAY64BE_TO_IEEE64_32();
#  else
	/* error -- no 32 or 64 bit int types defined */
#  endif
#elif NATIVE_DOUBLE == CRAY64BE
	GET_64_TO_64();
#endif
	break;
    }
}

/*
 * nexus_dc_check_lost_precision_...()
 *
 */

int nexus_dc_check_lost_precision_char(nexus_byte_t *buffer,
				       unsigned long count,
				       int format)
{
    DECLARE_CHK_VARS;
    
    if (format == NEXUS_DC_FORMAT_JAVA)
    { /* {}'s req'd */
	CHK_S16BE_TO_S8();
    }

    return (-1);
}

int nexus_dc_check_lost_precision_u_char(nexus_byte_t *buffer,
					 unsigned long count,
					 int format)
{
    DECLARE_CHK_VARS;

    if (format == NEXUS_DC_FORMAT_JAVA)
    { /* {}'s req'd */
	CHK_S16BE_TO_U8();
    }

    return (-1);
}

int nexus_dc_check_lost_precision_short(nexus_byte_t *buffer,
					unsigned long count,
					int format)
{
    DECLARE_CHK_VARS;

    switch (format)
    {
    case SOURCE_SHORT_S16LE:
    case SOURCE_SHORT_S16BE:
	break;
	
    case SOURCE_SHORT_S32BE:
#if   (NATIVE_SHORT == S16LE) || (NATIVE_SHORT == S16BE)
	CHK_S32BE_TO_S16();
#endif
	break;
	
    case SOURCE_SHORT_S64BE:
#if   (NATIVE_SHORT == S16LE) || (NATIVE_SHORT == S16BE)
	CHK_S64BE_TO_S16();
#elif (NATIVE_SHORT == S32BE)
	CHK_S64BE_TO_S32();
#endif
	break;
    }

    return (-1);
}

int nexus_dc_check_lost_precision_u_short(nexus_byte_t *buffer,
					  unsigned long count,
					  int format)
{
    DECLARE_CHK_VARS;

    switch (format)
    {
    case SOURCE_USHORT_U16LE:
    case SOURCE_USHORT_U16BE:
	break;
	
    case SOURCE_USHORT_S16BE:
	CHK_JAVA_S_TO_U(2);
	break;
	
    case SOURCE_USHORT_U32BE:
#if   (NATIVE_USHORT == U16LE) || (NATIVE_USHORT == U16BE)
	CHK_U32BE_TO_U16();
#endif
	break;
	
    case SOURCE_USHORT_U64BE:
#if   (NATIVE_USHORT == U16LE) || (NATIVE_USHORT == U16BE)
	CHK_U64BE_TO_U16();
#elif (NATIVE_USHORT == U32BE)
	CHK_U64BE_TO_U32();
#endif
	break;
    }

    return (-1);
}

int nexus_dc_check_lost_precision_int(nexus_byte_t *buffer,
				      unsigned long count,
				      int format)
{
    DECLARE_CHK_VARS;

    switch (format)
    {
    case SOURCE_INT_S32LE:
    case SOURCE_INT_S32BE:
	break;

    case SOURCE_INT_S64BE:
#if (NATIVE_INT == S32LE) || (NATIVE_INT == S32BE)
	CHK_S64BE_TO_S32();
#endif
	break;
    }

    return (-1);
}

int nexus_dc_check_lost_precision_u_int(nexus_byte_t *buffer,
					unsigned long count,
					int format)
{
    DECLARE_CHK_VARS;

    switch (format)
    {
    case SOURCE_UINT_U32LE:
    case SOURCE_UINT_U32BE:
	break;
	
    case SOURCE_UINT_S32BE:
	CHK_JAVA_S_TO_U(4);
	break;
	
    case SOURCE_UINT_U64BE:
#if   (NATIVE_UINT == U32LE) || (NATIVE_UINT == U32BE)
	CHK_U64BE_TO_U32();
#endif
	break;
    }

    return (-1);
}

int nexus_dc_check_lost_precision_long(nexus_byte_t *buffer,
				       unsigned long count,
				       int format)
{
    DECLARE_CHK_VARS;

    switch (format)
    {
    case SOURCE_LONG_S32LE:
    case SOURCE_LONG_S32BE:
	break;
	
    case SOURCE_LONG_S64LE:
#if   (NATIVE_LONG == S32LE) || (NATIVE_LONG == S32BE)
	CHK_S64LE_TO_S32();
#endif
	break;
	
    case SOURCE_LONG_S64BE:
#if   (NATIVE_LONG == S32LE) || (NATIVE_LONG == S32BE)
	CHK_S64BE_TO_S32();
#endif
	break;
    }

    return (-1);
}

int nexus_dc_check_lost_precision_u_long(nexus_byte_t *buffer,
					 unsigned long count,
					 int format)
{
    DECLARE_CHK_VARS;

    switch (format)
    {
    case SOURCE_ULONG_U32LE:
    case SOURCE_ULONG_U32BE:
	break;

    case SOURCE_ULONG_U64LE:
#if   (NATIVE_ULONG == U32LE) || (NATIVE_ULONG == U32BE)
	CHK_U64LE_TO_U32();
#endif
	break;
	
    case SOURCE_ULONG_U64BE:
#if   (NATIVE_ULONG == U32LE) || (NATIVE_ULONG == U32BE)
	CHK_U64BE_TO_U32();
#endif
	break;
	
    case SOURCE_ULONG_S64BE:
#if   (NATIVE_ULONG == U32LE) || (NATIVE_ULONG == U32BE)
	CHK_S64BE_TO_U32();
#elif (NATIVE_ULONG == U64LE) || (NATIVE_ULONG == U64BE)
	CHK_JAVA_S_TO_U(8);
#endif
	break;
    }

    return (-1);
}

int nexus_dc_check_lost_precision_float(nexus_byte_t *buffer,
					unsigned long count,
					int format)
{
    DECLARE_CHK_VARS;

    switch (format)
    {
    case SOURCE_FLOAT_IEEE32LE:
#if   NATIVE_FLOAT == CRAY64BE
	CHK_IEEE32LE_TO_CRAY64();
#endif
	break;
	
    case SOURCE_FLOAT_IEEE32BE:
#if   NATIVE_FLOAT == CRAY64BE
	CHK_IEEE32BE_TO_CRAY64();
#endif
	break;
	
    case SOURCE_FLOAT_CRAY64BE:
#if   (NATIVE_FLOAT == IEEE32LE) || (NATIVE_FLOAT == IEEE32BE)
	CHK_CRAY64BE_TO_IEEE32();
#endif
	break;
    }

    return (-1);
}

int nexus_dc_check_lost_precision_double(nexus_byte_t *buffer,
					 unsigned long count,
					 int format)
{
    DECLARE_CHK_VARS;

    switch (format)
    {
    case SOURCE_DOUBLE_IEEE64LE:
#if   NATIVE_DOUBLE == CRAY64BE
	CHK_IEEE64LE_TO_CRAY64();
#endif
	break;
	
    case SOURCE_DOUBLE_IEEE64BE:
#if   NATIVE_DOUBLE == CRAY64BE
	CHK_IEEE64BE_TO_CRAY64();
#endif
	break;
	
    case SOURCE_DOUBLE_CRAY64BE:
#if   (NATIVE_DOUBLE == IEEE64LE) || (NATIVE_DOUBLE == IEEE64BE)
	CHK_CRAY64BE_TO_IEEE64();
#endif
	break;
    }

    return (-1);
}
